#ifndef BIASED_RESIDUALS_PROCESSOR_H
#define BIASED_RESIDUALS_PROCESSOR_H 1

//LCIO
#include <lcio.h>

//MARLIN
#include "marlin/Processor.h"

namespace AIDA{
  class IHistogram1D;
  class IHistogram2D;
}


namespace marlintpc
{

    class TrackFitterFactory;
//    class TrackFitterBase;

    /** Fill the residuals of the hits into an AIDA histogram.
     *  Remember that LCIO::TRBIT_HITS has to be set, so the hits are stored.
     *  If ApplyCorrectionFactor is set the bias is corrected by \f$sqrt{n / (n-dof)}\f$,
     *  where dof is 2 for a straight line, 3 for a helix in xy, and 2 in z for all cases.
     *
     *  <h4>Input - Prerequisites</h4>
     *  \li A Collection of reconstructed tracks (LCIO::Track)
     *
     *  <h4>Output</h4>
     *  \li Histograms of the residuals in the xy-plane and in z
     *
     *  @param InputTrackes Name of input Tracks collection (default: TPCTracks)
     *  @param HistogramMinValueXY Minimum value in xy histogram (default: -0.3 mm)
     *  @param HistogramMaxValueXY Maximal value in xy histogram (default:  0.3 mm)
     *  @param HistogramMinValueZ  Minimum value in z histogram (default: -1.0 mm)
     *  @param HistogramMaxValueZ  Maximal value in z histogram (default:  1.0 mm)
     *
     *  @param ApplyCorrectionFactor Multiply residual with \f$sqrt{n / (n-dof)}\f$ (default: true)
     *  @param DoFXY Degrees of freedom in the XY plane (2 for straight line, 3 for helix) (default: 3)
     *
     *
     *  @author Martin Killenberg, University of Bonn
     */
    class BiasedResidualsProcessor : public marlin::Processor{
			      
    public:
    
	virtual Processor*  newProcessor() { return new BiasedResidualsProcessor ; }
    
	BiasedResidualsProcessor();

	~BiasedResidualsProcessor();

	virtual void init();

	virtual void processRunHeader( EVENT::LCRunHeader* run ) ;

	virtual void processEvent( LCEvent * evt ) ; 
    
    protected:

     std::string _inputColName ;    ///< the name of the input collection
     
     AIDA::IHistogram1D * _residualsXYHisto; ///< the residuals in the xy-plane
     AIDA::IHistogram1D * _residualsZHisto; ///< the residuals in z (time)
     
     AIDA::IHistogram2D * _residualsXY_vs_xr; ///< the residuals in the xy-plane vs x or r
     AIDA::IHistogram2D * _residualsXY_vs_yphi; ///< the residuals in the xy-plane vs y or phi     

     float  _histogramMinValueXY; ///< Minimum value in xy histogram
     float  _histogramMaxValueXY; ///< Maximum value in xy histogram

     float  _histogramMinValueZ; ///< Minimum value in z histogram
     float  _histogramMaxValueZ; ///< Maximum value in z histogram

     bool _applyCorrectionFactor;///< Multiply residual with \f$sqrt{n / (n-dof)}\f$ 
     int  _dofXY; ///<  Degrees of freedom in the XY plane 
    };
}

#endif // BIASED_RESIDUALS_PROCESSOR_H
