#ifndef LinearThreePointResolutionProcessor_h
#define LinearThreePointResolutionProcessor_h 1

#define N_1D_HISTOS_THREE_POINT_RESOLUTION_PROCESSOR 1
#define N_2D_HISTOS_THREE_POINT_RESOLUTION_PROCESSOR 0

// C++
#include <string>

// LCIO
#include "lcio.h"
#include <EVENT/TrackerHit.h>

// Marlin
#include "marlin/Processor.h"
#include "ZBinTemplateProcessor.h"


namespace AIDA{
  class IHistogram1D;
  class IHistogram2D;
}

namespace marlintpc{

    /* the following comment is tranfered to doxygen documentation
     * which begins with a second asterix, or in case of on line comment an extra third slash 
     */
    /**  Resoltuion calculator for straight tracks using the three point method.
     * 
     *   Determines the 3 point resolution.
     *   Only possible for straight line tracks!
     *
     *   \todo Currently only the histogram is filled. The resolution propper 
     *   has to be calculated manually.
     *
     *  <h4>Input - Prerequisites</h4>
     *  \li A collection of tracks (lcio::Track). <br>
     *  Remember that LCIO::TRBIT_HITS has to be set, so the hits are stored.
     *
     *  <h4>Output</h4> 
     *  \li A histogram with the 3 point residual distribution
     * 
     * @param InputCollectionName Name of input collection, TrackerRawData
     * 
     * @author Simone Zimmermann
     */
  
//  typedef lccd::ConditionsMap<tpcconddata::TimePixPixelMode::PixelKey, tpcconddata::TimePixPixelMode > TimePixModeMap ;
//  typedef std::map< tpcconddata::TimePixPixelMode::PixelKey, tpcconddata::TimePixPixelMode > stdTimePixModeMap;

  
    class LinearThreePointResolutionProcessor : 
    public ZBinTemplateProcessor<N_1D_HISTOS_THREE_POINT_RESOLUTION_PROCESSOR, 
	                         N_2D_HISTOS_THREE_POINT_RESOLUTION_PROCESSOR>{

//marlin::Processor {
	
    public:
	
	/* This method will be called by the marlin package
	 * It returns a processor of the currend type (here LinearThreePointResolutionProcessor)
	 */
	virtual Processor*  newProcessor() { return new LinearThreePointResolutionProcessor ; }
	
	/* the default constructor
	 * here the processor parameters are registered to the marlin package
	 * other initialisation should be placed in the init method
	 */
	LinearThreePointResolutionProcessor() ;
	
	/* Called at the beginning of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms
	 */
	virtual void init() ;
	
	/* Called for every run.
	 * in this processor it is not used
	 */
	virtual void processRunHeader( lcio::LCRunHeader* run ) ;
	
	/* Called for every event - the working horse.
	 * Here the real conversion take place
	 */
	virtual void processEvent( lcio::LCEvent * evt ) ; 
	
	/* This method is only called if the check flag is set (default) in the main processor
	 * (where the input files are given)
	 * It should be used to create check plots which are not needed in a complete analysis
	 * (where normaly the check flag is deactivated)
	 */
	virtual void check( lcio::LCEvent * evt ) ; 
	
	
	/* Called after data processing for clean up.
	 * e.g. saving control histogramms, delete created objects
	 * (which are not stored in a collection/event), etc.
	 */
	virtual void end() ;

	/* Sort function to sort the Hits along the track
	 */
	static bool compareHit(const std::pair<TrackerHit*, double> & pair1,
			       const std::pair<TrackerHit*, double> & pair2)
	  {
	    return pair1.second < pair2.second;
	  }

	
	
    protected:
	
	/* normally the doxygen documentation is palced before the method/variable
	 * but for one line comments in case of variables it may be more convenied for you
	 * to place it after the variable. In this case you use an arrow to tell doxygen
	 * to which variable the commet belongs to.
	 */
	
	std::string _inputColName ;  ///<- the name of the input collection
	int _chipID; ///<- number of current chip, 0 if single chip
	int _histoBin3P; ///<- d3P values will be binned with _histoBin3P
	int _histoBinTime; ///<- Time values will be binned with _histoBinTime
	int _histoBinNoTrack; ///<- Binning and range of histoNoTrack
	float _histoRange3P; ///<- d3P will have range (-_histoRange3P, _histoRange3P)
	float _histoRangeTime; ///<- Time values will have range (0, _histoRangeTime)

	/** The index of the template parameter for the hitsPerTrack histos
	 *  We want to have these values divided into z bins
	 */
	static const unsigned int _histo3PIndex;


	/** is true till the first output collection is wirten to an event
	 *  used to to store parameters only in the first collection
	 *  (also done, if collection is transient)
	 */
	bool isFirstCollection;



private:
	AIDA::IHistogram1D* histo3P; ///<-histogram of 3 point resolution
	AIDA::IHistogram2D* histo3P2D; ///<- 2D plot of d3P vs time value
	AIDA::IHistogram1D* histoNoTrack; ///<- histogram of number of tracks per event
    } ;
    
}

#endif
    
    

