#ifndef _SIMPLE_PEDESTAL_CALCULATOR_PROCESSOR_H
#define _SIMPLE_PEDESTAL_CALCULATOR_PROCESSOR_H

//C++
#include <map>
#include <utility>

//LCIO
#include <lcio.h>

//MARLIN
#include <marlin/Processor.h>

//TPCCondData
#include "Pedestal.h"

namespace marlintpc 
{
    
    /** The SimpePedestalCalculatorProcessor calculates 
     *  the pedestal for each channel from all events in the run.
     *  It just sums up all measured ADC values and calculates the mean
     *  value, so the data must not contain peaks. You need dedicated pedestal
     *  runs to record these data.
     *  The output is a simple lcio file with only one event containing the
     *  conditions data. Data base files or MySQL conditions data will not be
     *  supported.
     *
     *  Processor Parameters:
     *  @param InputCollectionName Name of the \c TrackerRawData input collection (default: TPCRawData)"
     *  @param OutputCollectionName Name of the output collection (default: TPCPedestals)
     *  @param OutputFileName Name of the lcio output file (existing file will be overwritten). 
     *         Default: TPCPedestals.slcio
     */

    class SimplePedestalCalculatorProcessor : public marlin::Processor
    {
			      
    public:
    
	virtual Processor*  newProcessor() { return new SimplePedestalCalculatorProcessor; }
    
	SimplePedestalCalculatorProcessor();

	/** Initialises the processor parameters
	 */
	virtual void init();
    
	/** Nothing to do here...	    
	 */
	virtual void processRunHeader( lcio::LCRunHeader* run );
    
	/** Sum up all pedestals in the pedestalCreator classes
	 */
	virtual void processEvent( lcio::LCEvent * evt ) ; 
    
    
	/** Nothing to check
	 */
	virtual void check( lcio::LCEvent * evt ) ;
    
    
	/** The end function writes the caclulated pedestals to the lcio file.
	 */
	virtual void end() ;
    
    protected:

	/** Internal helper class to sum up the pedestals and calculate the mean
	 *  value and width.
	 */
	class pedestalCreator
	{
	public:
	    pedestalCreator();

	    /// Add one measured ADC value to the calculation
	    void   addSample(int adcValue);

	    /// Get the number of samples that have been used to calculate the pedestal
	    size_t getNSamples() const;

	    /** Get the pedestal value.
	     *  It is \f$ (\sum_{i=0}^{n-1} x_i)/n \f$
	     */
	    double getPedestalValue() const;

	    /** Get the width of the pedestal.
	     *  It is \f$ (\sum_{i=0}^{n-1} x_i^2)/n - (\bar x)^2\f$
	     */
	    double getPedestalWidth() const;

	private:
	    size_t _nSamples; // The number number of samples that have been added
	    long long int _sumX; // The sum of all ADC values
	    long long int _sumXSquare; // The sum of the squares of all ADC values
	};

	/** For better readability: define a pedestal map type
	 */
	typedef std::map< std::pair<int,int> , pedestalCreator > PedestalCreatorMap_t;

	/** A map to associate a hardware ID (pair of ints for channel and readout group)
	 *  to the corresponding pedestal creator object.
	 */
	PedestalCreatorMap_t _pedestalCreatorMap;

	// the processor parameters
	std::string _inputColName, _outputColName, _outputFileName;

    };
}
#endif
