#ifndef ChargeDistributionProcessor_h
#define ChargeDistributionProcessor_h 1

#include "VoxelTPC.h"

//LCIO
#include "lcio.h"

//MARLIN
#include "marlin/Processor.h"

//C++
#include <string>

namespace marlintpc{




/* The following comment is tranfered to doxygen documentation
 * which begins with a second asterix, or in case of on line comment an extra third slash 
 */

/** This processor takes the coordinates of the drifted electrons 
 *  and the charge from amplification and smears the charge 
 *  on the padplane. Broadening of the charge cloud due to 
 *  diffussion between the GEMs is included as well as charge 
 *  transfer parameters in dependence on the GEM settings.
 *  The processor adds the transversale defocussing in the GEM stack 
 *  to the conditions data. If the conditions data is write
 *  protected and therefor the defocussing can not be added 
 *  a new writeable TPCConditions object is added to the data stream.
 *  The new object is a copy of the original TPCConditions and 
 *  additionally holds the value for the mean amplification of an single
 *  electron. Later processors should only use the newly created object.
 *
 *  Do not use the marlin::ConditionsProcessor to manage TPCConditions
 *  set or modified by the GEMProcessor!!!
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of SimTrackerHits with charge values set as dEdx
 *
 *  <h4>Output</h4> 
 *  Collection with TPCVoxels containing charges on pads
 *
 * @param InputCollectionName Name of input collection, SimTrackerHits
 * @param OutputCollectionName Name of output collection, TPCVoxel
 * @param Transient If not 0 the output collection is set transient
 * @param xBinLength Length of the bins in z direction in ns
 * @param ETrans1 Value of first transfer field in V/cm
 * @param ETrans2 Value of second transfer field in V/cm
 * @param EInd Value of Induction field in V/cm
 * @param Gas Gas mixture
 * @param BunchIntervall Time between two bunch crossings in ns
 * @param TPCConditions Name of the collection with TPCConditons to be read in 
 * @param AlternativeTPCConditionsName Name of newly created TPCCondtions collection if original object is writeable
 *
 *
 * @author Thorsten Krautscheid
 */




class ChargeDistributionProcessor : public marlin::Processor {
  
 public:
  
  virtual Processor*  newProcessor() { return new ChargeDistributionProcessor ; }
  
  
  ChargeDistributionProcessor() ;
  virtual ~ChargeDistributionProcessor(){}
  
  /** Called at the begin of the job before anything is read.
   *  Use to initialize the processor, e.g. book histograms.
   */
  virtual void init() ;
  
  /** Called for every run.
   */
  virtual void processRunHeader( EVENT::LCRunHeader* run ) ;
  
  /** Called for every event - the working horse.
   */
  virtual void processEvent( EVENT::LCEvent * evt ) ; 
  
  
  double CalculateTransverseDiffusion(double  EField);

  double CalculateGasParameter(double a, double b ,double c, double d, double e, 
                               double f, double g, double h, double i, double j, 
                               double k, double EField);
  
  double CalculateDiffTransBParam(double a, double b, double c, double d, double e);

protected:

  /** Input/output collection names.
   */
  std::string _inputCollectionName;///<Name of the input collection
  std::string _outputCollectionName;///<Name of the output collection
  std::string _tpcConditionsColName;///<Name of collection containing the TPCConditions
  std::string _altTPCConditionsName;///<Name of the newly created TPCConditonsCollection if the original one is not writeable
  std::string _gas;///<Name of the gas mixture

  float _ETrans1;///<First transfer field
  float  _ETrans2;///<Second transfer field
  float  _EInd;///<Induction field
  int _transient;///Values others than 0 set output collection to transient
  double _sigma;///<Value of the transverse defocussing
  float _BField;///<Strength of the magnetic field in T
  double _bunchIntervall;///<Time between two bunch crossings in ns
  double _zBinLength;///<Length of a z-bin in ns


  gear::PadRowLayout2D const * _padLayout;

  marlintpc::VoxelTPC* _trackerData;

  //For internal bookkeepin
  int _nEvt;
  int _nRun;


};

}//end of namespace marlintpc

#endif 
