#include <HitDistributor.h>

//C++
#include <cmath>
#include <complex>

//MARLIN
#include <streamlog/streamlog.h>


namespace marlintpc{

  //Constructor for FixedPadSizeDiskLayout
  HitDistributor::HitDistributor(gear::PadRowLayout2D  const* padLayout, 
                                 double zBinLength) {

    _padLayout     = padLayout;
    _zBinLength    = zBinLength;
    _padLayoutType = _padLayout->getPadLayoutType();
  }


  HitDistributor::~HitDistributor() {}

  // Steering function. The only function accessible from the outside.
  // Takes a hit for input and calls all other functios for the smearing
  // depending on the pad layout of the readout plane.
  std::vector<TPCVoxel> HitDistributor::distributeHit(SimTrackerHit* hit, 
						      double sigmaTrans, double sigmaLong){

    //Get hit values in cartesian coordinates
    _xHit = hit->getPosition()[0];
    _yHit = hit->getPosition()[1];   

    //Distribute charge of the hit according to the pad layout type (polar or rectangular geometry)
    switch( _padLayoutType ){
    case gear::PadRowLayout2D::POLAR: {
      std::vector<int> padIndexVec = getPadsPolar( sigmaTrans );
      if( padIndexVec.empty() )
	return std::vector<TPCVoxel>();
      else{
        return distributeHitOnPadsPolar( padIndexVec, hit, sigmaTrans, sigmaLong );
      }
      break;
    }
    case gear::PadRowLayout2D::CARTESIAN: {
      std::vector<int> padIndexVec = getPadsRect( sigmaTrans );
      if( padIndexVec.empty() )
	return std::vector<TPCVoxel>();
      else
        return distributeHitOnPadsRect( padIndexVec, hit, sigmaTrans, sigmaLong );
      break;
    }
    default:
      throw gear::Exception("Unknown pad layout type");
    };

    return std::vector<TPCVoxel>();    
  }

  // Gets all pads within the smearing radius and stores their indices 
  //in vector for circular pad layout. 
  std::vector<int> HitDistributor::getPadsPolar( double sigmaTrans ){

    std::vector<int> padIndexVec;

    std::complex<double> compHit( _xHit, _yHit );
    double rHit   = std::abs( compHit );
    double phiHit = std::arg( compHit );
    if( phiHit < 0 )
      phiHit = ( 2 * M_PI )+phiHit;

    int hitIndex;

    try{
       hitIndex = _padLayout->getNearestPad( rHit , phiHit );
    }catch(std::out_of_range& e){
      return padIndexVec;
    }

    int cursor = hitIndex;
    if( checkDistancePolar( hitIndex, sigmaTrans ) ){

      padIndexVec.push_back( hitIndex );
      padIndexVec = getPadsOnTheLeftPolar(  padIndexVec, hitIndex, sigmaTrans );
      padIndexVec = getPadsOnTheRightPolar( padIndexVec, hitIndex, sigmaTrans );
    }else{
      return padIndexVec;
    }

    //step up
    int row = _padLayout->getRowNumber( cursor ) ;

    while(1) {
      row++;
      if( row < _padLayout->getNRows() ) {
	try{
	  cursor = _padLayout->getNearestPad( _padLayout->getPadCenter( _padLayout->getPadIndex(row,0) )[0] , phiHit);
	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistancePolar( cursor, sigmaTrans) ) {
          padIndexVec.push_back( cursor );
	  padIndexVec = getPadsOnTheLeftPolar(  padIndexVec, cursor, sigmaTrans );
	  padIndexVec = getPadsOnTheRightPolar( padIndexVec, cursor, sigmaTrans );
	}else
	  break;
      }else
	break;
    }
    cursor = hitIndex;
    row = _padLayout->getRowNumber( cursor );

    //step down 
    while(1) {
      row--;
      if( row >= 0 ) {
	try{
	  cursor = _padLayout->getNearestPad(_padLayout->getPadCenter(_padLayout->getPadIndex(row,0))[0], phiHit);
	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistancePolar( cursor, sigmaTrans ) ) {
          padIndexVec.push_back( cursor );
	  padIndexVec = getPadsOnTheLeftPolar(  padIndexVec, cursor, sigmaTrans );
	  padIndexVec = getPadsOnTheRightPolar( padIndexVec, cursor, sigmaTrans );
	}else
	  break;
      }
      else
	break;
    }
    return padIndexVec;
  }

  // Gets all pads within the smearing radius and stores their indices 
  // in vector for rectangular pad layout. 
  std::vector<int> HitDistributor::getPadsRect( double sigmaTrans ){
    std::vector<int> padIndexVec;

    int hitIndex;
    try{
       hitIndex = _padLayout->getNearestPad( _xHit , _yHit );
    }catch(std::out_of_range& e){
      return padIndexVec;
    }

    int cursor = hitIndex;

    if( checkDistanceRect( hitIndex, sigmaTrans ) ){

      padIndexVec.push_back( hitIndex );
      padIndexVec = getPadsOnTheLeftRect(  padIndexVec, hitIndex, sigmaTrans );
      padIndexVec = getPadsOnTheRightRect( padIndexVec, hitIndex, sigmaTrans );

    }else
      return padIndexVec;

    int row = _padLayout->getRowNumber( cursor );

    // step up 
    while(1) {
      row++;
      if( row < _padLayout->getNRows() ) {
	try{
	  cursor = _padLayout->getNearestPad( _xHit , _padLayout->getPadCenter( _padLayout->getPadIndex(row,0) )[1] );
	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistanceRect( cursor, sigmaTrans) ) {
          padIndexVec.push_back( cursor );
	  padIndexVec = getPadsOnTheLeftRect(  padIndexVec, cursor, sigmaTrans );
	  padIndexVec = getPadsOnTheRightRect( padIndexVec, cursor, sigmaTrans );
	}else
	  break;
      }else
	break;
    }

    cursor = hitIndex;
    row = _padLayout->getRowNumber(cursor);

    // step down 
    while(1) {
      row--;
      if( row >= 0 ){
	try{
	  cursor = _padLayout->getNearestPad(_xHit, _padLayout->getPadCenter( _padLayout->getPadIndex(row,0)  )[1] );

	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistanceRect( cursor, sigmaTrans ) ){
	  padIndexVec.push_back( cursor );
	  padIndexVec = getPadsOnTheLeftRect(  padIndexVec, cursor, sigmaTrans );
	  padIndexVec = getPadsOnTheRightRect( padIndexVec, cursor, sigmaTrans );
	}else
	  break;
      }else
	break;
    }

    return padIndexVec;
  }

  // Gets the indices of all pads in the row, which are left from the given pad 
  //and wthin the smearing radius, and adds them to the vector which contains
  // all pad indices of hit. (for circular pad layout)
  std::vector<int> HitDistributor::getPadsOnTheLeftPolar(std::vector<int> padVec, 
							 int centerIndex, 
							 double sigmaTrans){
    int np = centerIndex;

    while(1){
      np = _padLayout->getLeftNeighbour( np );

      if( checkDistancePolar( np, sigmaTrans) && np != centerIndex )
	padVec.push_back( np );     
      else 
        return padVec;
    }
  }

  // Gets the indices of all pads in the row, which are left from the given pad 
  // and wthin the smearing radius, and adds them to the vector which contains
  // all pad indices of hit. (for rectangular pad layout)
  std::vector<int> HitDistributor::getPadsOnTheLeftRect( std::vector<int> padVec, int centerIndex, double sigmaTrans){
    int np = centerIndex;

    while(1){
      try{
        np = _padLayout->getLeftNeighbour( np );
      }catch(std::exception& e){   //exception handling has to be adjusted when the new gear version is installed!
	return padVec;
      }
      if( checkDistanceRect( np, sigmaTrans ) )
	padVec.push_back( np );
      else
	return padVec;
    }
  }

  // Gets the indices of all pads in the row, which are right from the given pad 
  //  and wthin the smearing radius, and adds them to the vector which contains
  //  all pad indices of hit. (for circular pad layout)
  std::vector<int> HitDistributor::getPadsOnTheRightPolar( std::vector<int> padVec, 
							   int centerIndex, 
							   double sigmaTrans){
    int np = centerIndex;

    while(1){
      np = _padLayout->getRightNeighbour( np );

      if( checkDistancePolar( np, sigmaTrans) && np != centerIndex )
	padVec.push_back( np );
      else
        return padVec;
    }
  }

  // Gets the indices of all pads in the row, which are right from the given pad 
  //  and wthin the smearing radius, and adds them to the vector which contains
  //  all pad indices of hit. (for rectangular pad layout)
  std::vector<int> HitDistributor::getPadsOnTheRightRect(std::vector<int> padVec, int centerIndex, double sigmaTrans){
    int np = centerIndex;

    while(1){
      try{
	np = _padLayout->getRightNeighbour( np );
      }catch(std::exception& e){
	return padVec;
      }
      if( checkDistanceRect( np, sigmaTrans) )
	padVec.push_back( np );
      else
	return padVec;
    }
  }

  // Checks whether the distance between the given pad and the hit is lower than the smearing 
  //  range. Returns true when any part of the pad is within the smearing radius, even when it 
  //  is only a little corner of the pad. (for circular pad layout)
  bool HitDistributor::checkDistancePolar( int padIndex, double sigmaTrans ){

    const double SMEAR = 4;

    double rad = _padLayout->getPadCenter( padIndex )[0];
    double phi = _padLayout->getPadCenter( padIndex )[1];

    std::complex<double> zPad = std::polar( rad, phi);
    double xPad = zPad.real();
    double yPad = zPad.imag();

    //Calculate half pad width and half pad height
    double halfPadWidth = _padLayout->getPadCenter(padIndex)[0]*tan( _padLayout->getPadWidth( padIndex ) / 2 );
    double halfPadHeight = _padLayout->getRowHeight( _padLayout->getRowNumber(padIndex) ) / 2 ;

    double uncorrDist = sqrt( pow(_xHit - xPad,2) + pow( _yHit-yPad,2) );

    if( (uncorrDist-std::min(halfPadWidth,halfPadHeight)) < SMEAR * sigmaTrans)
      return true;
     
    double xCloser = 0;
    double yCloser = 0;

    if( (xPad*yPad) >= 0 ){
      double xHeight = fabs(halfPadHeight * cos( phi ) );
      double xWidth  = fabs(halfPadWidth  * cos( ((M_PI) /2)-phi) );
      double yHeight = fabs(halfPadHeight * sin( phi ) );
      double yWidth  = fabs(halfPadWidth  * sin( (M_PI/2)-phi) );

      xCloser = std::max( xHeight, xWidth );
      yCloser = std::max( yHeight, yWidth );

    }else{
      double teta = M_PI - phi;

      double xHeight = fabs( halfPadHeight * cos( teta ) );
      double xWidth  = fabs( halfPadWidth * cos( (M_PI/2) - teta ) );
      double yHeight = fabs( halfPadHeight * sin( teta ) );
      double yWidth  = fabs( halfPadWidth * sin( (M_PI/2) - teta) );

      xCloser = std::max( xHeight, xWidth );
      yCloser = std::max( yHeight, yWidth );
    }

    if( xPad > _xHit )
      xPad = xPad - xCloser;
    else
      xPad = xPad + xCloser;

    if( yPad > _yHit )
      yPad = yPad - yCloser;
    else
      yPad = yPad + yCloser;

    double dist = sqrt( pow( _xHit-xPad, 2) + pow( _yHit-yPad,2) );

    if( dist < SMEAR * sigmaTrans )
      return true;
    else
      return false;
    
  }


  // Checks whether the distance between the given pad and the hit is lower than the smearing 
  //  range. Returns true when any part of the pad is within the smearing radius, even when it 
  //  is only a little corner of the pad. (for rectangular pad layout)
  bool HitDistributor::checkDistanceRect( int padIndex, double sigmaTrans ){

    const double SMEAR = 4;
    
    double xPad = _padLayout->getPadCenter( padIndex )[0];
    double yPad = _padLayout->getPadCenter( padIndex )[1];

    //claculate half pad width (with pitch!)
    int row = _padLayout->getRowNumber( padIndex );
    double halfPadWidth  = _padLayout->getPadWidth( padIndex ) / 2 ;
    if( _padLayout->getPadsInRow( row ).size() >= 2 ) {
      int one = _padLayout->getPadIndex( row, 0 );
      int two = _padLayout->getPadIndex( row, 1 );
      halfPadWidth = fabs( _padLayout->getPadCenter( one )[0] - _padLayout->getPadCenter( two )[0])/2.;
    }
    double halfPadHeight = _padLayout->getRowHeight( _padLayout->getRowNumber(padIndex) ) /2. ;

    if( sqrt( pow(fabs( _xHit - xPad )- halfPadWidth , 2)
             +pow(fabs( _yHit - yPad )- halfPadHeight ,2)) < (SMEAR*sigmaTrans) )
      return true;
    else
      return false;
  }


  //Smears the distributes the charge of the given hit in voxels according to the given pads, 
  //transversal and longitudinal diffusion for circular pad layout. 
  std::vector<TPCVoxel> HitDistributor::distributeHitOnPadsPolar( std::vector<int> padIndexVec, SimTrackerHit* hit, 
								  double sigmaTrans, double sigmaLong){

    std::vector<TPCVoxel> voxelVec;

    if( hit->getdEdx() > 0 ) {

      while(padIndexVec.begin()!=padIndexVec.end())  {

        int i = padIndexVec.front();

        //claculate the x,y coordinates of the pad center
        double xPad = _padLayout->getPadCenter(i)[0]*cos(_padLayout->getPadCenter(i)[1]);
        double yPad = _padLayout->getPadCenter(i)[0]*sin(_padLayout->getPadCenter(i)[1]);

	streamlog_out( DEBUG ) << "get pad coordinates " << std::endl;

        //rotate coordinate system so that the new x,y axis are parallel to pad widht,height
        double xHitRot =   _xHit*sin(_padLayout->getPadCenter(i)[1]) 
                         - _yHit*cos(_padLayout->getPadCenter(i)[1]);

        double yHitRot =    _xHit*cos(_padLayout->getPadCenter(i)[1]) 
                          - _yHit*sin(_padLayout->getPadCenter(i)[1]);
      

        double xPadRot =    xPad*sin(_padLayout->getPadCenter(i)[1]) 
                          - yPad*cos(_padLayout->getPadCenter(i)[1]);

        double yPadRot = +  xPad*cos(_padLayout->getPadCenter(i)[1]) 
                          - yPad*sin(_padLayout->getPadCenter(i)[1]);

	streamlog_out( DEBUG ) << "performed coordinate transformation " << std::endl;
      
        double halfPadWidth  = _padLayout->getPadCenter(i)[0]*tan(_padLayout->getPadWidth(i));/*_padLayout->getFixedPadWidth()/2;*/
        double halfRowHeight = _padLayout->getRowHeight(_padLayout->getRowNumber(i))/2;

	streamlog_out(DEBUG) << "calculated pad heigth and width: " << halfRowHeight << "  " << halfPadWidth << std::endl;

       //start and end values for the integration
	double xEnd  =xPadRot+halfPadWidth;
	double xStart=xPadRot-halfPadWidth;
	double yEnd  =yPadRot+halfRowHeight;
	double yStart=yPadRot-halfRowHeight;

        // integration over a 2d gaussian
        int charge= static_cast<int>(hit->getdEdx()/4 *((erf((xEnd  - xHitRot)/(sqrt(2.)*sigmaTrans))
                                                        -erf((xStart- xHitRot)/(sqrt(2.)*sigmaTrans)))
                                                      * (erf((yEnd  - yHitRot)/(sqrt(2.)*sigmaTrans))	
                                                        -erf((yStart- yHitRot)/(sqrt(2.)*sigmaTrans)))));

	streamlog_out(DEBUG) << "performed integration" << std::endl;

        int zBin = static_cast<int>( hit->getTime()  /_zBinLength ); 

        if( sigmaLong ){

	  const double tDiff = 4;

	  int minBin = static_cast<int>( (hit->getTime() - (tDiff*sigmaLong)) / _zBinLength  );
	  int maxBin = static_cast<int>( (hit->getTime() + (tDiff*sigmaLong)) / _zBinLength  );

	  if( minBin < 0 )
	    minBin = 0;

	  for(int j = minBin; j <= maxBin; j++){ 

	    int tDiffCharge = static_cast<int> ( charge / 2  * ( erf(((j+1)*_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))-
								 erf(( j   *_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))  ));

	    voxelVec.push_back( TPCVoxel(i,j,tDiffCharge) );
	  }
	  padIndexVec.erase( padIndexVec.begin() );
	}else{ 
          voxelVec.push_back( TPCVoxel(i,zBin,charge)); 
          padIndexVec.erase(padIndexVec.begin());
	}
      }
    }
    return voxelVec;
  }

  // Smears the distributes the charge of the given hit in voxels according to the given pads, 
  //transversal and longitudinal diffusion for rectangular pad layout. 
  std::vector<TPCVoxel> HitDistributor::distributeHitOnPadsRect( std::vector<int> padIndexVec, SimTrackerHit* hit, 
								 double sigmaTrans, double sigmaLong){

    std::vector<TPCVoxel> voxelVec;

    if( hit->getdEdx() > 0 ){

      //integration for all the stored pad indices
      while(padIndexVec.begin()!=padIndexVec.end()) {

        int i = padIndexVec.front();

	//calculate the width of the pad (with pitch!)
	int row = _padLayout->getRowNumber( i );
	double halfPadWidth = _padLayout->getPadWidth(i)/2;
	if( _padLayout->getPadsInRow( row ).size() >= 2) {
	  int one = _padLayout->getPadIndex( row, 0);
	  int two = _padLayout->getPadIndex( row, 1);
	  halfPadWidth = fabs( _padLayout->getPadCenter(one)[0] - _padLayout->getPadCenter(two)[0] ) /2.;
	}

        //start and end values for the integration
        double xStart, xEnd, yStart, yEnd; 
        if( _padLayout->getPadCenter(i)[0] < _xHit ) {
          xStart = _padLayout->getPadCenter(i)[0]+ halfPadWidth;
    	  xEnd   = _padLayout->getPadCenter(i)[0]- halfPadWidth;
        }else {
          xStart = _padLayout->getPadCenter(i)[0]- halfPadWidth;
	  xEnd   = _padLayout->getPadCenter(i)[0]+ halfPadWidth;
        }  
        if( _padLayout->getPadCenter(i)[1]<hit->getPosition()[1]) {
          yStart = _padLayout->getPadCenter(i)[1]+_padLayout->getRowHeight(_padLayout->getRowNumber(i))/2;
	  yEnd   = _padLayout->getPadCenter(i)[1]-_padLayout->getRowHeight(_padLayout->getRowNumber(i))/2;
        }else {
	  yStart = _padLayout->getPadCenter(i)[1]-_padLayout->getRowHeight(_padLayout->getRowNumber(i))/2;
	  yEnd   = _padLayout->getPadCenter(i)[1]+_padLayout->getRowHeight(_padLayout->getRowNumber(i))/2;
        }

        // integration over a 2d gaussian
        int charge= static_cast<int>(round(hit->getdEdx()/4 *((erf((xEnd  - _xHit)/(sqrt(2.)*sigmaTrans))
                                                              -erf((xStart- _xHit)/(sqrt(2.)*sigmaTrans)))
                                                            * (erf((yEnd  - _yHit)/(sqrt(2.)*sigmaTrans))	
                                                              -erf((yStart- _yHit)/(sqrt(2.)*sigmaTrans))))));
	charge = static_cast<int>(fabs(charge));   

        int zBin = static_cast<int>( hit->getTime()  /_zBinLength ); 

        if( sigmaLong ){

	  const double tDiff = 4;

	  int minBin = static_cast<int>( (hit->getTime() - (tDiff*sigmaLong)) / _zBinLength  );
	  int maxBin = static_cast<int>( (hit->getTime() + (tDiff*sigmaLong)) / _zBinLength  );

	  if( minBin < 0 )
	    minBin = 0;

	  for(int j = minBin; j <= maxBin; j++){ 

	    int tDiffCharge = static_cast<int> ( charge * ( erf(((j+1)*_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))-
                                                            erf(( j   *_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))  ));


	    voxelVec.push_back( TPCVoxel(i,j,tDiffCharge) );
	  }
	  padIndexVec.erase( padIndexVec.begin() );
	}else{ 
          voxelVec.push_back( TPCVoxel(i,zBin,charge)); 
          padIndexVec.erase(padIndexVec.begin());
	}
      } 
    }
    return voxelVec;

  }



} //namespace marlintpc
