#include "MokkaToVoxelProcessor.h"


//MarlinTPC
#include "LCObjectCopier.h"

//tpcconddata
#include "TPCConditions.h"

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <IMPL/SimTrackerHitImpl.h>


//GEAR
#include <gear/TPCParameters.h>

//Marlin
#include <marlin/Global.h>
#include <marlin/Exceptions.h>

//C++
#include <cmath>

using namespace std;
using namespace lcio;
using namespace marlintpc;
using namespace marlin;
using namespace tpcconddata;


namespace marlintpc{


  MokkaToVoxelProcessor aMokkaToVoxelProcessor ;


  MokkaToVoxelProcessor::MokkaToVoxelProcessor() : Processor("MokkaToVoxelProcessor"), _MokkaHitsInVoxelTPC(0), _padLayout(0) {
  
    _description = "MokkaToVoxelProcessor transforms Mokkahits in Voxels ..." ;
  
  
    // register steering parameters: name, description, class-variable, default value


    registerInputCollection(    LCIO::SIMTRACKERHIT,
				"InputCollectionName" , 
				"Name of the Input collection from Mokka (default: MokkaHits)"  ,
				_inputCollectionName ,
				string("MokkaHits") ) ;

    registerInputCollection(    LCIO::LCGENERICOBJECT,   
				"TPCConditionsCollection" ,
				"Name of the collection containing the TPCConditions (default: TPCConditions::getDefaultCollectionName())",
				_conditionsColName,
				TPCConditions::getDefaultColName() );
    
    registerOutputCollection(   LCIO::LCGENERICOBJECT ,
				"OutputCollectionName" ,
				"Name of the output voxel collection (default: MokkaHitsInVoxels)" ,
				_outputCollectionName ,
				string("MokkaHitsInVoxels") );
    
    registerProcessorParameter( "Transient" ,
				"Set write mode of output (default: 0)" ,
				_transient ,
				0 );
    
    registerProcessorParameter( "ReadoutFrequency" ,
				"Frequency of the readout electronics in MHz (default: 40)",
				_readoutFrequency,
				40. );

    registerProcessorParameter( "SigmaElectronics",
				"Width of the gaussian electronics shaping (default: 0)",
				_sigmaElectronics,
				0.);

    registerProcessorParameter( "Amplification",
				"Total amplification of charges (default: 1)",
				_amplification,
				1.);
   
    registerProcessorParameter( "IonisationEnergy" ,
				"The electron in GeV needed to ionize a gas particle (default: 26.E-9) ",
				_ionisationEnergy,
				26.E-9); //default value is for argon

    registerOptionalParameter(  "DriftVelocity",
				"Optional: electron drift velocity (in [mm/mus]) in case it is not provided by TPCConditions",
				_vDrift,
				0. );

    registerOptionalParameter(  "TransverseDiffusion",
				"Optional: transverse diffusion (in [mm/sqrt(mm)]) in case it is not provided by TPCConditions",
				_diffTrans,
				24. );

    registerOptionalParameter(  "LongitudinalDiffusion",
				"Optional: longitudinal diffusion (in [mm/sqrt(mm)]) in case it is not provided by TPCConditions",
				_diffLong,
				24. );

    registerOptionalParameter(  "TransverseDefocussing",
				"Optional: transverse defocussing (in []) in case it is not provided by TPCConditions ",
				_defocusTrans,
				24. );

    
  }



  void MokkaToVoxelProcessor::init() { 

    streamlog_out(MESSAGE2) << _description << endl;

    //Get pad layout from GEAR
    _padLayout = &(Global::GEAR->getTPCParameters().getPadLayout());

    //Get max drift distance from gear
    _maxDriftLength = Global::GEAR->getTPCParameters().getMaxDriftLength ();
    
    //Calculate length of time bins in [ns]
    double timeBin = 1. / ( _readoutFrequency * 0.001 );

    //Initialize a VoxelTPC
    if( _MokkaHitsInVoxelTPC )
      delete _MokkaHitsInVoxelTPC;
    _MokkaHitsInVoxelTPC = new marlintpc::VoxelTPC( _padLayout, timeBin );

    //Usually a good idea
    printParameters() ;

  }


 
  void MokkaToVoxelProcessor::processRunHeader( LCRunHeader* run) {
    
    //Revision logging
    run->parameters().setValue( _processorName + "_revision", "$Rev$" );
  
    //Parameter logging
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
    
      if ( ! i->second->isOptional() || i->second->valueSet() ) 
	run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
    }
    
  } 

 

  void MokkaToVoxelProcessor::processEvent( LCEvent * evt ) {

 
    //Get the TPCConditions
    try{

      LCCollection* tpcConditionsCol = evt->getCollection( _conditionsColName );
      TPCConditions myTPCConditions( tpcConditionsCol->getElementAt( 0 ) );


      //TPCConditions are available but can be overwritten by processor
      //parameters. Give a warning in this case.

      //Get drift velocity in [mnm/mus]
      if(  parameterSet("DriftVelocity") ){ 

	streamlog_out( WARNING0 ) << _processorName 
				  << ": WARNING: drift velocity from conditions data is overwritten by processor parameter!"
				  <<std::endl
				  << "Remove parameter \"DriftVelocity\" from steering file to use drift velocity from TPCCondData."
				  <<std::endl;
      }else
	_vDrift=myTPCConditions.getDriftVelocity();


      //Get transverse diffusion in [  ]
      if( parameterSet("TransverseDiffusion") ){ 

	streamlog_out( WARNING0 ) << _processorName 
				  << ": WARNING: transverse diffusion from conditions data is overwritten by processor parameter!"
				  <<std::endl
				  << "Remove parameter \"TransverseDiffusion\" from steering file to use transverse diffusion from TPCCondData."
				  <<std::endl;
      }else
	_diffTrans = myTPCConditions.getTransDiffusionCoef();

      
      //Get transverse defocussing in [  ]
      if( parameterSet("TransverseDefocussing") ){

	streamlog_out( WARNING0 ) << _processorName 
				  << ": WARNING: transverse defocussing from conditions data is overwritten by processor parameter!"
				  <<std::endl
				  << "Remove parameter \"TransverseDefocussing\" from steering file to use transverse defocussing from TPCCondData."
				  <<std::endl;
      }else
	_defocusTrans = myTPCConditions.getTransDefocussing();
      

      //Get longitudinal diffusion in [  ]
      if( parameterSet("LongitudinalDiffusion") ){

	streamlog_out( WARNING0 ) << _processorName 
				  << ": WARNING: longitudinal diffusion from conditions data is overwritten by processor parameter!"
				  <<std::endl
				  << "Remove parameter \"LongitudinalDiffusion\" from steering file to use longitudinal diffusion from TPCCondData."
				  <<std::endl;
      }else
	_diffLong = myTPCConditions.getLongDiffusionCoef();
      
  
      

    }catch( DataNotAvailableException &e ){

      //All values have to be set either by conditions data or as processor parameter, 
      //throw an exception if any value is not set.
       if( !parameterSet("DriftVelocity") )
       {  
	 streamlog_out( ERROR2 ) << "DriftVelocity has to be set "
				 <<"either in conditions data or "
				 <<"as processor parameter!" << std::endl;
	 throw DataNotAvailableException("No DriftVelocity");
       }

       if( !parameterSet("LongitudinalDiffusion") )
       {  
	 streamlog_out( ERROR2 ) << "LongitudinalDiffusion has to be set "
				 <<"either in conditions data or "
				 <<"as processor parameter!" << std::endl;
	 throw DataNotAvailableException("No LongitudinalDiffusion");
       }

       if( !parameterSet("TransverseDiffusion") )
       {  
	 streamlog_out( ERROR2 ) << "TransverseDiffusion has to be set "
				 <<"either in conditions data or "
				 <<"as processor parameter!" << std::endl;
	 throw DataNotAvailableException("No TransverseDiffusion");
       }

       if( !parameterSet("TransverseDefocussing") )
       {  
	 streamlog_out( ERROR2 ) << "TransverseDefocussing has to be set "
				 <<"either in conditions data or "
				 <<"as processor parameter!" << std::endl;
	 throw DataNotAvailableException("No TransverseDefocussing");
       }


    }// catch DataNotAvailableException


   
    //Set sigma for the transverse broadening of the charge
    _sigmaTrans = sqrt( pow(_diffTrans,2) + pow(_defocusTrans,2) );

    //Set sigma for the longitudinal broadening of the charge
    _sigmaLong = sqrt( pow(_diffLong,2) + pow(_sigmaElectronics,2) );


    try{
      
      //Get input collection with SimTrackerHits from Mokka
      LCCollection* inputMokkaHitsCol = evt->getCollection( _inputCollectionName );
        
      //Loop over all entries in input collection
      for(int i =0; i < inputMokkaHitsCol->getNumberOfElements() ; i++)
      {

	//To get the ionsiation caused by an energy deposition
	//dE/dx of the Mokka hits has to be divided by the mean 
	//ionizatin energy. Therefore get a copy of the Mokka
	//hit to reset its dE/dx to the corresponding charge
	//value
	SimTrackerHitImpl* myMokkaHit = LCObjectCopier::copy( dynamic_cast<SimTrackerHitImpl*>(inputMokkaHitsCol->getElementAt(i)) );
	
	//Calculate charge corresponding to the energy of the Mokka hit
	int charge = static_cast<int>( myMokkaHit->getdEdx() / _ionisationEnergy  );
	
	//Reset Mokka dEdx value to amplified charge
	myMokkaHit->setdEdx( _amplification * charge );

	//Set time to drift time plus time of creation after the beginning of the event in [ns]
	myMokkaHit->setTime( myMokkaHit->getTime() + 1000 * ( (_maxDriftLength - myMokkaHit->getPosition()[2]) / _vDrift ) );
	
	//Put charge in voxels
	_MokkaHitsInVoxelTPC->putChargeInVoxels( myMokkaHit , _sigmaTrans, _sigmaLong );     
      }

      //Get output collection with all charges binned in voxels
      LCCollectionVec* outputVoxelColVec = _MokkaHitsInVoxelTPC->fetchVoxelCollection();

      //Add output collection to event
      if( outputVoxelColVec != 0 )
	evt->addCollection( outputVoxelColVec , _outputCollectionName );

    }catch(DataNotAvailableException& ex){

      streamlog_out( ERROR ) << "An exception occurred: " << ex.what() << endl;
    }
  
  }



  void MokkaToVoxelProcessor::check( LCEvent * evt ) { 

    //Nothing to do here 

  }



  void MokkaToVoxelProcessor::end(){ 

    //Nothing to do here

  }


} //end of namespace marlintpc
