//CLASS
#include <VoxelTPC.h>

//MARLIN
#include "streamlog/streamlog.h"


using namespace marlin ;

namespace marlintpc{

  /** Constructor of VoxelTPC
   */
  VoxelTPC::VoxelTPC(gear::PadRowLayout2D const * padLayout , double lengthOfZBin){

    _lengthOfZBin = lengthOfZBin;
    
    // Instantiate HitDistributor.
    _distributeHitPtr = new HitDistributor( padLayout, _lengthOfZBin );
  }

  /** Destructor of VoxelTPC
   */
  VoxelTPC::~VoxelTPC(){
  
    delete _distributeHitPtr;
  }

  /** Constructor of VoxelIndex
   */
  VoxelTPC::VoxelIndex::VoxelIndex(int _padIndex, int  _zBin) {

    padIndex = _padIndex;
    zBin     = _zBin;
  }

  /** Destructor of VoxelIndex
   */
  VoxelTPC::VoxelIndex::~VoxelIndex(){}


  /** Returns a pointer on the collction which contains all voxels with a 
   *  zBin below or equal to the given outputLength. These voxels are 
   *  deleted from the map. Default outputLength is set to zero. If 
   *  outputLength is zero all voxels in the map are returned (and deleted).
   */
  LCCollectionVec* VoxelTPC::fetchVoxelCollection( double outputLength ) {

    if(voxelMap.empty()){
      streamlog_out(DEBUG4) << "no entries in voxel map!" << std::endl;
      return 0;
    }

    /** Instanciate the output collection.
     */
    float zBinLength=_lengthOfZBin;
    LCCollectionVec* voxelPtr = new LCCollectionVec(LCIO::LCGENERICOBJECT);
    voxelPtr->parameters().setValue("zBinLength",zBinLength);

    /** All voxels with a zBin smaller or equal to this bin will be added to the collection.
     */
    int outputZBin;

    /** Get the zBin according to the given output length. If output length is set
     *  to zero set outputZBin to the highest zBin in the map in order to get all
     *  entries. 
     */
    if( outputLength >= 0 )
      outputZBin = getZBin( outputLength );      
    else       
      outputZBin = voxelMap.rbegin()->first.zBin;   
    
    /** Add all voxels with zBin smaller or equal to the output zBin to the output collection.
     *  Delete all voxels added to the collection from the map.
     */
    for(std::map<VoxelIndex, int>::iterator iter = voxelMap.begin();
	(iter->first.zBin <= outputZBin) && iter != voxelMap.end();
	iter++)
      {
	TPCVoxel* myVoxel = new TPCVoxel(iter->first.padIndex, iter->first.zBin, iter->second);
	voxelPtr->addElement( myVoxel);
	voxelMap.erase(iter);
      }
    return voxelPtr;
  }


  /** Returns a pointer on the collection which contains all voxels with a zBin below or equal
   *  to the given outputLength. Default outputLength is set to zero. If outputLength is zero all
   *  voxels in the map are returned.
   */
  LCCollectionVec* VoxelTPC::getVoxelCollection( double outputLength )  {

    if(voxelMap.empty()){
      streamlog_out(DEBUG4) << "no entries in voxel map!" << std::endl;
      return 0;
    }

    /** Instanciate the output collection.
     */
    LCCollectionVec* voxelPtr = new LCCollectionVec(LCIO::LCGENERICOBJECT);

    float zBinLength=_lengthOfZBin;
    voxelPtr->parameters().setValue("zBinLength",zBinLength);

    /** All voxels with a zBin smaller or equal to this bin will be added to the collection.
     */
    int outputZBin;

    /** Get the zBin according to the given output length. If output length is set
     *  to zero set outputZBin to the highest zBin in the map in order to get all
     *  entries. 
     */
    if( outputLength >= 0)
      outputZBin = getZBin( outputLength );
    else 
      outputZBin = voxelMap.rbegin()->first.zBin;

    /** Add all voxels with zBin smaller or equal to the output zBin to the output collection.
     */
    for(std::map<VoxelIndex, int>::iterator iter = voxelMap.begin();
	(iter->first.zBin <= outputZBin) && iter != voxelMap.end();
	iter++)
      {
	TPCVoxel* myVoxel = new TPCVoxel(iter->first.padIndex, iter->first.zBin, iter->second);
	voxelPtr->addElement( myVoxel);	
      }
    return voxelPtr;
  }


  /** Puts charges in voxels according to a the input SimTrackerHit. Smearing is done in the x,y-plane
   *  according to transversal diffusion and if wanted also in z according to longitudinal diffusion
   */
  void VoxelTPC::putChargeInVoxels(SimTrackerHit *hit, double sigmaTrans, double sigmaLong = 0) {

    std::vector<TPCVoxel> voxelVec = _distributeHitPtr->distributeHit( hit, sigmaTrans, sigmaLong);

    for(unsigned int i = 0; i < voxelVec.size(); i++) 
      putChargeInVoxel( voxelVec[i] );
	
    voxelVec.clear();
  }


  /** Puts charge in given voxel
   */
  void VoxelTPC::putChargeInVoxel(TPCVoxel const &voxel){

    if(voxel.getCharge())
      voxelMap[ VoxelIndex( voxel.getPadIndex(), voxel.getTimeIndex() )] += voxel.getCharge();
  }


  /** Determines the zBin corresponding to the given z length and then 
   *  puts charge in voxel with given padIndex and zBin
   */
  void VoxelTPC::putChargeInVoxel(int padIndex, double zLength, int charge) {

    int zBin = getZBin( zLength );
    if(charge)
      putChargeInVoxel( TPCVoxel( padIndex, zBin, charge) );
  }


  /**returns voxel with given padIndex and zBin
   */
  TPCVoxel VoxelTPC::getVoxel(int padIndex, int zBin) {

    return( TPCVoxel( padIndex, zBin , voxelMap.find(VoxelIndex(padIndex,zBin))->second) );
  }


  /**returns vector containing all voxels with a given padIndex
   */
  std::vector<TPCVoxel> VoxelTPC::getVoxelsOnPad(int padIndex)  {

      streamlog_out(DEBUG2)<<"----------------------------"<<std::endl;
      streamlog_out(DEBUG2)<<"VoxelTPC::getVoxelsOnPad "<<std::endl;

    std::vector<TPCVoxel> aTPCVoxelVec;

    int maxTimeBin=voxelMap.rbegin()->first.zBin;

    streamlog_out(DEBUG1) << "Maximum zBin: " << maxTimeBin << std::endl;

    streamlog_out(DEBUG1) << " map.size() = " <<voxelMap.size()<<std::endl;
    
    streamlog_out(DEBUG1) <<" zBin begin: "<<voxelMap.begin()->first.zBin
			  <<" end: "<<voxelMap.rbegin()->first.zBin<<std::endl;

    streamlog_out(DEBUG1) <<" padIndex begin: "<<voxelMap.begin()->first.padIndex
			  <<" end: "<<voxelMap.rbegin()->first.padIndex<<std::endl;
    
    for(int i = voxelMap.begin()->first.zBin; 
	i <= voxelMap.rbegin()->first.zBin && voxelMap.begin() != voxelMap.end(); 
	i++ ){

      std::map<VoxelIndex,int>::iterator it = voxelMap.find( VoxelIndex(padIndex,i) );
      
      if ( i%10000 == 0 ) 
	  streamlog_out(DEBUG0)<<"i = "<<i<<std::endl;

      if(it!=voxelMap.end())
        aTPCVoxelVec.push_back( TPCVoxel(padIndex,i,it->second) );

    }
    /**The old code
    for(std::map<VoxelIndex, int>::iterator iter = voxelMap.begin(); iter != voxelMap.end(); iter++){
      
      if(padIndex== iter->first.padIndex)
	aTPCVoxelVec.push_back(TPCVoxel( padIndex, iter->first.zBin, iter->second));
    }
    */

    streamlog_out(DEBUG2)<<"VoxelTPC::getVoxelsOnPad  done!! "<<std::endl;
    streamlog_out(DEBUG2)<<"----------------------------"<<std::endl;

    return(aTPCVoxelVec);
  }





  /** Rreturns vector containing all voxels with a given zBin.
   */
  std::vector<TPCVoxel> VoxelTPC::getVoxelsWithZBin(int zBin){

    std::vector<TPCVoxel> aTPCVoxelVec;

    for(std::map<VoxelIndex,int>::iterator iter = voxelMap.lower_bound(VoxelIndex(zBin,0));
	*iter < *voxelMap.lower_bound(VoxelIndex(0, zBin+1));iter++)
      aTPCVoxelVec.push_back(TPCVoxel(iter->first.padIndex, zBin, iter->second));
      
    return(aTPCVoxelVec);
  }

  void VoxelTPC::addCollection(LCCollection* voxelCol){

    if(voxelCol==0)
      return;

    double colZBinLength=voxelCol->getParameters().getFloatVal("zBinLength");
  
    for(int i =0; i < voxelCol->getNumberOfElements(); i++) {
  
      TPCVoxel* myVoxel = new TPCVoxel(dynamic_cast<LCGenericObject*>( voxelCol->getElementAt(i) ));
  
      double zLength = myVoxel->getTimeIndex()*colZBinLength;

      putChargeInVoxel(myVoxel->getPadIndex(),zLength,myVoxel->getCharge());
    }
  }



  /** Removes the given voxel from the map.
   */
  void VoxelTPC::removeVoxel(TPCVoxel const &voxel) {

    removeVoxel(voxel.getPadIndex() , voxel.getTimeIndex());   
  }


  /**removes the voxel with given padIndex and zBin from the map.
   */
  void VoxelTPC::removeVoxel(int padIndex, int zBin) {

    voxelMap.erase(VoxelIndex(padIndex, zBin));
  }    


  /** Removes all voxels in the given vector from the map.
   */
  void VoxelTPC::removeVoxels(std::vector<TPCVoxel> const &voxelVec){
    for(unsigned int i=0; i<voxelVec.size(); i++) 
      removeVoxel(voxelVec[i].getPadIndex(), voxelVec[i].getTimeIndex());
  }


  /** Removes all voxels with given padIndex from the map.
   */
  void VoxelTPC::removeVoxelsOnPad(int padIndex) {

    for( std::map<VoxelIndex, int>::iterator iter = voxelMap.lower_bound(VoxelIndex(padIndex,0)) ; 
	 *iter < *voxelMap.lower_bound(VoxelIndex(padIndex+1,0));
	 iter++) 
       voxelMap.erase(iter);
  }


  /** Removes all voxels with given zBin from the map.
   */
  void VoxelTPC::removeVoxelsWithZBin(int zBin) {

    for( std::map<VoxelIndex, int>::iterator iter = voxelMap.begin(); 
         iter != voxelMap.end(); 
         iter++) 
      if(iter->first.zBin == zBin) {voxelMap.erase(iter);}
  }


  /** Returns the number of the zBin corresponding to the given length.
   */
  int VoxelTPC::getZBin( double zLength ){

    int zBin = static_cast<int>( zLength/_lengthOfZBin );
    return zBin;
  }


} //end of namespace marlintpc

