#ifndef CHANNELCORRECTIONHANDLER_H
#define CHANNELCORRECTIONHANDLER_H

#include <IMPL/LCCollectionVec.h>
#include <ChannelCorrection.h>

namespace marlintpc {

/** 
 *  A helper class for convenient usage of the channel correction information.
 *  It allows to obtain information about calibration constants, time offsets,
 *  broken or noisy channel, etc.
 *
 *  @author Peter Wienemann, University of Bonn
 */
class ChannelCorrectionHandler {
 public:

  /**
   * Constructor
   *
   * @param corrCol collection containing ChannelCorrection objects
   */
  ChannelCorrectionHandler( EVENT::LCCollection* corrCol = 0 );

  /**
   * Set channel correction collection
   *
   * @param channel correction collection
   */
  void setChannelCorrectionCollection( EVENT::LCCollection* corrCol );
    
  /**
   * Get channel quality
   *
   * @param channelID hardware channel ID
   *
   * @return integer with channel quality bits
   */
  int getChannelQuality( int channelID );

  /**
   * Get number of calibration factors
   * (there can be an arbitrary number of calibration factors per channel)
   *
   * @param channelID hardware channel ID
   *
   * @return number of calibration factors
   */
  int getNChannelCalibrationFactors( int channelID );

  /**
   * Get calibration factor
   * (there can be an arbitrary number of calibration factors per channel)
   *
   * @param channelID hardware channel ID
   * @param index index of calibration factor
   *
   * @return channel calibration factor
   */
  float getChannelCalibrationFactor( int channelID, int index = 0 );

  /**
   * Get calibration factor vector
   * (there can be an arbitrary number of calibration factors per channel)
   *
   * @param channelID hardware channel ID
   *
   * @return vector with calibration factors
   */
  std::vector<float> getChannelCalibrationFactors( int channelID );

  /**
   * Get channel time offset
   *
   * @param channelID hardware channel ID
   *
   * @return time offset in ns
   */
  float getChannelTimeOffset( int channelID );

  /**
   * Get information about channel operability
   *
   * @param channelID hardware channel ID
   *
   * @return true if channel is broken
   */
  bool isBroken( int channelID );

  /**
   * Get information about noise in channel
   *
   * @param channelID hardware channel ID
   *
   * @return true if channel is noisy
   */
  bool isNoisy( int channelID );

 private:

  IMPL::LCCollectionVec _sortedChanCorrColl; ///< channel correction collection sorted by channel ID

  /**
   * Get iterator to ChannelCorrection object with given channel ID
   *
   * @param channelID hardware channel ID
   */
  IMPL::LCCollectionVec::iterator   getChannelCorrection( int channelID );

  /**
   * A nested helper class for sorting ChannelCorrection
   * objects according to channel ID
   */
   class  sortChannelID {
   public:
     bool operator()( EVENT::LCObject* chancorr1, EVENT::LCObject* chancorr2 ) {
       return ( tpcconddata::ChannelCorrection(chancorr1).getChannel() <
		tpcconddata::ChannelCorrection(chancorr2).getChannel() );
     }
   };
   
   /**
    * A nested helper class for comparing an int to a ChannelCorrection
    * object's channel ID
    */
  class  cmpChannelID {
  public:
    bool operator()( EVENT::LCObject* chancorr1, int chanID ) {
      return ( tpcconddata::ChannelCorrection(chancorr1).getChannel()
	       < chanID );
    }
  };

};

} // namespace marlintpc

#endif // CHANNELCORRECTIONHANDLER_H
