#ifndef PedestalSubtractorProcessor_h
#define PedestalSubtractorProcessor_h 1

// C++
#include <string>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "marlin/Exceptions.h"

#ifdef USE_LCCD
//LCCD
#include "lccd/IConditionsChangeListener.hh"
#endif


// MarlinTPC
#include "PedestalHandler.h"
#include "ADCElectronicsHandler.h"


/// the MarlinTPC Package: reconstruction of TPC data
namespace marlintpc{

    /** Pedestal subtracting processor of the MarlinTPC package.
     * 
     *  Thisprocessor substracts pedestals provided via LCCD (simple file, database file or database).
     *
     *  <h4>Input - Prerequisites</h4>
     *  Collection of TrackerData
     *
     *  <h4>Output</h4> 
     *  Collection of TrackerData.
     * 
     * @param InputCollectionName Name of input data collection
     * @param OutputCollectionName Name of output data collection
     * @param SetOutputTransient If not 0 the output collection is set transient
     * @param PedestalCollectionName Name of the collection containing the pedestals
     * @param MaximumADCValue Maximum ADC Value of the read-out electronic to detect overflow samples, if this parameter is set it overides this information from the conditions database (optional parameter)
     **/
     /** @param ElectronicsParametersObjectName Name of the conditions object in which the parameters of the read out electronics is stored. For example "GenericADCElectronicsParameters". **/
     /** @param ElectronicsParametersCollectionName Name of the conditions collection in which the parameters of the read out electronics is stored **/
     /**
     *
     * @author M. E. Janssen, DESY; R. Diener, DESY
     */
                
     class PedestalSubtractorProcessor : public marlin::Processor {
    
        public:
        
        virtual Processor*  newProcessor() { return new PedestalSubtractorProcessor ; }
        
        PedestalSubtractorProcessor() ;
        
        virtual void init() ;
        
        virtual void processRunHeader( lcio::LCRunHeader* run ) ;
        
        virtual void processEvent( lcio::LCEvent * evt ) ; 
        
        virtual void check( lcio::LCEvent * evt ) ; 
        
        virtual void end() ;
        
        
        protected:
        
        /// the name of the input collection            
        std::string _inputColName ;    
        /// the name of the output collection
        std::string _outputColName ;  
        /// the name of the pedestal collection 
        std::string _pedestalColName ;
        /// the name of the collection with the read out electronics parameters
        std::string _electrParaObjName ;
        /// the name of the collection with the read out electronics parameters
        std::string _electrParaColName ;
    
        /** give the status which is set to the transient flag of the output collection
        *  it is an int instead of a boolean, because the processor parameter cannot be a boolean
        */
        int _outputIsTransient;
        
        /** is true till the first output collection is written to an event
        *  used to to store parameters only in the first collection
        *  (also done, if collection is transient)
        */
        bool isFirstCollection;
    
        /// counts the processed runs
        int _nRun ; 
        /// counts the processed events
        int _nEvt ; 
        
        /// maximum ADC value, needed for overflow detection
        float _maximumADCValue ;
        
        private:
                     
        /// Handler which provides the pedestals
        PedestalHandler* _pedestalHandler; 
        ADCElectronicsHandler* _adcElectronicsHandler; 

     } ;
}

#endif



