#ifndef PULSEFINDERPROCESSOR_H
#define PULSEFINDERPROCESSOR_H

//C++
#include <string>

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "EVENT/TrackerData.h"
#include <IMPL/TrackerDataImpl.h>
#include "EVENT/TrackerPulse.h"

//Marlin
#include "marlin/Processor.h"

// MarlinTPC
#include "PedestalHandler.h"

namespace marlintpc
{
  /** This processor creates lcio::TrackerPulse objects from lcio::TrackerData
   *
   *  First the pulses are searched using a threshold based method, where the start and end thresholds are parameters.
   *  Then these candidates are further processed, by splitting multiple pulses.
   *  In a last step quality flags are assigned to account for
   *    - multiple pulses (changed shape/charge information)
   *    - overflows, where the ADC values are saturated
   *    - anomalous shape, currently only too long pulses (FIXME: now 3 times the minimal pulse length)
   *  For the storage the integrated charge in ADC values is computed and the pulse time; 
   *  the pulse time can be computed in two ways (choose by a parameter): 
   *   (default:) the point of change in curvature in the rising slope
   *        or:   the charge weighted average/center of gravity
   *  As a default also the initial TrackerData (=ADC) spectrum is stored (choose by parameter),
   *  which is always forced when any quality flag is set.
   *
   *  <h4>Input - Prerequisites</h4>
   *  Collection of TrackerData (can be zero supressed)
   *
   *  <h4>Output</h4>
   *  Collection of TrackerPulses
   *  Collection of TrackerData which belong to a pulse
   *
   *  @param InputCollectionName name of the input TrackerData collection
   *  @param OutputTrackerPulseCollectionName name of the output TrackerPulse collection
   *  @param OutputTrackerDataCollectionName  name of the output TrackerData collection
   *  @param PedestalCollectionName name of the collection containing the pedestal information
   *  @param PedestalCollectionFile name of the file containing the pedestals (if not provided via LCCD)
   *  @param ForceSpectrumSave if not 0 the ADC spectrum will be stored for all pulses
   *  @param PulseStartThreshold a pulse start if ADC value is the given value times width of the pedestal
   *  @param PulseEndThreshold the pulse ends if the ADC value drops below this value
   *  @param DoMultiplePulseSplitting multiple pulses are split if a change in curvature of the adc values exists
   *  @param MinimumPulseHeight to accept a pulse its maximal ADC entry must exceed this cut
   *  @param MinumumPulseLength to accept a pulse its length in time bins must be larger than this value without Start/End bins
   *  @param SaveNBinsBeforeStart # bins saved before the start bin (exept there is another pulse)
   *  @param SaveNBinsAfterStart # bins saved after the start bin (exept there is another pulse)
   *  @param ReadoutFrequency frequency of the readout in Hz (only needed if not povided by GEAR) (FIXME: this should be condDB data)
   *  @param UseChargeAveragedPulseTime (optional) set this to use weighted mean for pulse time determination
   *
   * @author Ch. Rosemann, DESY
   */
  class PulseFinderProcessor : public marlin::Processor
  {
    
  public:
    
    virtual Processor*  newProcessor() { return new PulseFinderProcessor; }
    
    PulseFinderProcessor();

    virtual void init();
    
    virtual void processRunHeader(lcio::LCRunHeader* run);
    
    virtual void processEvent(lcio::LCEvent* evt);
    
    virtual void check(lcio::LCEvent* evt);

    virtual void end();
    
  protected:
    
    int _nRun;          //< run counter
    int _nEvt;          //< event counter
	      
    ///< name of the collection, where input data are stored
    std::string  _inputColName;
    ///< name of the collection, where TrackerData are stored
    std::string  _outputDataColName;
    ///< name of the collection, where TrackerPulse are stored
    std::string  _outputPulseColName;
    ///<- the name of the pedestal collection
    std::string  _pedestalColName;
    
    // the following is only needed if compiled without LCCD
#ifndef USE_LCCD
    //<- the name of the pedestal file
    std::string _pedestalFileName;
#endif
    
    /// set the status which is set to the transient flag of the output collection
    bool _outputIsTransient;
    
    /// if not 0 the ADC spectrum will be saved for all pulses
    bool _forceSpectrumSave;
    
    /** set if you want the default pulse time calculation (by using the sign change of the derivative of the charge spectrum)
	or if set 'true' then the charge weighted average is used
    */
    bool _useChargeAveragePulseTime;
    
    /// HACK: (LCCD problem for now)
    /// if set, then this will be the noise value, which is used for the pulse finding
    /// additionally this turns (_FOR NOW_) the pedestal handling OFF
    /// _this is subject to change !!!_
    int _noiseOverrideValue;

    //< set wether multiple pulses are split (default) or not
    bool _doMultiplePulseSplitting;
    //< start threshold to detect a pulse in adc values
    float _pulseStartThreshold;
    //< end threshold to detect a pulse in adc values
    float _pulseEndThreshold;
    //< minimal value for the maximum adc value
    float _minimumPulseHeight;
    //< minimal length for a pulse in number of time bins
    int _minimumPulseLength;
    //< # of bins stored before the start threshold
    int _saveNBinsBeforeStart;
    //< # of bins stored after the end threshold
    int _saveNBinsAfterEnd;
    //< maximum adc value, needed to determine if a pulse is maxed out
    float _maximumADCValue;
    //< readout frequency in Hz (note: the time will be stored in ns)
    double _readoutFrequency;
    
    //< will provide the acutal pedestal width
    PedestalHandler* _pedestalHandler;
    
    /** Find pulses by threshold in the ADC information and closer inspection.
     *  This is the somewhat "master" function that determines all aspects of the pulses,
     *  it calls some further functions for easier reading.
     *  @param thisData 'raw' data as TrackerData
     *  @param tempPulse a helper data construct, a pair of the TrackerData of a pulse and the quality bit
     *  @param noiseValue the width of the pedestal
     */
    
    void findPulses(lcio::TrackerData* thisData, std::vector<std::pair<lcio::TrackerDataImpl*, int> >* tempPulse, float noiseValue);
    /** The first function that is called during the pulse finding,
	it simply looks for values above the start threshold and below the end thresholds,
	the start and end positions are stored by their iterators in the original TrackerData ADC value vector.
	Here already pulses which are too short are suppressed.
	@param ADCSpectrum the ADC spectrum to be searched
	@param theCandidatePulses a vector of the pairs of start and end positions of the pulse candidates (as iterators)
    */
    
    void brutePulseSearch(std::vector<float>& ADCSpectrum, 
			  std::vector<std::pair<std::vector<float>::iterator, std::vector<float>::iterator> >* theCandidatePulses);
    
    /** The second part in the pulse finding process.
	Check the pulse candidates for possible multiple pulses by looking more closely at the ADC spectrum.
	Then replace the /old/ pairs of start/end iterators with the appropriate new number of pairs.
	TODO: adjust the charges in the bins at the joined boundary of the split pulses 
	@param theCandidatePulses a vector of the pairs of start and end positions of the pulse candidates (as iterators)
	@param noiseValue the pedestal width
    */
    
    void splitMultiplePulses(std::vector<std::pair<std::vector<float>::iterator, std::vector<float>::iterator> >* 
			     theCandidatePulses, float noiseValue);
    
    /** indicates  wether the electronics polarity is negative
     */
    bool _negativePolarity;
    
    /** Calculate the time information of the given pulse spectrum.
     *  Two different ways are implemented, which can be chosen by a processor parameter.
     *  The default way is the point of change in the curvature of the rising slope of the pulse.
     *  The second possibility is to take the charge weighted mean of the time (center of gravity).
     *  @param thisPulse ADC spectrum from which the time will be calculated
     *  @param NoiseValue variation smaller than this value will be ingnored during the calculation
     *  @return the time information of the pulse in ns
     */
    float calcPulseTime(lcio::TrackerData* thisPulse, float noiseValue = 0.0);
    
    /** Calculate the integrated charge of the given pulse by summing up all bins of the ADC spectrum.
     *  The polarity of the pulse is taken into acount
     *  @param thisPulse ADC spectrum for charge calculation
     */
    float calcPulseCharge(lcio::TrackerData* thisPulse);
  };
}
#endif // PULSEFINDERPROCESSOR
