#ifndef ROWBASEDHITFINDERPROCESSOR_H
#define ROWBASEDHITFINDERPROCESSOR_H

// MARLIN
#include "marlin/Processor.h"
#include "marlin/Global.h"

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "IMPL/LCCollectionVec.h"
#include "EVENT/TrackerPulse.h"

//GEAR
#include "gear/TPCParameters.h"

// C++
#include <string>
#include <vector>
#include <map>

namespace marlintpc
{
  
/** A hit finder for row based (pad) geometries 
 *
 *  @author Christoph Rosemann, DESY
 *
 *  <h4>Input - Prerequisites</h4>
 *  \li TrackerPulse collection
 *  \li tpcconddata::TPCConditions for the drift velicity (optionally provided as processor parameter)
 *  \todo Handling of dead and noisy channels.
 *
 *  <h4>Output</h4>
 *  \li TrackerHit collection
 *
 *  @param InputTrackerPulses Name of input TrackerPulses collection (default: TPCPulses)
 *  @param OutputTrackerHits Name of output TrackerHits collection (default: TPCHits)
 *  @param InputTPCConditions \c OPTIONAL: Name of the input collection containing the TPC 
 *         conditions data (default: tpcconddata::TPCConditions::getDefaultColName() )
 *  @param SetOutputHitsTransient If true the output collection is set transient (default: false)
 *  @param MaxEmptyPads Maximum number of consecutive empty pads in a row (default: 1)
 *  @param MinHitSize Minimum size of hit in number of pads (default: 1)
 *  @param MaxTimeSpread Maximum time between pulses in a hit [in ns] (default: 200)
 *  @param VDrift \c OPTIONAL: Set drift velocity in case there is no conditions data [in mm/us] (default: 0)
 *  @param DeadPads \c OPTIONAL: a vector of geometry indices of dead channels, default should be conditions data
 *
 * List of open issues/questions:
 * - error calculation/storage
 * - hit quality flag: definition of bit map (?) and LCIO implementation
 * - extend debug output
 * - incorporate damaged pad info
 * - incorporate inheritance of pulse quality
 * - separation of hits?
 * +++ needs testing!
 */

  class RowBasedHitFinderProcessor : public marlin::Processor
  {
    
  public:
    
    virtual Processor*  newProcessor() { return new RowBasedHitFinderProcessor; }
    
    RowBasedHitFinderProcessor() ;
    
    /** Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;

    /** Called for every run.
     */
    virtual void processRunHeader( EVENT::LCRunHeader* run ) ;
    
    /** Called for every event - the working horse.
     */
    virtual void processEvent( EVENT::LCEvent * evt ) ; 
    
    
    virtual void check( EVENT::LCEvent * evt ) ; 
    
    
    /** Called after data processing for clean up.
     */
    virtual void end() ;
    
  private:
    /** the name of the collection the input pulses will be stored under
     */
    std::string _inputTrackerPulsesCollectionName ;
    
    /** the name of the collection the output hits will be stored under
     */
    std::string _outputTrackerHitsCollectionName ;
    
    /** the name of the collection the TPC conditions are stored with
     */
    std::string _inputTPCConditionsCollectionName;
    
    /** a list of dead pads, given in the GearPadIndex,
     *  for now: just a single module
     * this is deprecated from the beginning, 
     * just a workaround until the conditions object is there
     */
    std::vector<int> _deadPadIndices;

    /** Give the status which is set to the transient flag of the output collection.
     *  It is an int instead of a boolean, because the processor parameter cannot be a boolean.
     *  Hits and tracks can be set separately.
     */
    bool _outputHitsTransient;    //< if true, the hits will not be stored

    /** maximum number of empty pads (noisy/dead) in a row
     */
    int _maxEmptyPads;
    
    /** minimum size of hit in number of pads
     */
    int _minHitSize;

    /** minimum height for the maximal pulse in a hit to be accepted
     */
    float _minMaxPulseSliceHeight;
    
    /** maximum time between pulses belonging to a single hit [in ns]
     */
    float _maxTimeSpread;
    
    /** drift velocity in [mm/mu s]
     */
    float _driftVelocity;

    /** conversion factor from ADC counts to number of primary electrons
     */
    float _chargeConversionFactor;

    // a helper class -- does it make sense to externalise it from the HitFinder?
    class hitCandidate
    {
    public:
      hitCandidate(EVENT::TrackerPulse*);
      
      void addPulse(EVENT::TrackerPulse*);
      void addDeadPad();

      void setBoundaryFlag();
      
      
      const EVENT::TrackerPulse* getMaximumPulse() const { return _maxPulse; };
      double getCharge() const;
      double getMaxPulseHeight();
      const std::vector<EVENT::TrackerPulse*>& getPulses() const { return _thePulses; };
      unsigned int getWidth() const;
      bool   isAtBoundary() {return _isAtBoundary;};

    private:
      hitCandidate() {;};
      std::vector<EVENT::TrackerPulse*> _thePulses;
      EVENT::TrackerPulse* _maxPulse;
      double _maxPulseCharge;
      int _numberOfDeadPads;
      bool _isAtBoundary;
    };
    
    /** helper functions:
     */
    void _sortPulses(EVENT::LCCollection*, const gear::TPCParameters&, 
		     std::map<int, std::map<int, std::vector<EVENT::TrackerPulse*>* >* >&);
    void _findHitCandidates(std::map<int, std::map<int, std::vector<EVENT::TrackerPulse*>* >* >&,
			    const gear::TPCParameters&,
			    std::map<int, std::vector<RowBasedHitFinderProcessor::hitCandidate> >&);
    void _createHitCollection(std::map<int, std::vector<RowBasedHitFinderProcessor::hitCandidate> >&,
						const gear::TPCParameters&,IMPL::LCCollectionVec*);

    bool _pulseComparator(EVENT::TrackerPulse*,EVENT::TrackerPulse*);
    EVENT::TrackerPulse* _findHighestPulseInRow(std::vector<EVENT::TrackerPulse*>&);
    bool _maxPulseHeightCheck(EVENT::TrackerPulse*);
    std::vector<EVENT::TrackerPulse*> _getNeighbourPulses(EVENT::TrackerPulse*, const gear::TPCParameters&,
							  std::vector<EVENT::TrackerPulse*>&,bool&);
    std::vector<EVENT::TrackerPulse*>::iterator _findBestNeighbourPulseMatch
      (std::vector<std::vector<EVENT::TrackerPulse*>::iterator>&,float);

    bool _hitIsValid(const RowBasedHitFinderProcessor::hitCandidate&);
    double _calculateFixedCoordinate(const RowBasedHitFinderProcessor::hitCandidate&,const gear::TPCParameters&);
    double _calculateFreeCoordinate(const RowBasedHitFinderProcessor::hitCandidate&,const gear::TPCParameters&);
    double _calculateZCoordinate(const RowBasedHitFinderProcessor::hitCandidate&);
    double _calculateFixedError(const RowBasedHitFinderProcessor::hitCandidate&,const gear::TPCParameters&);
    double _calculateFreeError(const RowBasedHitFinderProcessor::hitCandidate&,const gear::TPCParameters&);
    double _calculateZError(const RowBasedHitFinderProcessor::hitCandidate&);
    double _convertCharge(const RowBasedHitFinderProcessor::hitCandidate&);
  };
}//namespace marlintpc
#endif//ROWBASEDHITFINDERPROCESSOR_H
