#ifndef TRACK_FITTER_FACTORY_H
#define TRACK_FITTER_FACTORY_H
#include <EVENT/LCParameters.h>
#include <Exceptions.h>
#include <map>

namespace EVENT
{
    class LCRunHeader;
    class LCEvent;
}

namespace marlintpc{

    class TrackFitterBase;

/**  The TrackFitterFactory is an interface to handle track fitters.
 *
 *
 *   Analysis processors use the \c calculateResiduals function of the track fitters 
 *   so this code does not have to be implemented twice. Especially for analysis processors the 
 *   abstract interface means that the analysis code is independent from the track fitter used,
 *   but all the specialities of a track fitter like geometry dependence, dependence on the gas mixture
 *   and whatever the special fitter requires are taken into accout.
 *   
 *   The getTrackFitter function is called in every event.
 *   It is passed an instance of LCParametes,
 *   which are the collection parameters of the collection containing the tracks.
 *   These parameters contain
 *   information about the parameters the track fitter might need.
 *   The TrackFitter is only created once and with every new call the parameters are reset.
 *   This code is neither reentrant nor thread safe.
 *
 *   Every developer of a new track fitter has to add his processor to the getTrackFitter function.
 */

class TrackFitterFactory
{
 public:

    /** Get a track fitter and reinitialise its parameters.
     *  A fitter of the given type is returned
     *  Parameters for the fitter are defined in the LCParameters.
     *
     *  Some fitter types might need information from the conditions data or
     *  the full collection of hits to revisit which hit belongs to the track.
     *  This is transferred by giving a pointer to the event.
     *  If the specific fitter type does not need the event it can be ommitted.
     *  Note: The fitter shall not modify the event, this it is a pointer to a const
     *  object. !TO BE DISCUSSED!
     *  Throws !!!FIXME what kind of exception in case the fitter need conditon data 
     *  but \c event = 0.
     */
    static TrackFitterBase * getTrackFitter(unsigned char fitterType,
					    EVENT::LCParameters const * collectionParameters,
					    EVENT::LCEvent const * event = 0);


    /** Define a new exception class for unkonwn fitter types.
     *  It's empty, just a new type for catching.
     */
    class FitterTypeException : public EVENT::Exception
    {
      public:
        FitterTypeException(std::string const & initString);
    };
    
 private:
    /// The private constructor. It cannot be called, everything is static
    TrackFitterFactory(){}
    
    /// The private copy constructor. Never used, but implemented to prevent usage 
    // (default copy copystructor is public).
    TrackFitterFactory( TrackFitterFactory const & ){}
};


} // namespace marlintpc

#endif // TRACK_FITTER_FACTORY_H
