#ifndef TRACKFITTERLIKELIHOODPROCESSOR_H
#define TRACKFITTERLIKELIHOODPROCESSOR_H 1

#include "marlin/Processor.h"

#include "TrackFitterBase.h"

#include "lcio.h"
#include <string>
#include <vector>

#include "IMPL/TrackImpl.h"

/* NOTE this uses the C++ version of Minuit - not the Fortran version */
#include "Minuit2/FCNBase.h"

namespace AIDA {
	class ITuples;
}

namespace EVENT {
	class Track;
}

namespace marlintpc{

/** TrackFitterLikelihoodProcessor : This class takes post-amplified data from a readout pad and
 * reconstructs the track based on the location of each signal. The algorithm was developed by Dean Karlen at UVic
 *    @author J. Abernathy (UVic)<br>
 */
class TrackFitterLikelihoodProcessor : public marlin::Processor, public ROOT::Minuit2::FCNBase {
  
public:

	/** Default constructor
	*/
	TrackFitterLikelihoodProcessor();


	/** Destructor
	*/
	~TrackFitterLikelihoodProcessor();


	/** Return a new instance of this processor
	*/
	Processor * newProcessor();


	/** Called at the begin of the job before anything is read.
	*/
	void init();


	/** Called once per run to process the event's header
	*/
	void processRunHeader (lcio::LCRunHeader *run);


	/** Called for every event - the working horse.
	*/
	virtual void processEvent(lcio::LCEvent *evt);


	// From FCNBase
	/**
		The meaning of the vector of parameters is of course defined by the user,
		who uses the values of those parameters to calculate his function value.
		The order and the position of these parameters is strictly the one specified
		by the user when supplying the starting values for minimization. The starting
		values must be specified by the user, either via an std::vector<double> or the
		MnUserParameters supplied as input to the MINUIT minimizers such as
		VariableMetricMinimizer or MnMigrad. Later values are determined by MINUIT
		as it searches for the minimum or performs whatever analysis is requested by
		the user.

		@param par function parameters as defined by the user.

		@return the value of the function.

		@see MnUserParameters
		@see VariableMetricMinimizer
		@see MnMigrad
	*/
	virtual double operator()(const std::vector<double>& x) const;

	/**
	Error definition of the function. MINUIT defines parameter errors as the
	change in parameter value required to change the function value by up. Normally,
	for chisquared fits it is 1, and for negative log likelihood, its value is 0.5.
	If the user wants instead the 2-sigma errors for chisquared fits, it becomes 4,
	as Chi2(x+n*sigma) = Chi2(x) + n*n.
	\todo Comment a little bit better with links!!!!!!!!!!!!!!!!! Idem for errorDef()
	*/
	virtual double Up() const;


	/** Called for every event - right after processEvent() has been called for all processors.
	*/
	virtual void check (lcio::LCEvent *evt);


	/** Called after data processing for clean up in the inverse order of the init() method so that resources allocated in the first processor also will be available for all following processors.
	*/
	virtual void end();


private:

	/** Returns a Track which contains the expected number of electron in the pulse data (get total charge) of the trackerhits of the track. It uses the parameters of the input track to calculate the likelihood and to determine which pads to use (it uses the pads we have data for). It <b>ONLY</b> works for FixedPadSizeDiskLayout!!!!!!
	*/
	IMPL::TrackImpl calculate_expected_track(const EVENT::Track&, const double) const;

	/** A helper function for calculate_expected_distribution
	*/
	IMPL::TrackImpl calculate_expected_track_polar(const EVENT::Track&, const double) const;

	/** A helper function for calculate_expected_distribution
	*/
	IMPL::TrackImpl calculate_expected_track_cartesian(const EVENT::Track&, const double) const;
	
	/* function for fitting in (r,phi), (x,y) */
	
	/** Calculates the likelihood of observing this distribution of charge on the readout pads in the readout geometry plane (r-phi/x-y)
	*/
	const double calculate_charge_distribution_log_likelihood(const EVENT::Track&) const;

	/* functions for fitting in s-z */

	/** Calculates the likelihood of observing this distribution of primary ionizations in the z direction with a specified z deviation
	*/
	const double calculate_primary_ionization_location_log_likelihood(const EVENT::Track&, const double) const;


	/** Uses MINUIT to find the most likely track which fits the data supplied in the seed track
	*/
	IMPL::TrackImpl find_most_likely_fit(EVENT::Track* seed_track);

	/** Calculates the expected distribution of charge on one pad by the line-gaussian parameterized using Dean Karlen's calculations. see "TPC Performance in Magnetic Fields with GEM and Pad Readout"
	* b is the distance from the centre of the pad to the track in the x dimension (negative is left) (mm)
	* phi is the local azimuthal angle of the straight line segment in the pad row (relative to [0,1,0], clockwise is negative)
	* sigma is the standard deviation of the distribution
	* height is the height of the pad (mm)
	* width is the width of the pad (mm)
	*/
	double calculate_expected_distribution(double b, double phi, double sigma,
		double height, double width) const;

	/** The name of the LCCollection to obtain the seed track from.
	*/
	std::string _seed_track_collection_name;

	/** The name of the LCCollection to store the fitted track into the event with.
	*/
	std::string _fitted_track_collection_name;

	/** the name of the collection the TPC conditions are stored with
	*/
	std::string _tpc_conditions_collection_name;

	/** The constant diffusion of the GEMs and similar TPC structures which does not vary depending on distance traveled
	*/
	double _sigma0;

	/** Should _sigma_0 be a free parameter in the minimization?
	* true : free parameter
	* false : constant (default)
	*/
	bool _fit_sigma0;

	/** The transverse diffusion constant of the gas in the TPC in sqrt(mm)
	*/
	double _gas_trans_diffusion_constant;

	/** The transverse defocussing constant of the gas in the TPC in mm
	*/
	double _gas_trans_defocussing_constant;

	/** A noise factor to reduce the affect of spurious signals
	*/
	double _p_noise;

	/** The track being minimized currently
	*/
	EVENT::Track* _current_seed_track;

	/** The innermost position of the readout electronics
	**/
	double _maximum_drift_length;
	
	/** The standard deviation of the primary ionizations in the z-direction
	*/
	double _sigma_z;

	/** a parameter which chooses whether the track tuples are produced
	 */
	bool _produce_tuples;

	/** the track parameters stored as tuples
	*/
	AIDA::ITuple *_track_result_tuples;
};

} // namespace marlintpc

#endif // TRACKFITTERLIKELIHOODPROCESSOR_H
