#ifndef TRACKERRAWDATATODATACONVERTERPROCESSOR_H
#define TRACKERRAWDATATODATACONVERTERPROCESSOR_H

// C++
#include <string>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"

namespace marlintpc
{

  /* the following comment is tranfered to doxygen documentation
   * which begins with a second asterix, or in case of on line comment an extra third slash 
   */

  /**  raw data converting Processor of the MarlinTPC package .
   * 
   *   converts TrackerRawData into TrackerData without processing the data
   *
   *  <h4>Input - Prerequisites</h4>
   *  a collection of TrackerRawData
   *
   *  <h4>Output</h4> 
   *  Collection containing TrackerData.
   * 
   * @param InputCollectionName Name of raw data collection
   * @param OutputCollectionName Name of data collection
   * @param SetOutputTransient If not 0 the output collection is set transient
   * 
   * @author M. E. Janssen, DESY, Ch. Rosemann, DESY
   * @version $Id: TrackerRawDataToDataConverterProcessor.h,v 1.1 2006/07/07 08:52:04 mjanssen Exp $ 
   */

  class TrackerRawDataToDataConverterProcessor : public marlin::Processor
  {
  public:
    
    /* This method will be called by the marlin package
     * It returns a processor of the currend type (here TrackerRawDataToDataConverterProcessor)
     */
    virtual Processor* newProcessor() { return new TrackerRawDataToDataConverterProcessor; }
    
    /* the default constructor
     * here the processor parameters are registered to the marlin package
     * other initialisation should be placed in the init method
     */
    TrackerRawDataToDataConverterProcessor();
    
    /* Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms
     */
    virtual void init();
    
    /* Called for every run.
     * in this processor it is not used
     */
    virtual void processRunHeader(lcio::LCRunHeader* run);
    
    /* Called for every event - the working horse.
     * Here the real conversion take place
     */
    virtual void processEvent(lcio::LCEvent* evt);
    
    /* This method is only called if the check flag is set (default) in the main processor
     * (where the input files are given)
     * It should be used to create check plots which are not needed in a complete analysis
     * (where normaly the check flag is deactivated)
     */
    virtual void check(lcio::LCEvent* evt);
    
    
    /* Called after data processing for clean up.
     * e.g. saving contol histograms, delete created objects
     * (which are not stored in a collection/event), etc.
     */
    virtual void end();
    
  protected:
    
    /* normally the doxygen documentation is placed before the method/variable
     * but for one line comments in case of variables is may be more convenient to you
     * to place it after the variable. In this case you use an arrow to tell doxygen
     * to which variable the comment belongs to.
     */
    
    std::string _inputColName;  // the name of the input collection
    std::string _outputColName; // the name of the output collection
    
    /** give the status which is set to the transient flag of the output collection
     */
    bool _outputIsTransient;
    
    int _nRun; // counts the runs processed
    int _nEvt; // counts the events processed
  };
} // class declaration
#endif
    
    

