#ifdef USE_LCCD

#include "CountsToPrimaryElectronsProcessor.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

// tpcconddata
#include "TPCConditions.h"
#include "ChannelCorrection.h"


using namespace lcio ;
using namespace marlin ;

namespace marlintpc {


CountsToPrimaryElectronsProcessor aCountsToPrimaryElectronsProcessor;


CountsToPrimaryElectronsProcessor::CountsToPrimaryElectronsProcessor() : Processor("CountsToPrimaryElectronsProcessor"),

	_input_tracker_pulses_collection_name("TPCPulses"),
	_output_tracker_pulses_collection_name("TPCPulses"),
	_tpc_conditions_collection_name( tpcconddata::TPCConditions::getDefaultColName() ),
	_channel_corrections_collection_name( tpcconddata::ChannelCorrection::getDefaultColName() ), 
	_output_is_transient(true) {

	/* description of the processor which will be displayed in the
	* steering file autometicly generated by
	* marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	*/

	_description = "CountsToPrimaryElectronsProcessor converts FADC counts into primary electrons, the conversion factors are provided by LCCD. Gain corrections are provided by the GainCorrectorProcessor";

	/* register steering parameters: name, description, class-variable, default value
	* the type will by definded by the type of the default value
	* string, double, float and int are possible
	* plaese note to get a float you must cast in C++: (float) 1.3
	*/
	registerInputCollection( LCIO::TRACKERDATA ,
					"InputTrackerPulses",
					"Name of the input collection as stored in the LCIO file",
					_input_tracker_pulses_collection_name,
					std::string("TPCPulses") );

	registerOutputCollection( LCIO::TRACKERDATA ,
					"OutputTrackerPulses",
					"Name of the collection to be stored in the LCIO file",
					_output_tracker_pulses_collection_name,
					std::string("TPCPulses") );

	registerOptionalParameter( "SetOutputTransient",
				    "if not 0 the output collection is set transient",
				    _output_is_transient,
				    true);

	registerOptionalParameter( "TPCConditionsCollection" ,
				    "Name of the collection containing the TPC conditions",
				    _tpc_conditions_collection_name,
				    tpcconddata::TPCConditions::getDefaultColName() );
					
	registerOptionalParameter( "ChannelCorrectionCollection" ,
				    "Name of the collection containing the channel corrections",
				    _channel_corrections_collection_name,
				    tpcconddata::ChannelCorrection::getDefaultColName() );
}


void CountsToPrimaryElectronsProcessor::init() {
	
	// usually a good idea to
	printParameters();
	
	_number_of_runs = 0;
	_number_of_events = 0;
}
    

void CountsToPrimaryElectronsProcessor::processRunHeader( LCRunHeader* run) {
	
	_number_of_runs++ ;
}

void CountsToPrimaryElectronsProcessor::processEvent( LCEvent * evt ) {
	
	// get the input collection form the event
	try {


	    LCCollection* input_pulses = evt->getCollection( _input_tracker_pulses_collection_name );
	    //		LCCollection* tpc_conditions_collection = evt->getCollection( _tpc_conditions_collection_name );

		//retrieve the channel corrections
		LCCollection* channel_correction_collection = evt->getCollection( _channel_corrections_collection_name );
		
		tpcconddata::ChannelCorrection channel_correction( channel_correction_collection->getElementAt(0) );

		const double fC_per_electron = 1.6022E-04;
		
		// these calibration factors are specific to the UVic TPC simulation. They will change for others
		const double amp_mV_per_fC = channel_correction.getChannelCalibrationFactor(1);
		const double adc_mV_per_chan = channel_correction.getChannelCalibrationFactor(3);

		const double electronic_correction_factor = adc_mV_per_chan / (amp_mV_per_fC * fC_per_electron);

		streamlog_out( MESSAGE3 ) << "LCCD parameters: amp mv per fC = " << amp_mV_per_fC
			<< ", adc mv per channel = " << adc_mV_per_chan
			<< ", total electronic factor = " << electronic_correction_factor << "\n";
		streamlog_out( MESSAGE3 ) << "Processing " << input_pulses->getNumberOfElements() << " pulses.\n.";

	    // create the output collection with a string defining the TYPE (see lcio.h)
	    LCCollectionVec* output_pulses = new LCCollectionVec( LCIO::TRACKERDATA );
	    // set the transient flag as definded by the processor parameter
	    output_pulses->setTransient( (_output_is_transient != 0) );

	    // set the flags that cellID1 should be stored
	    lcio::LCFlagImpl trkFlag(0) ;
	    trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
	    output_pulses->setFlag( trkFlag.getFlag() );

	    for(int iRawData = 0; iRawData< input_pulses->getNumberOfElements(); iRawData++) {

			/* getElement return only a LCObject
			* you have to cast it the get the right type
			* please use dynamic_cast which will return 0 if the object can not be casted
			*/
			TrackerPulse* input_pulse = dynamic_cast<TrackerPulse*>( input_pulses->getElementAt( iRawData ) ) ;

			if (input_pulse) {

				/* create a new object of the TrackerData type
				* you have to use the IMPL, otherwise you can not set values
				*/
				TrackerPulseImpl* output_pulse = new TrackerPulseImpl();

				// transver the values

				//first the simple
				output_pulse->setCellID0( input_pulse->getCellID0() );
				output_pulse->setCellID1( input_pulse->getCellID1() );// normally 0 for TPC data
				output_pulse->setTime( input_pulse->getTime() );
				output_pulse->setQuality( input_pulse->getQuality() );

				// now convert from fadc values to primary electrons
				output_pulse->setCharge( input_pulse->getCharge() * electronic_correction_factor );

				streamlog_out( MESSAGE ) << "Pad #" << input_pulse->getCellID0() << ", amplified electrons: " << input_pulse->getCharge() << ", original electrons: " << output_pulse->getCharge() << ".\n";

				// add the converted pulse to the collection
				output_pulses->addElement( output_pulse );
			}
			else {

				throw( Exception("Input collection " + _input_tracker_pulses_collection_name + " does not contain TrackerPulses.") );
			}
	    }

	    
	    /* after convertin all raw data objects
	     * add the collection to the eventand asigning a name to it
	     * please note again: DO NOT DELETE the collection  this will be done by LCIO
	     */

		evt->addCollection( output_pulses, _output_tracker_pulses_collection_name );
	}
	catch(DataNotAvailableException e) {
	    // an error occured
	    streamlog_out( ERROR ) << e.what() << std::endl;
	}

	_number_of_events++;
}
    
    

void CountsToPrimaryElectronsProcessor::check( LCEvent * evt ) {
	// nothing to check here - could be used to fill checkplots in reconstruction processor
}
    
    
void CountsToPrimaryElectronsProcessor::end() {

}

}

#endif // USE_LCCD
