#include "PedestalSubtractorProcessor.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/LCFlagImpl.h>
#include <UTIL/LCFixedObject.h>

// tpcconditions
#include "GenericADCElectronicsParameters.h"

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

    PedestalSubtractorProcessor aPedestalSubtractorProcessor ;

    PedestalSubtractorProcessor::PedestalSubtractorProcessor() : Processor("PedestalSubtractorProcessor")
    {

        /* description of the processor which will be displayed in the
         * steering file autometicly generated by
         * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
         */
        _description = "PedestalSubtractorProcessor to correct the TrackerData by subtracting the pedestals provided via LCCD";

        /* register steering parameters: name, description, class-variable, default value
         * the type will by definded by the type of the default value
         * string, double, float and int are possible
         * plaese note to get a float you must cast in C++: (float) 1.3
         */
        registerInputCollection( LCIO::TRACKERDATA,
                                 "InputCollectionName" ,
                                 "Name of the input collection"  ,
                                 _inputColName ,
                                 std::string("TPCConvertedRawData") 
                               );

        registerOutputCollection( LCIO::TRACKERDATA,
                                  "OutputCollectionName" ,
                                  "Name of the output collection"  ,
                                  _outputColName ,
                                  std::string("TPCData") 
                                );

        registerProcessorParameter( "SetOutputTransient" ,
                                    "if not 0 the output collection is set transient"  ,
                                    _outputIsTransient,
                                    1
                                  );

        registerProcessorParameter( "PedestalCollectionName" ,
                                    "Name of the collection containing the pedestals",
                                    _pedestalColName,
                                    PedestalHandler::getDefaultColName() 
                                  );
        
        registerProcessorParameter( "ElectronicsParametersObjectName" ,
                                    "Name of the conditions object in which the parameters of the read out electronics is stored",
                                    _electrParaObjName,
                                    std::string("GenericADCElectronicsParameters")
                                  );
        
        registerProcessorParameter( "ElectronicsParametersCollectionName" ,
                                    "Name of the conditions collection in which the parameters of the read out electronics is stored",
                                    _electrParaColName,
                                    tpcconddata::GenericADCElectronicsParameters::getDefaultColName() 
                                  );
        
        registerOptionalParameter( "MaximumADCValue" ,
                                   "Maximum ADC Value of the read-out electronic",
                                   _maximumADCValue,
                                   float(-1.0)
                                  );

    }

    void PedestalSubtractorProcessor::init()
    {
        printParameters() ;
        _nRun = 0 ;
        _nEvt = 0 ;

        // the pedestal handler with and without LCCD
        _pedestalHandler = new PedestalHandler( _pedestalColName );
        
        // give warning message if optional maxADCValue parameter overrides conditions data
        if ( parameterSet("MaximumADCValue") ) 
        {
            m_out(WARNING)<<"Maximum ADC value ("<<_maximumADCValue<<") taken from steering file. This overrides parameter from conditions database!"<<std::endl;                
        }
        else
        {
            _adcElectronicsHandler = new ADCElectronicsHandler( _electrParaObjName, _electrParaColName );
        }
    }
    
    
    void PedestalSubtractorProcessor::processRunHeader( LCRunHeader* run)
    {
        run->parameters().setValue( _processorName + "_revision", "$Rev: 1822 $");
        for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ )
        {
            if ( ! i->second->isOptional() || i->second->valueSet() )
            {
                run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
            }
        }
        _nRun++ ;
    }

    void PedestalSubtractorProcessor::processEvent( LCEvent * evt )
    {
        // get the input collection from the event
        try
        {
            LCCollection* inputCol = evt->getCollection( _inputColName ) ;
            // create the output collection with a string defining the TYPE (see lcio.h)
            LCCollectionVec* outputCol = new LCCollectionVec( LCIO::TRACKERDATA );
            // set the transient flag as definded by the processor parameter
            outputCol->setTransient( (_outputIsTransient != 0) );
            // set the flags that cellID1 should be stored
            lcio::LCFlagImpl trkFlag(0) ;
            trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
            outputCol->setFlag( trkFlag.getFlag() );
            
            // get the number of elements in the input collection
            int nRawData = inputCol->getNumberOfElements();
            // loop over the number of elements of the input collection
            for(int iRawData=0; iRawData< nRawData ; iRawData++)
            {
                /* getElement return only a LCObject
                 * you have to cast it the get the right type
                 * please use dynamic_cast which will return 0 if the object can not be casted
                 */
                TrackerData* RawData =
                    dynamic_cast<TrackerData*>( inputCol->getElementAt( iRawData ) ) ;
                if ( RawData != 0 )
                {
                    //only poceed if the cast to TrackerData was successful

                    // create a new object of the TrackerData type
                    TrackerDataImpl* Data = new TrackerDataImpl();

                    // transver the values from the "RawData" to the "Data" object

                    // first the simple
                    Data->setCellID0( RawData->getCellID0() );
                    Data->setCellID1( RawData->getCellID1() );
                    Data->setTime( RawData->getTime() );

                    m_out(DEBUG)<<"Processing Spectrum at Pad:"<<RawData->getCellID0()<<"/Module:"<<RawData->getCellID1()<<std::endl;
                    // subtract the pedestals 
                    float PedestalVal = _pedestalHandler->getPedestalValue( RawData->getCellID0(),
                                                                            RawData->getCellID1() );
                    m_out(DEBUG)<<"Pedestal Value here is: "<<PedestalVal<<std::endl;
                    FloatVec RawDataADCVal = RawData->getChargeValues();
                    m_out(DEBUG)<<"Values (before/after) : ";
                    // read in the maximum possible ADC value, either from LCCD conditions if parameter not set
                    if ( !parameterSet("MaximumADCValue") )
                    {
                        _maximumADCValue=_adcElectronicsHandler->getMaxADCValue(RawData->getCellID1());
                    }
                    for ( FloatVec::iterator iSlice=RawDataADCVal.begin();
                          iSlice!=RawDataADCVal.end(); iSlice++ )
                    {
                        float outSlice=0.0;
                        if ( (*iSlice)==_maximumADCValue ) // test if slice value is maximum of read-out electronics == overflow
                        {
                            outSlice = (*iSlice); // if slice in overflow, do not substract pedestal to preserve the overflow information
                            m_out(DEBUG)<<" ! Slice not pedestal substracted: ";
                        }
                        else if ( (*iSlice) > _maximumADCValue ) 
                        {
                            m_out(ERROR)<<std::endl<<"Error: Value of time slice (="<<(*iSlice)<<") is bigger than maximum ADC value (="<<_maximumADCValue<<") !"<<std::endl;
                            throw StopProcessingException(this);
                        } else {
                            outSlice=(*iSlice)-PedestalVal; // else substrect pedestal value
                        }
                        // debug output
                        m_out(DEBUG)<<(*iSlice)<<"/"<<outSlice<<" ";
                        // fill the data vector with the pedestal substracted data
                        (Data->chargeValues()).push_back( outSlice );
                    }
                    m_out(DEBUG)<<std::endl<<std::endl;
                    /* fill the TrackerData in the collection
                     * please note: DO NOT DELETE the object! this will be done by LCIO
                     */
                    outputCol->addElement( Data );
                }
            } // end of loop over the number of elements of the input collection

            // save the processor parameters in the first collection stored
            if (isFirstCollection)
            {
                outputCol->parameters().setValue("InputCollectionName",_inputColName);
                // TO DO: make this sense?
                outputCol->parameters().setValue("SetOutputTransient",_outputIsTransient);
                outputCol->parameters().setValue("PedestalCollectionName",_pedestalColName);
                #ifndef USE_LCCD
                outputCol->parameters().setValue("PedestalFileName",_pedestalFileName);
                #endif
                isFirstCollection = false;
            }

            /* after converting all raw data objects
             * add the collection to the event and asigning a name to it
             * please note again: DO NOT DELETE the collection  this will be done by LCIO
             */
            evt->addCollection( outputCol, _outputColName );

        }
        catch(DataNotAvailableException)
        {
            // if there is not input collection give a warning to the user
            std::cerr<<"Warning of PedestalSubtractorProcessor::processEvent: input collection "
                     <<_inputColName<<" cannot be found in event "<<_nEvt<<std::endl;
        }
        _nEvt++ ;
    }

    void PedestalSubtractorProcessor::check( LCEvent * evt )
    {
        // nothing here currently - could be extended to fill checkplots
    }

    void PedestalSubtractorProcessor::end()
    {
        m_out(MESSAGE)<< "PedestalSubtractorProcessor::end()  " << name()
                      << " processed " << _nEvt << " events in " << _nRun << " runs "<< std::endl ;
    }

}
