#include "PulseFinderProcessor.h"

//C++
#include <iostream>
#include <algorithm>

//LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

#ifdef USE_GEAR
// STUFF needed for GEAR
#include <marlin/Global.h>
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>
#endif

using namespace lcio;
using namespace marlin;
using namespace std;

namespace marlintpc
{
  PulseFinderProcessor aPulseFinderProcessor;
  
  PulseFinderProcessor::PulseFinderProcessor() : Processor("PulseFinderProcessor")
  {
    _description = "PulseFinderProcessor: finds pulses out of the raw data and calculate the time and charge (in FADC) information";
    
    registerInputCollection(LCIO::TRACKERDATA,
			    "InputCollectionName",
			    "Name of the input TrackerData collection",
			    _inputColName,
			    string("TPCData"));
    
    registerOutputCollection(LCIO::TRACKERDATA,
			     "OutputTrackerDataCollectionName",
			     string("Name of the output TrackerData collection ")+
			     string("only the data related to a pulse will be stored"),
			     _outputDataColName,
			     string("TPCDataInPulses"));
	
    registerOutputCollection(LCIO::TRACKERPULSE,
			     "OutputTrackerPulseCollectionName",
			     "Name of the output TrackerPulse collection",
			     _outputPulseColName,
			     string("TPCPulses"));

    registerProcessorParameter("SetOutputTransient",
			       "if true the output collection is set transient",
			       _outputIsTransient,
			       true);
	
    registerProcessorParameter("ForceSpectrumSave",
			       "if true the ADC spectrum will be stored for all pulses",
			       _forceSpectrumSave,
			       true);
	
    registerProcessorParameter("PulseStartThreshold",
			       string("A pulse starts if the ADC value is the given value times the pedestal width."),
			       _pulseStartThreshold,
			       (float)7.0);

    registerProcessorParameter("PulseEndThreshold",
			       string("A pulse ends if the ADC value drops below this value times the pedestal width."),
			       _pulseEndThreshold,
			       (float)3.0);

    registerProcessorParameter("MinimumPulseHeight",
			       string("Declares the minimal height of the maximum entry of a pulse in ADC counts."),
			       _minimumPulseHeight,
			       (float)12.0);
    
    registerProcessorParameter("MinimumPulseLength",
			       string("Declares the minimal length of a pulse in time bins above the thresholds."),
			       _minimumPulseLength,
			       5);

    registerProcessorParameter("SaveNBinsBeforeStart",
			       "# bins saved before start threshold (except there is another pulse)",
			       _saveNBinsBeforeStart,
			       2);
	
    registerProcessorParameter("SaveNBinsAfterEnd",
			       "# bins saved after the end threshold (except there is another pulse)",
			       _saveNBinsAfterEnd,
			       8);

    registerProcessorParameter("MaximumADCValue",
			       "maximum number of the adc spectrum; to determine wether a pulse is maxed out",
			       _maximumADCValue,
			       (float)255.);

    registerProcessorParameter("NegativeElectronicsPolarity",
			       "Set to true, if the electronics register negative values; default is /false/.",
			       _negativePolarity,
			       false);
    
    registerProcessorParameter("DoMultiplePulseSplitting",
                               "If set to true, multiple pulses will be split by detecting a change in curvature of the adc values",
                               _doMultiplePulseSplitting,
                               true);
    
    registerProcessorParameter("ReadoutFrequency",
			       "Frequency of the readout in Hz",
			       _readoutFrequency,
			       12.5E6);


    registerOptionalParameter("UseChargeAveragedPulseTime",
			      "Switch the default way the pulse time is calculated from the derivative sign change to charge average.",
			      _useChargeAveragePulseTime,
			      false);
    
    registerOptionalParameter("NoiseOverrideValue",
			      "HACK! for now this sets the default value of the noise _AND_ turns off the pedestal collection usage!",
			      _noiseOverrideValue,
			      1);
    

    registerProcessorParameter("PedestalCollectionName",
			       "Name of the collection containing the pedestal information",
			       _pedestalColName,
			       PedestalHandler::getDefaultColName() );
	
    /* if MARLIN is compiled without LCCD a file is needed which contains the pedestal information
     */
#ifndef USE_LCCD
    if(!parameterSet("NoiseOverrideValue")) // HACK! change, when better solution available
      {
	registerProcessorParameter("PedestalCollectionFile",
				   "File containing the pedestals",
				   _pedestalFileName,
				   string("Pedestal.slcio"));
      }
#endif
  }
  
  
  
  void PulseFinderProcessor::init()
  {
    
    // usually a good idea to
    printParameters();
    
    _nRun = 0;
    _nEvt = 0;
    
    // the pedestal handler with and without LCCD
#ifdef USE_LCCD
    if(!parameterSet("NoiseOverrideValue")) // HACK! change, when better solution available
      {
	_pedestalHandler = new PedestalHandler( _pedestalColName );
      }
#else
    if(!parameterSet("NoiseOverrideValue")) // HACK! change, when better solution available
      {
	_pedestalHandler = new PedestalHandler( _pedestalColName, _pedestalFileName );
      }
#endif
  }



void PulseFinderProcessor::processRunHeader(LCRunHeader* run)
{
  run->parameters().setValue( _processorName + "_revision", "$Rev: 1879 $");
  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ )
    {
      if ( ! i->second->isOptional() || i->second->valueSet() )
	{
	  run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
	}
    }
  
    _nRun++;
  }



void PulseFinderProcessor::processEvent(LCEvent* evt)
  {
    // give a life sign if you are in debug mode
    _nEvt++;
    m_out(DEBUG) << "Debugging at event " <<_nEvt << endl;
    
    try
      {
	LCCollection* InputCollection = evt->getCollection(_inputColName);
	
	// create the output collections for pulses and zero suppressed data (ADC spectra)
	IMPL::LCCollectionVec* outputColPulse = new IMPL::LCCollectionVec(LCIO::TRACKERPULSE);
	IMPL::LCCollectionVec* outputColData  = new IMPL::LCCollectionVec(LCIO::TRACKERDATA);
	
	// set the transient flag as definded by the processor parameter
	outputColPulse->setTransient(_outputIsTransient);
	outputColData->setTransient(_outputIsTransient);
	
	// set the flags that cellID1 should be stored
	lcio::LCFlagImpl trkFlag(0);
	trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1);
	outputColPulse->setFlag(trkFlag.getFlag());
	outputColData->setFlag(trkFlag.getFlag());
	
	/* temporary data object, to be filled by /raw/ pulse finding 
  	   first pair object is the actual tracker data,
	   second pair object is the quality
	*/
	vector<pair<TrackerDataImpl*, int> > tempPulses;
	tempPulses.clear();

	unsigned int nData = InputCollection->getNumberOfElements();

	// first loop over all channels within the event to search for the /raw/ pulses
	for(unsigned int iData = 0; iData < nData; ++iData)
	  {
	    TrackerData* thisData = dynamic_cast<TrackerData*>( InputCollection->getElementAt(iData) );

	    if (thisData)    // cast is OK
	      {
		m_out(DEBUG) << "Looking at Pad:" << thisData->getCellID0() 
			     << " module:" << thisData->getCellID1() << endl;
		if(!parameterSet("NoiseOverrideValue")) // HACK! change, when better solution available
		  {
		    m_out(DEBUG) << "Pedestal Value for pad:" << thisData->getCellID0() 
				 << " module:" << thisData->getCellID1() << " = " 
				 << _pedestalHandler->getPedestalValue( thisData->getCellID0(), thisData->getCellID1() ) << endl;
		    m_out(DEBUG) << "Pedestal Width for pad:" << thisData->getCellID0() 
				 << " module:" << thisData->getCellID1() << " = " 
				 << _pedestalHandler->getPedestalWidth( thisData->getCellID0(), thisData->getCellID1() ) << endl;
		  }
		// special meaning of noise: the triple value of the pedestal width! MAGIC NUMBER!
		float noiseValue;
		if(!parameterSet("NoiseOverrideValue")) // HACK! change, when better solution available
		  {
		    noiseValue = _pedestalHandler->getPedestalWidth(thisData->getCellID0(), thisData->getCellID1()) * 3.;
		  }
		else
		  {
		    noiseValue = _noiseOverrideValue;
		  }

		// if pedestal width is undefined, its default value is -1
		//  then use pedestal width = 1
		if (noiseValue < 0.)
		  noiseValue = 1.;
		
		// this is the master call: search in the spectrum for pulses and
		// store them into the collection vector that is passed to the call
		findPulses(thisData, &tempPulses, noiseValue);
	      }
	  }
	
	// use the found pulses to calculate time and charge information and store the data
	for(vector<pair<TrackerDataImpl*, int> >::const_iterator pulseCandidate = tempPulses.begin();
	    pulseCandidate < tempPulses.end(); ++pulseCandidate)
	  {
	    TrackerDataImpl* trackerData = pulseCandidate->first;
	    int quality                  = pulseCandidate->second;
	    
	    if( streamlog::out.write< streamlog::DEBUG >() )
	      {
		m_out(DEBUG) << "++{ Found a pulse on [Pad/Module]: [" << trackerData->getCellID0() 
			     << "/" << trackerData->getCellID1() << "], at the time " << trackerData->getTime() << " of length " 
			     << (trackerData->getChargeValues()).size() << " and with quality flag \"" << quality << "\"" << endl;

		m_out(DEBUG4) << "The pulse spectrum is : ";
		vector<float> theADCSpectrum = trackerData->getChargeValues();
		for(vector<float>::const_iterator adcVal = theADCSpectrum.begin(); adcVal < theADCSpectrum.end(); ++adcVal)
		  {
		    m_out(DEBUG4) << *adcVal << " ";
		  }
		m_out(DEBUG4) << "}+++" << endl;
	      }

	    float noiseValue;
	    if(!parameterSet("NoiseOverrideValue")) // HACK! change, when better solution available
	      noiseValue =  _pedestalHandler->getPedestalWidth(trackerData->getCellID0(), trackerData->getCellID1()) * 3.;
	    else 
	      noiseValue = _noiseOverrideValue;
	    
	    // if pedestal width is not defined its default value is -1
	    // then set pedestal width equal to 1
	    if (noiseValue < 0.)
	      noiseValue = 1.;
	    
	    //save Pulse
	    TrackerPulseImpl* newPulse = new TrackerPulseImpl();
	    newPulse->setCellID0(trackerData->getCellID0());
	    newPulse->setCellID1(trackerData->getCellID1());
	    newPulse->setTime(calcPulseTime(trackerData, noiseValue));
	    newPulse->setCharge(calcPulseCharge(trackerData));
	    newPulse->setQuality(quality);
	    
	    // store the adc spectrum connected to the pulse
	    // this is implicitly done by incorporating /_forceSpectrumSave/ into /quality/
	    if ( quality != 0 || _forceSpectrumSave)
	      {
		// make a copy of the TrackerData
		TrackerDataImpl* newData = new TrackerDataImpl();
		newData->setCellID0(trackerData->getCellID0());
		newData->setCellID1(trackerData->getCellID1());
		newData->setTime(trackerData->getTime());
		vector<float> thisDataChargeVal = trackerData->getChargeValues();
		for (vector<float>::const_iterator iSlice = thisDataChargeVal.begin(); iSlice != thisDataChargeVal.end(); ++iSlice)
		  (newData->chargeValues()).push_back( (*iSlice) );
		
		// add the spectrum to the <tracker data> collection
		outputColData->addElement(newData);
		// and add the tracker data to the <tracker pulse> collection
		newPulse->setTrackerData(newData);
	      }
	    
	    outputColPulse->addElement(newPulse);
	  }
	
	evt->addCollection(outputColPulse, _outputPulseColName);
	evt->addCollection(outputColData, _outputDataColName);
      }
    catch(DataNotAvailableException)
      {
	// if there is not input collection give a warning to the user
	cerr << "Warning of PulseFinderProcessor::processEvent: input collection "
	     << _inputColName << " cannot be found in event " << _nEvt << endl;
      }
  }



void PulseFinderProcessor::check(LCEvent* evt)
  {
  }



void PulseFinderProcessor::end()
  {
  }



  void PulseFinderProcessor::findPulses(TrackerData* thisData, vector<pair<TrackerDataImpl*,int> >* tempPulses, float noiseValue)
  {
    /** the basic idea to find pulses is split into three steps:
       (1) simply look for candidate pulses: above values between start and end threshold 
          due to the high frequency of this already check here for a minimum length
       (2) search the candidate pulses for possible multiple pulse and split, if necessary
       (3) finally check the pulse boundaries and assemble the pulse out of the bins before and after the pulse, if possible
    */

    float startTime  = thisData->getTime(); 

    vector<float> ADCSpectrum = thisData->getChargeValues();
    m_out(DEBUG) << " The current pad has an adc vector of length " << ADCSpectrum.size() << endl;

    /* this is a help construct for the data (including a typedef)
       simply store the start and end position of the pulse
    */
    typedef std::vector<float>::iterator timeIter;
    vector<pair<timeIter, timeIter> > theCandidatePulses;
    
    /* do the first task: 
       simply look for the values above the start threshold and below end threshold
    */
    brutePulseSearch(ADCSpectrum, &theCandidatePulses);

    /* do the check for multiple pulses,
       technically: divide the start- and end-iterator into a new pair
       TODO: in addition to pulse splitting adjust the charge at the boundaries
    */
    if(_doMultiplePulseSplitting)
      splitMultiplePulses(&theCandidatePulses, noiseValue);
    
    /* as last part determine the possible boundaries of the pulse;
       - check wether the storing of the time bins before and after the end is possible
       - set the flags accordingly
    */

    // this is the supposed minimum time between two pulses
    unsigned int nominalSeparation =  _saveNBinsBeforeStart + _saveNBinsAfterEnd;
    
    /* now prepare the pulse candidates for storage:
       - assign quality flags (according to agreement: document in preparation)
       - check boundaries at the beginning and the end
       - return the data construct for storage
    */
    m_out(DEBUG) << "There are " << theCandidatePulses.size() << " PulseCandidates dedicated to storage" << endl;
    for(vector<pair<timeIter, timeIter> >::const_iterator theTimes = theCandidatePulses.begin();
	theTimes < theCandidatePulses.end(); theTimes++)
      {
	m_out(DEBUG) << "Checking pulse " << theTimes - theCandidatePulses.begin() << " of " << theCandidatePulses.size() << endl;
	int quality = 0;
	timeIter endTimePulseBefore;
	timeIter startTimePulseAfter;

	// two special cases: the first and the last pulse candidate
	if(theTimes == theCandidatePulses.begin())
	  endTimePulseBefore = ADCSpectrum.begin();//(theCandidatePulses.begin())->second;
	else
	  endTimePulseBefore  = (theTimes - 1)->second;

	if(theTimes == (theCandidatePulses.end() - 1) )
	  startTimePulseAfter = ADCSpectrum.end();//(theCandidatePulses.end() - 1)->first;
	else	
	  startTimePulseAfter = (theTimes + 1)->first;
	
	// get the distance in time bins to the neighboring pulse candidates
	unsigned int deltaTBefore = theTimes->first - endTimePulseBefore;
	unsigned int deltaTAfter  = startTimePulseAfter - theTimes->second;

	if( (deltaTBefore < nominalSeparation) || (deltaTAfter < nominalSeparation) )
	  {
	    m_out(DEBUG) << "The distance in time to its neighbor pulses indicates a multiple pulse." << endl;
	    quality |= (1 << 1); // multiplepulsecandidate, bit 1
	  }

	int pulseLength  = theTimes->second - theTimes->first;
	if( pulseLength > 3*_minimumPulseLength) // FIXME: is that a good condition?
	  {
	    m_out(DEBUG) << "The pulse is exceptionally long, flagged for anomalous shape." << endl;
	    quality |= ( 1 << 2); // anomalousshape, bit 2
	  }
	for(timeIter checkValues = theTimes->first; checkValues < theTimes->second; ++checkValues)
	  {
	    if( !(quality & ( 1 << 4)) && *checkValues == _maximumADCValue )
	      {
		quality |= ( 1 << 4); // overflow, bit 4
		m_out(WARNING) << "The pulse candidate has reached its overflow value, flagging overflow." << endl;
	      }
	    if(*checkValues > _maximumADCValue )
	      m_out(ERROR) << "The adc value EXCEEDED the set maximum adc value; better check your math or your detector." << endl;
	  }
	
	unsigned int binsToStoreBefore = _saveNBinsBeforeStart;
	unsigned int binsToStoreAfter  = _saveNBinsAfterEnd;

	if(deltaTBefore < static_cast<unsigned int>(_saveNBinsBeforeStart) )
	  binsToStoreBefore = deltaTBefore;
	if(deltaTAfter < static_cast<unsigned int>(_saveNBinsAfterEnd) )
	  binsToStoreAfter = deltaTAfter;

	// calculate the start time of the pulse to be stored:
	float currentTime = startTime + (theTimes->first - ADCSpectrum.begin()) - binsToStoreBefore; 

	TrackerDataImpl* toStoreTrackerData;
	toStoreTrackerData = new TrackerDataImpl();
	toStoreTrackerData->setCellID0(thisData->getCellID0());
	toStoreTrackerData->setCellID1(thisData->getCellID1());
	toStoreTrackerData->setTime(currentTime);
	
	for(timeIter getValues = theTimes->first - binsToStoreBefore; getValues < theTimes->second + binsToStoreAfter; ++getValues)
	  (toStoreTrackerData->chargeValues()).push_back( *getValues );

	pair<TrackerDataImpl*, int> tempPair;
	tempPair.first = toStoreTrackerData;
	tempPair.second = quality;
	tempPulses->push_back(tempPair);
      }

    return;
  }



  void PulseFinderProcessor::brutePulseSearch(vector<float>& ADCSpectrum, vector<pair<vector<float>::iterator, vector<float>::iterator> >* theCandidatePulses)
  {
    typedef vector<float>::iterator timeIter;
    timeIter candStart, candEnd;
    
    bool insidePulse    = false;
    bool foundCandPulse = false;
    
    /* first search the spectrum for values, distinguish:
     (A) OUTSIDE a pulse and looking for a value above the start threshold
     (B) INSIDE a pulse and looking for a value below the end threshold
    */

    // also look if the pulse candidate meets the minimum requirements of length and height
    float maxFoundPulseHeight = 0.;
    for(vector<float>::iterator adcValue = ADCSpectrum.begin(); adcValue < ADCSpectrum.end(); ++adcValue)
      {
	if(!insidePulse && (*adcValue > _pulseStartThreshold))
	  {
	    m_out(DEBUG) << "Found an adc value above the threshold." << endl;
	
	    insidePulse = true;
	    candStart   = adcValue;
	  }
	if(insidePulse)
	  {
	    if(*adcValue > maxFoundPulseHeight)
              {
		maxFoundPulseHeight = *adcValue;
		m_out(DEBUG) << "The current maximal adc value in the pulse candidate is " << maxFoundPulseHeight << endl;
	      }

	    if(*adcValue < _pulseEndThreshold)
	      {
		insidePulse    = false;
		candEnd        = adcValue;
		foundCandPulse = true;
	      }
	    if( (adcValue + 1) == ADCSpectrum.end() )	
	      {
		candEnd        = ADCSpectrum.end();
		foundCandPulse = true;
	      }
	  }
	
	// check wether a pulse candidate has been found
	if(foundCandPulse)
	  {
	    m_out(DEBUG) << "A possible pulse candidate is checked for its height and length: ";

	    // pulses are plenty, do a simple length and height check before any further computation
	    int pulseLength = candEnd - candStart;
	    if(! (pulseLength >= _minimumPulseLength) || (maxFoundPulseHeight < _minimumPulseHeight )) // the minimum requirements
	      {
		m_out(DEBUG) << "It is discarded." << endl;
		foundCandPulse = false;
		continue;
	      }

	    // now store the candidate pulse
	    pair<timeIter, timeIter> tmpPair;
	    tmpPair.first  = candStart;
	    tmpPair.second = candEnd;
	    theCandidatePulses->push_back(tmpPair);
	    
	    m_out(DEBUG) << "It's stored for closer inspection." <<endl;
	    foundCandPulse = false;
	  }
      } // end of the first task

    return;
  }



  void PulseFinderProcessor::splitMultiplePulses(vector<pair<vector<float>::iterator, vector<float>::iterator> >* theCandidatePulses, float noiseValue)
  {
    typedef vector<float>::iterator timeIter;
    vector<pair<timeIter, timeIter> > temporaryPulseStorage;
    
    /* now check the candidates for multiple pulses
       first find the positions where the candidate needs to be split
                  for this: construct the derivative vector for the pulse in question
    */
    
    for(vector<pair<timeIter, timeIter> >::iterator aPulse = theCandidatePulses->begin();aPulse < theCandidatePulses->end(); ++aPulse)
      {
	vector<float> derivativeVec;
	vector<unsigned int> splitPositions;
	
	/* for each candidate in the lot:
	   loop from start to end of pulse and check wether
	   a change in the derivative from positive to negative occurs 
	*/
	for( timeIter aBin = aPulse->first; aBin < (aPulse->second - 1); ++aBin)
	  {
	    double derivative = *(aBin + 1) - *aBin;
	    derivativeVec.push_back(derivative);
	  }
	
	// now take a closer look at the candidate
	for(vector<float>::const_iterator dVal = derivativeVec.begin(); dVal < (derivativeVec.end() - 1); ++dVal)
	  {
	    // check the sign change and if the absolute change is above the noise threshold
	    // store the position of the /end/ of the split pulse + 1 (this is the split value +2)
	    if( (*dVal < 0) && (*(dVal+1) > 0) && (fabs(*dVal - *(dVal+1)) > noiseValue) )
	      {
		splitPositions.push_back(dVal-derivativeVec.begin() + 2);
		/* still under construction:
		   for a more exact charge evaluation the charge needs to be spread between the split pulses
		   
		   double factor_change = *(dVal+1) / (*dVal - *(dVal + 1) );
		*/
	      }
	  }

	if(splitPositions.empty())
	  temporaryPulseStorage.push_back(*aPulse);
	else
	  {
	    timeIter currentStart = aPulse->first;
	    pair<timeIter, timeIter> tmpPair;
	    
	    for(vector<unsigned int>::const_iterator split = splitPositions.begin(); split < splitPositions.end(); ++split)
	      {
		tmpPair.first  = currentStart;
		tmpPair.second = aPulse->first + *split;
		currentStart   = aPulse->first + *split;
		
		temporaryPulseStorage.push_back(tmpPair);
	      }
	    tmpPair.first  = currentStart;
	    tmpPair.second = aPulse->second;

	    temporaryPulseStorage.push_back(tmpPair);
	  }
      }// end loop over pulses

    theCandidatePulses->clear();
   
    // create the new Pulses
    for(vector<pair<timeIter,timeIter> >::iterator aPulse = temporaryPulseStorage.begin();aPulse < temporaryPulseStorage.end(); ++aPulse)
      {
	int pulseLength = aPulse->second - aPulse->first;
	float maxFoundPulseHeight = 0.;
	for(int testValue = 0; testValue < pulseLength; ++testValue)
	  {
	    float adcValue = *(aPulse->first + testValue);
	    if( adcValue > maxFoundPulseHeight)
	      maxFoundPulseHeight = adcValue;
	  }
	if( (pulseLength >= _minimumPulseLength) 
	    && (maxFoundPulseHeight > _minimumPulseHeight ) )
	  theCandidatePulses->push_back(*aPulse);

      }
  }
  
  
  
  float PulseFinderProcessor::calcPulseTime(TrackerData* thisPulse, float noiseValue)
  {
    vector<float> chargeVec = thisPulse->getChargeValues();
    float _timeBin  = 1e+9/_readoutFrequency;
    
    if(parameterSet("UseChargeAveragedPulseTime") &&  _useChargeAveragePulseTime)
      {
        float chargeSum = 0;
        float timeSum   = 0;
	
	// sum up the charge and calculate the charge weighted time sum
        for(vector<float>::const_iterator charge = chargeVec.begin(); charge < chargeVec.end(); ++charge)
	  {
            chargeSum += (*charge);
	    timeSum   += (*charge) * _timeBin * ( (charge - chargeVec.begin()) + 1 );
	  }

        timeSum  = timeSum/chargeSum;
        timeSum += thisPulse->getTime();
        return timeSum;
      }
    
     float PulseTime    = 0.0;
     float Derivative   = 0.0;
     float Norm         = 0.0;
     float SliceTime    = thisPulse->getTime();
     
     if (chargeVec.size()>1)
       {
	 for(vector<float>::const_iterator timeBin = chargeVec.begin() + 1; timeBin < chargeVec.end(); ++timeBin)
	   { 
	     Derivative  = (*timeBin)-(*(timeBin-1));

	     // make derivative sign insensitive
	     if(_negativePolarity)
	       Derivative *= (-1.); 
	     
	     SliceTime += 1.;

	     // break out of calculation if decrease is larger than the noise
	     if (Derivative < (-noiseValue))
	       break;
	     
	     Norm      += Derivative;
	     PulseTime += Derivative * SliceTime;  
	   }
	 if(Norm != Norm) // check for nan; exception is not readily available
	   {
	     m_out(WARNING) << "Pulse time calculation failed, default to beginning of pulse" << endl;
	     PulseTime = SliceTime;
	   }
	 else if(Norm > 0.)
	   PulseTime /= Norm;
       }
     else
       {
	 PulseTime = SliceTime;
       }

     return PulseTime*_timeBin; //conversion in ns
  }
  
  
  
  float PulseFinderProcessor::calcPulseCharge(TrackerData* thisPulse)
  {
    float         PulseCharge  = 0.0;
    vector<float> ChargeValues = thisPulse->getChargeValues();

    for (vector<float>::const_iterator adcVal = ChargeValues.begin(); adcVal < ChargeValues.end(); ++adcVal)
      PulseCharge += (*adcVal);
      
    /// flip sign if polarity is negative
    if (_negativePolarity) 
      PulseCharge *= -1;
    
    return PulseCharge;
  }
}
