
// STL
#include <vector>
#include <map>
#include <algorithm>
#include <iomanip>
#include <cassert>

// LCIO
#include "EVENT/LCIO.h"
#include "EVENT/LCCollection.h"
#include "EVENT/TrackerHit.h"
#include "IMPL/LCCollectionVec.h"
#include "IMPL/TrackImpl.h"
#include <IMPL/LCFlagImpl.h>

// GEAR
#include "gear/GEAR.h"
#include "gear/TPCParameters.h"
#include "gear/PadRowLayout2D.h"
#include "gearxml/GearXML.h"

// CLHEP
#include "CLHEP/Matrix/Matrix.h"
#include "CLHEP/Vector/LorentzVector.h"
#include "CLHEP/Vector/TwoVector.h"

// Marlin
#include "marlin/Global.h"
#include "SimpleTrackSeederProcessor.h"

using namespace marlin;
using namespace lcio;

namespace marlintpc
{

SimpleTrackSeederProcessor aSimpleTrackSeederProcessor;

SimpleTrackSeederProcessor::SimpleTrackSeederProcessor() : Processor("SimpleTrackSeederProcessor")
{

	_description = "Calculates seed track parameters from  in the TrackerHits in the track candidates collection!";

	registerInputCollection(    LCIO::TRACK,
				    "InputTrackCandidates",
				    "The the name of the input collection of track candidates (default: TPCTrackCandidates)",
				    _input_track_candidates_collection_name,
				    std::string("TPCTrackCandidates"));

	registerOutputCollection(    LCIO::TRACK,
				    "OutputSeedTracks",
				    "The name of the collection the seed tracks be stored under (default: TPCSeedTracks)",
				    _output_seed_tracks_collection_name,
				    std::string("TPCSeedTracks"));
	registerProcessorParameter( "SetOutputTransient" , 
				    "if not 0 the output hits collection is set transient (default: 0)"  ,
				    _outputIsTransient,
				    int(0));
  
}


SimpleTrackSeederProcessor::~SimpleTrackSeederProcessor() {

}


void SimpleTrackSeederProcessor::init() {

  // FIXME!!! This if statement will become obsolete / should be removed 
  //          once the new logging mechanism is implemented in printParameters()
    if( streamlog::out.write< streamlog::MESSAGE3 >() )
	printParameters();

  // get the pad layout from gear
//  try
//  {
//    _padLayoutType = Global::GEAR->getTPCParameters().getPadLayout().getPadLayoutType();
//    switch ( _padLayoutType ) {
//    case PadRowLayout2D::CARTESIAN:
//      _padLayout = &dynamic_cast<const RectangularPadRowLayout&>( Global::GEAR->getTPCParameters().getPadLayout() );
//      break;
//    case PadRowLayout2D::POLAR:
//      _padLayout = &dynamic_cast<const FixedPadSizeDiskLayout&>( Global::GEAR->getTPCParameters().getPadLayout() );
//      break;
//    default:
//	throw gear::Exception( "Unknown pad layout type" );
//    };
////    _maxDriftDistance = Global::GEAR->getTPCParameters().getMaxDriftLength ();
//  }
//  catch(std::exception& e)
//  {
//      streamlog_out(ERROR)<< e.what() << std::endl;
//    throw e;
//  }
//
}


void SimpleTrackSeederProcessor::processRunHeader(LCRunHeader *run) {

    run->parameters().setValue( _processorName + "_revision", "$Rev: 781 $");

    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
        if ( ! i->second->isOptional() || i->second->valueSet() ) {
            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }

}

void SimpleTrackSeederProcessor::processEvent(LCEvent *evt) {

    streamlog_out(DEBUG4) << "SimpleTrackSeederProcessor"<<std::endl;
    
    IMPL::LCCollectionVec* trackCandidates=0;
    try {
	// the tracker hit collection
	trackCandidates = dynamic_cast<IMPL::LCCollectionVec*>(evt->getCollection(_input_track_candidates_collection_name));
    }
    catch (DataNotAvailableException &) {
	streamlog_out(MESSAGE3) << "No seed tracks collection \""<<_input_track_candidates_collection_name
				<<"\" in event "<< evt->getEventNumber()<<std::endl;
	return;
    }
    catch (std::bad_cast)
    {
	streamlog_out(WARNING1) << "Could not convert collection \""<<_input_track_candidates_collection_name
				<<"\" to LCCollectionVec, skipping event "<< evt->getEventNumber()<<std::endl;
	return;
    }
    
//		const double innermost_electronics_z = tpc_parameters.getMaxDriftLength();
    


    // a collection to hold the new track
    IMPL::LCCollectionVec* seedTrackCollection = new IMPL::LCCollectionVec(EVENT::LCIO::TRACK);

    LCFlagImpl trkFlag(0) ;
    trkFlag.setBit( LCIO::TRBIT_HITS ) ;
    seedTrackCollection->setFlag( trkFlag.getFlag() );

    //loop all tracks
    for (IMPL::LCCollectionVec::iterator trackIter=trackCandidates->begin(); 
	 trackIter < trackCandidates->end() ; trackIter++)
    {
	EVENT::Track *track = 0;
	try{
	    track = dynamic_cast<EVENT::Track *>(*trackIter);
	}
	catch(std::bad_cast &)
	{
	    streamlog_out(ERROR3) << "bad_cast: Track candidates collection"
				  << _input_track_candidates_collection_name
				  << " does not contain EVENT::Tracks!"<< std::endl;
            throw DataNotAvailableException("No track candidates found.");
	}
	
	// The TrackerHitVec is a vector of pointers to TrackerHits:  std::vector< TrackerHit * >
	// trackerHits is a copy (!) of the TrackerHitVec in the track, so we can resort the pointers
	// using the helpe classes defined in the header file
	EVENT::TrackerHitVec trackerHits = track->getTrackerHits();

	// check whether there are at least 4 hits stored in the track, otherwise continue with next track
	if (trackerHits.size() < 4)
	{
	    streamlog_out(WARNING2) << "Less than 4 hits in track candidates. "
				    << "Check your settings of the track finder!"<< std::endl;
	    continue;
	}

	// the actual track estimate
	IMPL::TrackImpl* seedTrack = new IMPL::TrackImpl();

	// fill all hits into the seed track:
	for (EVENT::TrackerHitVec::iterator hit = trackerHits.begin();
	     hit < trackerHits.end(); hit++)
	{
	    seedTrack->addHit(*hit);
	}

	// set the origin as the reference point
	const CLHEP::Hep3Vector reference_point(0., 0., 0.);
	{
	    float refPoint[3];
	    refPoint[0] = reference_point.x();
	    refPoint[1] = reference_point.y();
	    refPoint[2] = reference_point.z();
	    seedTrack->setReferencePoint(refPoint);
	}

	// ok, up to now there only was getting the data from lcio
	// now the real work starts, calculations of the parameters
	
	// step one: calculate Omega, phi and d_0
	// it is Jasons code ;-)

	// sort the hits in y 
	std::sort( trackerHits.begin(), trackerHits.end(), sort_by_y() );

	if( streamlog::out.write< streamlog::DEBUG1 >() )
	{
	    streamlog::out() << "\tSorted the hits: x y z"<< std::endl;
	    for(EVENT::TrackerHitVec::iterator hit = trackerHits.begin(); hit < trackerHits.end(); hit++)
	    {
		streamlog::out() << (*hit)->getPosition()[0] <<"\t" 
				 << (*hit)->getPosition()[0] <<"\t"
				 << (*hit)->getPosition()[0] << std::endl;
	    }
	}

	// estimate Omega based on three points mapped into x,y
	const double * hit1Pos= trackerHits.front()->getPosition();
	CLHEP::Hep3Vector circle_point_one( hit1Pos[0],  hit1Pos[1], 0. );
	const double * hit2Pos= trackerHits[trackerHits.size() / 2]->getPosition();
	CLHEP::Hep3Vector circle_point_two(  hit2Pos[0],  hit2Pos[1], 0.  );
	const double * hit3Pos=  trackerHits.back()->getPosition();
	CLHEP::Hep3Vector circle_point_three(  hit3Pos[0],  hit3Pos[1], 0. );

	/*CLHEP::Hep3Vector circle_point_one( -1, 0, 0);
	  CLHEP::Hep3Vector circle_point_two(  0,  1, 0);
	  CLHEP::Hep3Vector circle_point_three( -1, 2, 0);*/
	
	streamlog_out(DEBUG3) << "calculating Omega based on three points: {"
				<< circle_point_one << "mm, " 
				<< circle_point_two << "mm, " 
				<< circle_point_three << "mm } "
				<< std::endl;
	
	CLHEP::Hep3Vector p1_minus_p2 = circle_point_one - circle_point_two;
	CLHEP::Hep3Vector p1_minus_p3 = circle_point_one - circle_point_three;
	CLHEP::Hep3Vector p2_minus_p1 = circle_point_two - circle_point_one;
	CLHEP::Hep3Vector p2_minus_p3 = circle_point_two - circle_point_three;
	CLHEP::Hep3Vector p3_minus_p1 = circle_point_three - circle_point_one;
	CLHEP::Hep3Vector p3_minus_p2 = circle_point_three - circle_point_two;
	
	CLHEP::Hep3Vector denominator = p1_minus_p2.cross(p2_minus_p3);
	
	const double param_one = p1_minus_p2.mag() * p2_minus_p3.mag() * p3_minus_p1.mag();
	const double param_two = 2 * denominator.mag();
	
	double estimated_omega=0;
	double estimated_phi=0;
	double estimated_d0=0;
	CLHEP::Hep3Vector helix_centre;
	CLHEP::Hep3Vector point_of_closest_approach;
	if(param_two < 1E-08) // numerator of curvature is very small, assume straight line
	{
	    streamlog_out(DEBUG2) << "Curvature is very small, assuming straight line" << std::endl;
	    
	    // estimate Omega
	    estimated_omega = 0.;
	    
	    // estimate phi
	    estimated_phi = atan2(circle_point_two[1] - circle_point_one[1], circle_point_two[0] - circle_point_one[0]);
	    
	    // estimate d0
	    CLHEP::Hep3Vector n_pca( -1 * sin(estimated_phi), cos(estimated_phi) );
	    CLHEP::Hep3Vector r( circle_point_one[0] - reference_point[0], circle_point_one[1] - reference_point[1] );
	    
	    estimated_d0 = n_pca.dot(r);

	    point_of_closest_approach = n_pca * estimated_d0; // this will be needed later to calculate tan(lambda)
	    streamlog_out(DEBUG3) << "pca is " << point_of_closest_approach  << std::endl;
/* calculation of tan(lambda) can only be done after resorting in z,
   it might be a curler where the current sorting in y is not a function in z
			// estimate tan(lambda)
			double d_s = 0;
			double d_z = 0;

			for(std::vector<CLHEP::Hep3Vector>::const_iterator hit = hit_locations.begin()+1; hit != hit_locations.end(); hit++) {

				d_s += (*hit - *(hit-1)).mag();
				d_z += (*hit).z() - (*(hit-1)).z();
				#ifdef BUILD_DEBUG
				std::cout << "\t\td_s: " << d_s << "mm, d_z: " << d_z << "mm." << std::endl;
				#endif
			}

			const double estimated_tan_lambda = d_z / d_s;
*/

//			seed_track->setTanLambda( estimated_tan_lambda );
	}
	else {

	    // calculate the 1 / radius
	    const double inverse_radius = param_two / param_one;

	    streamlog_out(DEBUG2)<< "estimating inverse radius as " 
				 << inverse_radius << "mm^-1 (radius = " << 1 / inverse_radius << ")"<< std::endl;

	    const double alpha = ( p2_minus_p3.mag2() * p1_minus_p2.dot( p1_minus_p3 ) ) / ( 2 * denominator.mag2() );
	    const double beta = ( p1_minus_p3.mag2() * p2_minus_p1.dot( p2_minus_p3 ) ) / ( 2 * denominator.mag2() );
	    const double gamma = ( p1_minus_p2.mag2() * p3_minus_p1.dot( p3_minus_p2 ) ) / ( 2 * denominator.mag2() );

	    // calculate the centre (in x,y - relative to global (0,0)
	    helix_centre = alpha*circle_point_one + beta*circle_point_two + gamma*circle_point_three;

	    streamlog_out(DEBUG2) << "estimating helix centre at " << helix_centre << "mm." <<std::endl;

	    // Helix centre, reference point and pca are alway on one straigt line (definition of pca).
	    // The vector along phi has to be perpendicular to the centre - pca vector,
	    // and thus to the  centre - ref vector. 
	    // Definition: The motion along the helix/circle is always clockwise, resulting in a positive Omega.
	    // The parametrisation leaves this freedom since it has three signed variables 
	    // Omega, d_0, tan(lambda) and phi in [0..2pi], which gives an additional freedom (i.e. sign)
	    // in the direction. These are four signs ^= 2^4 = 16 dof for 8 different cases. We realy cannot tell which
	    // way the particle travelled as we don't know its charge, so we can choose a direction.
	    // FIXME: One might want to implement something like "choose the direction of phi that points with small s
	    // are closer to the reference point".

	    const bool track_moving_clockwise = true;

/*	    // estimate the rotational direction !!FIXME!! the best way to do this would be comparing fitting two sine waves to the locations based on the estimated radius of curvature. Instead, it is assumed the radius of curvature is large enough that it does not overlap itself inside the TPC. It also doesn't work for tracks travelling in the opposite direction (yet)
	    
	    double d_phi = 0;
	    double d_z = 0;

	    for(std::vector<CLHEP::Hep3Vector>::const_iterator hit = hit_locations.begin()+1; hit != hit_locations.end(); hit++) {
		
		d_phi += ( *hit - helix_centre ).phi() - ( *(hit-1) - helix_centre ).phi();
				d_z += (*hit).z() - (*(hit-1)).z();
				
				streamlog_out(DEBUG1) << "\From: " << (*hit).z() << " to " << (*(hit-1)).z() << " mm." << std::endl;
				
				streamlog_out(DEBUG1) << "d_phi: " << d_phi << "rads, d_z: " << d_z << "mm." << std::endl;
	    }
	    const bool track_moving_clockwise = (d_phi > 0)? false: true;
	    
	    streamlog_out(DEBUG2)  << "estimating the track is moving "
					       << (track_moving_clockwise?"clockwise": "anti-clockwise")
				   << std::endl;
*/	    
	    estimated_omega = inverse_radius * (track_moving_clockwise? 1: -1);

	    // keep the unit vector in 2D
//no, should be 0 and stay 0, every z is 0	    helix_centre.setZ( reference_point.z() );
	    streamlog_out(DEBUG0) << "helix_centre.z() should be 0. and is " << helix_centre.z() <<std::endl;
	    point_of_closest_approach = helix_centre - (helix_centre - reference_point).unit() / inverse_radius;
	    
	    streamlog_out(DEBUG2) << "calculated the point of closest approach at " << point_of_closest_approach << std::endl;
	    
 	    // estimate phi (the angle in r-phi between the slope of the tangent going through the p.c.a. and the vector(1,0,0))
	    const CLHEP::Hep3Vector pca_relative_helix_centre = point_of_closest_approach - helix_centre;
	    streamlog_out(DEBUG2) << "pca relative to centre phi: " << pca_relative_helix_centre.phi() << " rads."<<std::endl;
			
	    estimated_phi = fmod( pca_relative_helix_centre.phi() + (M_PI / 2) * (track_moving_clockwise? -1: 1) + M_PI * 2, M_PI * 2 );
			
	    // estimate d0
	    const CLHEP::Hep3Vector d = (point_of_closest_approach - reference_point);
	    const CLHEP::Hep3Vector n_pca( -1 * sin( estimated_phi ) , cos( estimated_phi ) );
	    estimated_d0 = n_pca.dot( d );
	    
/* This does not work for curlers, do it later
	    // estimate tan lambda
	    const double estimated_tan_lambda = d_z * inverse_radius / d_phi;
*/	    
	} 


	// step two: estimate tan-lambda

	// start by calculating s/z pairs (s in [-pi*r .. pi*r])
	// store them in a vector

	// Also determine whether track is in positive or negative half TPC.
	// In principle all hits on one track should be either positive or negative,
	// but due to a wrong drift velocity there might be hits with the wrong sign.
	// Just count how many positive and negative there are and take sign of majority.
	int signOfZCounter=0;

	std::vector< std::pair<double,double> > szPairs;
        // unit vector along phi, needed in case of straight line
	const CLHEP::Hep3Vector n_phi(cos(estimated_phi),sin(estimated_phi)); 
	// unity vector at pca perpendicular to track, pointing away from centre, needed in case of real helix
	const CLHEP::Hep3Vector n_pca( -1.*sin(estimated_phi), cos(estimated_phi) );
	
	for (EVENT::TrackerHitVec::iterator hitIter = trackerHits.begin(); 
	     hitIter<trackerHits.end() ; hitIter++)
	{
	    double const *hitPosition=(*hitIter)->getPosition();
	    CLHEP::Hep3Vector hit(hitPosition[0],hitPosition[1],0);

	    double s = 0;
	    if (estimated_omega==0) // a straight line
	    {
		streamlog_out(DEBUG2) << "Omega == 0" << std::endl;
		
		CLHEP::Hep3Vector  pca2hit = hit - point_of_closest_approach;
		// n_phi and pca2hit are parallel, use dot product to get
		// right sign of z 
		s= pca2hit.dot(n_phi); // the distance in the xy plane
	    }
	    else // estimated_omega!=0 , a real helix
	    {
		CLHEP::Hep3Vector centre2hit(hit - helix_centre);
		
		s= ( centre2hit.phi() -  n_pca.phi() )/fabs(estimated_omega);
	    }

	    szPairs.push_back(std::pair<double,double> (s, hitPosition[2]));
	    if ( hitPosition[2] < 0) 
		signOfZCounter--;
	    else
		signOfZCounter++;
	    streamlog_out(DEBUG1)<< "Adding szPair "<<s<<"\t"<< hitPosition[2] << std::endl;
	}
	
	// sort the whole thing in z
        // preserve the y ordering for equal zs
	std::stable_sort( szPairs.begin(), szPairs.end(), sort_by_z() );

	// now sum up all slopes m if fabs(delta_s) is smaller than pi*r 
	// (otherwise a jump in s is assumed, eg +pi -> -pi or vice versa)
	double sum_m = 0;
	int mCounter =0; // count how many 'm's have been summed for averaging
	for (std::vector< std::pair<double,double> >::iterator szIter = szPairs.begin() ;
	     szIter < szPairs.end()-1; szIter++ )
	{
	    
	    if ( estimated_omega != 0) // avoid divide_by_zero in case of straight line
		if ( std::fabs(szIter->first - (szIter+1)->first) > M_PI/estimated_omega  )
		    continue;

	    sum_m += ((szIter+1)->second - szIter->second) / ((szIter+1)->first - szIter->first);
	    mCounter++;
	}
	double estimatedTanLambda = 0.; 
	if( mCounter)
	    estimatedTanLambda =  sum_m / mCounter ;
	else // something really went wong, not one valid slope
	{
	    streamlog_out(ERROR2) << "Could not find seed value for tan(lambda), setting to 0."
				    << std::endl;
	    // estimatedTanLambda already is 0
	}
	
	// def: z = z0 + z_ref + s * tan(lambda)
	// -> z0 = z - z_ref- s * tan(lambda)
	// estimate it for the first point in the vector

	double estimatedZ0 =  szPairs.begin()->second - reference_point.z() 
			     - szPairs.begin()->first * estimatedTanLambda;

	// now check for correct sign of curvature. Assuming the particle started at the 
	// origin tan(lambda) should be positice for z > 0 and negative for z < 0
	if ( estimatedTanLambda * signOfZCounter < 0 ) // we have the wrong curvature
	{ // correct Omega, phi, d0 and tan(lambda)
	    estimated_omega *= -1.;
	    estimated_d0 *= -1.;
	    estimatedTanLambda *= -1.;
	    estimated_phi = fmod( estimated_phi + M_PI, M_PI*2. );
	}

        // now set the estimated values in the seed track
	seedTrack->setOmega( estimated_omega );
	seedTrack->setPhi( estimated_phi );
	seedTrack->setD0( estimated_d0 );
	seedTrack->setTanLambda( estimatedTanLambda );
	seedTrack->setZ0( estimatedZ0 );

	seedTrackCollection->push_back(seedTrack);

	
	streamlog_out(MESSAGE3) << std::setprecision(5) << "\tSeedTrack phi:" << std::setw(8) << seedTrack->getPhi() << " rads ,\tOmega: " 
				<< std::setw(8) << seedTrack->getOmega() << " mm^-1 ,\td0: " << std::setw(8) << seedTrack->getD0() 
				<< " mm ,\ttan(lambda): " << std::setw(8) << seedTrack->getTanLambda() 
				<< " ,\tz0: " << std::setw(8) << seedTrack->getZ0() << "mm." << std::endl;;

    } // loop tracks

    // now add the collection to the event!
    try {
	
	evt->addCollection(seedTrackCollection, _output_seed_tracks_collection_name);
    }
    catch (EVENT::EventException e) {
	
	std::cout << "SimpleTrackSeederProcessor::processEvent(). " << e.what()
		  << "Collection " << _output_seed_tracks_collection_name << " is being deleted." << std::endl;
	
	std::vector<EVENT::LCObject*>::iterator iter = seedTrackCollection->begin();
	
	for(; iter != seedTrackCollection->end(); iter++) {
	    
	    if(*iter) {
		
		delete (*iter);
		(*iter) = NULL;
	    }
	}
	
	seedTrackCollection->clear();
	delete seedTrackCollection;
    }
}// processEvent


Processor* SimpleTrackSeederProcessor::newProcessor() {

	return new SimpleTrackSeederProcessor();
}


void SimpleTrackSeederProcessor::check(LCEvent *evt) {


}


void SimpleTrackSeederProcessor::end() {


}

} // namespace marlintpc
