#include <TimeShiftCorrectorProcessor.h>

#include <IMPL/TrackerDataImpl.h>

#include <marlin/Exceptions.h>

#include <LCObjectCopier.h>

namespace marlintpc {

TimeShiftCorrectorProcessor::TimeShiftCorrectorProcessor() : marlin::Processor( "TimeshiftCorrector" ) {

  // the processor description
  _description = "TimeShiftCorrectorProcessors applies time offset corrections to individual channels." ;

  // register steering parameters: name, description, class-variable, default value
  registerInputCollection( LCIO::TRACKERDATA,
                           "InputCollectionName",
                           "Name of the input collection",
                           _inputColName,
                           std::string("TPCData") );

  registerProcessorParameter( "ChannelCorrectionCollectionName",
			      "Name of the collection with the channel corrections",
			      _channelCorrColName,
			      std::string("TPCChannelCorrection") );

  registerOptionalParameter( "OutputCollectionName",
                              "Optional: Name of output collection",
                              _outputColName,
                              std::string("TPCDataTimeShiftCorrected") );

  registerProcessorParameter( "SetOutputTransient" , 
			      "If not 0 the output collection is set transient (only if OutputCollectionName defined)"  ,
			      _outputIsTransient,
			      0);

}


void TimeShiftCorrectorProcessor::processRunHeader( LCRunHeader* run ) 
{
    run->parameters().setValue( _processorName + "_revision", "$Rev: 762 $" );

    for ( marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
        if ( ! i->second->isOptional() || i->second->valueSet() ) {
            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
}


void TimeShiftCorrectorProcessor::processEvent( LCEvent * evt ) 
{ 
    // Get input collection and check if there are data in it
    // Cast the input to LCCollectionVec if we want to use the iterator or modify it 
    IMPL::LCCollectionVec* inputColl = 0;
    try {
        inputColl = dynamic_cast<IMPL::LCCollectionVec*>( evt->getCollection( _inputColName ) );

	if ( inputColl->size() == 0 )
	    throw DataNotAvailableException( std::string("Empty TrackerData collection" )
					     + _inputColName);
    }
    catch( EVENT::DataNotAvailableException& e ) {
        std::cout << _processorName << ": Empty event "<< evt->getEventNumber() << std::endl;
	throw marlin::SkipEventException( this );
    }

    // The TrackerData collection is the "output".
    // If the input collection is writeable and no output collection name
    // is given we use the input collection
    IMPL::LCCollectionVec* trackerDataColl = 0;

    // Test if the collection is writeable
    try {
        dynamic_cast<IMPL::TrackerDataImpl*>( inputColl->getElementAt(0) )->setTime(
		     dynamic_cast<EVENT::TrackerData*>( inputColl->getElementAt(0) )->getTime() );
    }
    catch ( lcio::ReadOnlyException& e ) {
        // Check if output collection name is given
        if ( ! parameterSet( "OutputCollectionName" ) ) {
	    std::cerr << "Error: Collection is not writeable."<< std::endl;
	    std::cerr << "You have to specify an OutputCollectionName." << std::endl;
	    std::cerr << "Remember to adapt the input collection name of the next processor!" << std::endl;
	    throw lcio::ReadOnlyException( _inputColName + " not writeable" );
	}
    }

    // Clone input if OutputCollectionName is given
    // (note that this is mandatory if input is writeOnly)
    if ( parameterSet( "OutputCollectionName" ) ) {
        trackerDataColl = new LCCollectionVec( "TrackerData" );
	for ( IMPL::LCCollectionVec::iterator it = inputColl->begin(); it < inputColl->end() ; it++ ) {
	    trackerDataColl->addElement( lcio::LCObjectCopier::copy( dynamic_cast<IMPL::TrackerDataImpl*>(*it) ) );
	}

	// set the transient flag as definded by the processor parameter
	trackerDataColl->setTransient( ( _outputIsTransient != 0 ) );
    }
    else {
        trackerDataColl = inputColl;
    }

    EVENT::LCCollection* chanCorrColl = evt->getCollection( _channelCorrColName );

    _channelCorrectionHandler.setChannelCorrectionCollection( chanCorrColl );

    for ( IMPL::LCCollectionVec::iterator it = trackerDataColl->begin();
	 it < trackerDataColl->end(); it++ ) {

      IMPL::TrackerDataImpl* trackerData = dynamic_cast<IMPL::TrackerDataImpl*>( *it );

      float oldtime = trackerData->getTime();
      float timeshift = _channelCorrectionHandler.getChannelTimeOffset( trackerData->getCellID0() );
      trackerData->setTime( oldtime - timeshift );

    }
    
    // if the pulses collection has been created by cloning
    // we have to add it to the event
    if ( parameterSet( "OutputCollectionName" ) ) {
      evt->addCollection( chanCorrColl, _outputColName );
    }

}

} // namespace marlintpc
