#include "TrackFinderRectangularProcessor.h"

#include <iostream>
#include <stdexcept>
#include <string>

//LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerHit.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackImpl.h>
//GEAR
#include <gear/PadRowLayout2D.h>
#include <marlin/Global.h>
#include <gear/TPCParameters.h>


using namespace lcio ;
using namespace marlin ;
using namespace std ;

namespace marlintpc{

TrackFinderRectangularProcessor aTrackFinderRectangularProcessor;
  

TrackFinderRectangularProcessor::TrackFinderRectangularProcessor() 
    : Processor("TrackFinderRectangularProcessor") ,
      isInTrack(NULL), isInTrackCandidate(NULL), nHits(0)
{
  // modify processor description
  _description = "TrackFinderRectangularProcessor searches straight tracks using linear regression" ;
  

  // register steering parameters: name, description, class-variable, default value
  registerInputCollection(    LCIO::TRACKERHIT,
			      "InputCollectionName" , 
			      "Name of the input collection"  ,
			      _inputColName ,
			      std::string("TPCHits") ) ;
  
  registerOutputCollection(   LCIO::TRACK,
			      "OutputCollectionName" , 
			      "Name of the output collection"  ,
			      _outputColName ,
			      std::string(" TPCSeedTracks") ) ;
  
  registerProcessorParameter( "SetOutputTransient" , 
			      "if not 0 the output collection is set transient"  ,
			      _outputIsTransient,
			      0);
	
  registerProcessorParameter( "MinTrackHits" , 
			      "Minimum number of hits on track",
			      _minTrackHits ,
			      int(5) ) ;

  registerProcessorParameter( "MaxSkipRows" , 
			      "Maximum number of subsequnetly missing hits",
			      _maxSkipRows ,
			      int(1) ) ;

  registerProcessorParameter( "DeltaX" , 
			      "Maximum distance of predicted and measured hit in x in mm",
			      _deltaX ,
			      float(3.0) ) ;
  registerProcessorParameter( "DeltaY" , 
			      "Maximum distance of predicted and measured hit in y in mm",
			      _deltaY ,
			      float(0.01) ) ;
  registerProcessorParameter( "DeltaZ" , 
			      "Maximum distance of predicted and measured hit in z in ns",
			      _deltaZ ,
			      float(500) ) ;

  registerProcessorParameter( "VerboseLevel" , 
			      "Level of prozessor verbosity (0 .. 5)",
			      _verbosityLevel ,
			      int(1) ) ;

  // initialise the straight line (just set everyting to 0)
  straightLine.a=0; 
  straightLine.b=0;
  straightLine.c=0;
  straightLine.d=0;

  _padLayout = NULL;
  
}
 

void TrackFinderRectangularProcessor::init() { 
 
 _padLayout = &( Global::GEAR->getTPCParameters().getPadLayout() );

 if (_verbosityLevel>0)  
    printParameters() ;

  _nRun = 0 ;
  _nEvt = 0 ;
  
}

void TrackFinderRectangularProcessor::processRunHeader( LCRunHeader* run) 
{ 
    run->parameters().setValue( _processorName + "_revision", "$Rev: 762 $");

    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
        if ( ! i->second->isOptional() || i->second->valueSet() ) {
            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }

    _nRun++ ;
} 

void TrackFinderRectangularProcessor::zeroFlags(bool *IsIn) 
{
  if (_verbosityLevel>=5)
    cout << "zeroing "<<nHits<<endl;//DEBUG

  for (unsigned int i=0; i<nHits; i++) IsIn[i]=false ;
}

void TrackFinderRectangularProcessor::addTrackCandidateFlags()
{
  if (_verbosityLevel>=5)
    cout << "adding tracks flags"<<endl;//DEBUG
    
  for (unsigned int i=0; i<nHits; i++)
  {
    if (isInTrackCandidate[i])
	    isInTrack[i]=true;
  }
}

void TrackFinderRectangularProcessor::processEvent( LCEvent * evt ) 
{ 
  _nEvt++;

  // if verbose mode is not 0 print every 100th event number
  // if verbose mode is >=2 (every event) print every event number
    if ( ( (evt->getEventNumber() % 100 == 0 ) && _verbosityLevel > 0 ) ||
	 ( _verbosityLevel >=2)
       )
      cout << "\r proceccing event "<<evt->getEventNumber()<<flush;

    EVENT::LCCollection *TPCHits=evt->getCollection(_inputColName);

     // get number of hits in this event
    nHits=TPCHits->getNumberOfElements();

    //Collection containing  all found seed track
    LCCollectionVec  *seedTracksCol=new LCCollectionVec( LCIO::TRACK );
    
    // set the transient flag as definded by the processor parameter
    seedTracksCol->setTransient( (_outputIsTransient != 0) );

    // if the number of hits is smaller than the minimum
    // number fo track hits add an empty collection
    // and return
    if (nHits < static_cast<unsigned int>(_minTrackHits) )
    { 
	evt->addCollection( seedTracksCol,_outputColName );
	return;
    }

    //DEBUG: Dump all hits
    if (_verbosityLevel >=2)
    {
      cout << "looking for tracks in event with "<< nHits<<" hits." <<endl;
      cout << "hits are (x,y,z):"<<endl;
      for (unsigned int i=0; i<nHits ; i++)
      {
	  TrackerHit *hit=static_cast<TrackerHit*>(TPCHits->getElementAt(i));
	  const double *coords=hit->getPosition();
	  cout <<i<<"\t"<<coords[0]<<"\t"<<coords[1]<<"\t"<<coords[2]<<endl;//DEBUG
      }
    }

    // Dynamically generate variables neede during one run
    TrackerHitImpl *xhit = new TrackerHitImpl;//exptrapolated hit
    TrackImpl  *trackCandidate=new TrackImpl;

    // generate flag arrays with the size of nHits
    // and set them all false
    isInTrack = new bool[nHits];
    isInTrackCandidate = new bool[nHits];
    zeroFlags(isInTrack);
    zeroFlags(isInTrackCandidate);

 // How this algorithm works:
 // To find a straight track, the algorithm takes the first and the 
 // last hit of a track and looks for hits on this track comparing the
 // the position of the next extrapolated hit with all measured hits.
 // If a hit has already been found to be member of a track, it is not
 // used in further steps. This information is saved in the IsInTrack
 // array.
 // The hits have to be sorted in y (rows ), within this in x, 
 // and within x in z
 //
 // There are four loops for searching:
 // loop over i: running down from nHits-1 to 0 containing the (index of
 //              the) last hit of the track
 // loop over j: running up from 0 to i-1 containing the first hit
 //              of the track
 // loop over k: running up from j+1 to i-1 testing all hits between
 //              first and last hit of the track
 // loop over skipRows: running from 0 to _maxSkipRows counting the 
 //                     number of missing hits
  

    // loop all possible end points of the track
    // we need a singed counter because of the range check 
    for (int signed_i=nHits-1; signed_i>=0; signed_i--)
    { //i holds geometrically last hit
      unsigned int i=signed_i; // this should be save due to the range of the loop
      // within the loop we want an unsigned int as array index and to
      // avoid compiler warnings

      // get last not tested hit in event
      TrackerHit *lastHit=static_cast<TrackerHit*>(TPCHits->getElementAt(i));
      
      // test if hit is already in a track
      if (isInTrack[i])
      {
	if (_verbosityLevel >= 4)
	  cout << "skipping last hit "<<i<<", already in track"<<endl;
	continue;// try next hit if hit i already in another track
      }
	
      // loop all possible starting points of the track
      for (unsigned int j=0 ; j<i; j++) //j holds geometrically first hit
      {

	// get first not tested hit in event
	TrackerHit *firstHit=static_cast<TrackerHit*>(TPCHits->getElementAt(j));
	    
	// test if hit is already in a track
	if (isInTrack[j])
	{
	  if (_verbosityLevel >= 4)
	    cout << "skipping first hit "<<j<<", already in track"<<endl;
	  continue;// try next hit if hit j already in another track
	}
	    
	// the two seed hits have been found, now we can initialise a new track candidate

	// reset track candidate's flags and set flags for the seed hits
	zeroFlags(isInTrackCandidate);
	isInTrackCandidate[i]=true;
	isInTrackCandidate[j]=true;

	// some check code, the conditions should never be true
	// throw exceptions, which avoid segemtation faults and simplifies debugging
	if (firstHit==NULL)
	  {
	    cerr << "something went wrong, empty first hit "<<i<<endl;
	    throw std::runtime_error(string("TrackFinderRectangularProcessor::processEvent hit is NULL "));
	  }
	if (lastHit==NULL)
	  {
	    cerr << "something went wrong, empty last hit "<<i<<endl;
	    throw std::runtime_error(string("TrackFinderRectangularProcessor::processEvent hit is NULL "));
	  }

	// actually add the two seed hits to the track candidate
	trackCandidate->addHit(lastHit);
	trackCandidate->addHit(firstHit);

	// we have to remember which hit was the last added one
	int lastAddedHit=j;

	// some debug output which helps to follow what the algorithm is doing
	if (_verbosityLevel >= 3)
	{
	  cout << "starting track with hit "<<j<<" and "<<i<<endl;//DEBUG 
	  const double *firstHitCoords = firstHit->getPosition();//DEBUG
	  cout << j<< " "<<firstHitCoords[0] <<" "//DEBUG
	       << firstHitCoords[1] <<" "<< firstHitCoords[2] <<endl;//DEBUG
	  const double *lastHitCoords = lastHit->getPosition();//DEBUG
	  cout << i<< " "<<lastHitCoords[0] <<" "//DEBUG
	       << lastHitCoords[1] <<" "<< lastHitCoords[2] <<endl;//DEBUG
	}

	// calculate the track parameters a, b, c and d for the seed track
	doRegression(trackCandidate);

	// a flag whether to recalculate the retrapolated hit 
	// saves computing time if only performed when needed
	bool redoExtrapolation=true;

	// now loop the rows until the maximum number of missing hits is reached
	// the reason why this is the outer loop is that first all hits in the next
        // row have to be checked before the hits in the next to next row is considered
	for (unsigned int skipRows=0; skipRows <= static_cast<unsigned int>(_maxSkipRows); skipRows++)
 	{

	  // the innermost loop checks all hits from the last added hit to the 
	  // geometrically last hit with index i
	  for (unsigned int k=lastAddedHit+1; k < i; k++)  
	  {
	    //extrapolate the next hit if necessary
	    if (redoExtrapolation)
	    {
	      // get all hits from the track candidate
	      EVENT::TrackerHitVec hitVec=trackCandidate->getTrackerHits();
	 
	      
		
	      // Calculate hit from the y position.
	      // The extrapolated y position is the y position ( getPosition()[1] )
	      // of the last hit (  hitVec.back()  ) 
	      // in the hit vector of the track candidate,
	      // plus the height of skipRows to go
	      	


		//get the index of the current hit
                int currentHitIndex =_padLayout->getNearestPad(hitVec.back()->getPosition()[0],hitVec.back()->getPosition()[1]);

		//get the row number of the current hit
		int row = _padLayout->getRowNumber(currentHitIndex);

		//add the number of skipRows to the row number and get the pad index 
		int firstPadIndexRow=0;
		try
		{
		   firstPadIndexRow = _padLayout->getPadIndex ( row + (skipRows+1),0);
		}
		catch(std::out_of_range &e) // There is no row number ( row + (skipRows+1) in the pad layout
		{  // stop scanning for further points
		

		   break;

 		}
		//get the y coordinate and extrapolate the next hit
		*xhit= extrapolateNextHit( _padLayout->getPadCenter(firstPadIndexRow)[1]);
		


		// no need to redo the extrapolation until a new row is tested
	      redoExtrapolation=false;

	      // debug output to follow what the algorithm is doing
	      if (_verbosityLevel>=4) 
	      {
		const double *xhitCoords = xhit->getPosition();//DEBUG
		cout << "extrapolated hit "<< xhitCoords[0] <<" "//DEBUG
		     << xhitCoords[1] <<" "<< xhitCoords[2] <<endl;//DEBUG	
	      }
	    }
		    
	    // now get the hit to be tested and it's coordinates
	    // as well a the extrapolated hit's coordinates
	    TrackerHit *testHit=static_cast<TrackerHit*>(TPCHits->getElementAt(k));
	    const double *testHitCoords=testHit->getPosition();
	    const double *xhitCoords=xhit->getPosition();

	    // again the debug output to follow the algorithm
	    if (_verbosityLevel>=4) 
	      cout << "testing hit "<< testHitCoords[0] <<" "//DEBUG
		   << testHitCoords[1] <<" "<< testHitCoords[2] <<endl;//DEBUG

	    //check if there is the measured hit that fits extrapolation
	    // within the search window (+- deltaX, +- deltaY, +-deltaZ)
	    // The hit is only valid if not in already another track
	    if ( (testHitCoords[0] >= xhitCoords[0] - _deltaX )
		 &&(testHitCoords[0] <= xhitCoords[0] + _deltaX)
		 &&(testHitCoords[1] >= xhitCoords[1] - _deltaY)
		 &&(testHitCoords[1] <= xhitCoords[1] + _deltaY)
		 &&(testHitCoords[2] >= xhitCoords[2] - _deltaZ)
		 &&(testHitCoords[2] <= xhitCoords[2] + _deltaZ)
		 && !isInTrack[k] )
	      { // hit is OK, add it

		// some check code, the conditions should never be true
		// throw exceptions, which avoid segemtation faults and simplifies debugging
		if (testHit==NULL)
		{
		  cerr << "something went wrong, empty hit "<<k<<endl;
		  throw std::runtime_error(string("TrackFinderRectangularProcessor::processEvent hit is NULL "));
	        }
		trackCandidate->addHit(testHit);
		isInTrackCandidate[k]=1;

		// Remember which was the last added hit, so the search continues at the correct position.
		// This is needed in case there are no matching hits for this y coordinate and we
		// have to search in the next row.
		lastAddedHit=k;

		if (_verbosityLevel>=4) 
		  cout << "adding hit "<< testHitCoords[0] <<" "//DEBUG
		       << testHitCoords[1] <<" "<< testHitCoords[2] <<endl;//DEBUG
		
		// Now that a hit has been found search in the next row.
		// To calculate the extrapolated hit the redoExtrapolation
		// flag has to be set
		skipRows=0; // start searching without row skip
		redoExtrapolation=true;

		continue; // with next hit in sample
	      }//if hit fits
		    
	    // If y of testhit ist larger than y of the extrapolated hit
	    // there won't be any more mathing hits for this xhit
	    // as hits are sortet in y.
	    // In this case quit the k-loop and continue search in next row
	    if (testHitCoords[1] > xhitCoords[1] + _deltaY)
	      break;

	  }// for k
		
	  // To calculate the extrapolated hit in the next row, set the 
	  // redoExtrapolation ftag to true
	  redoExtrapolation=true;
		
	} // for skipRows

	// Now the search for this starting and end point is finished.
	// Check whether the track candidate is a valid track

	// Get the vector of hits on the track
	EVENT::TrackerHitVec hitVec=trackCandidate->getTrackerHits();
	
			
		//get the Index of the first seed hit
		 int firstSeedIndex =_padLayout->getNearestPad(hitVec[0]->getPosition()[0],hitVec[0]->getPosition()[1]);
              	 //get the rownumber of the the first seed hit
		int firstSeedRow = _padLayout->getRowNumber(firstSeedIndex); 



		//get the Index of last added hit
		int lastAddedIndex =_padLayout->getNearestPad(hitVec.back()->getPosition()[0],hitVec.back()->getPosition()[1]);
			//getthe row number of the last added hit		
		int lastAddedRow = _padLayout->getRowNumber(lastAddedIndex); 

	// Debug output: dump some parameters of the track candidate
	if (_verbosityLevel>=3) 
	  cout << "testing track candidate with length "<<hitVec.size()
	       << " y1 " <<  firstSeedRow
	       << " y2 " << lastAddedRow
                    	    + _maxSkipRows+1    << endl;

	

	// The track is accepted if four conditions are fulfilled:
	// 1. It is long enough, i. e. it has minimum minTrackHits hits
	// 
	//    The geometrically last hit,  which was the first seed hit,
	// 2. is not more than max SkipRows away from the last added hit
	// 3. is not at the same y-position as the last added hit
	// 4  is within the deltaX and deltaZ window around the fitted track
	
	

	if 
	( 
	 //  1.track is long enough
	(hitVec.size() >= static_cast<unsigned int>(_minTrackHits))

	 &&
	//   2.the seed hit [0] is not more than _maxSkipRows away
	//      from the last added hit
	(firstSeedRow  <=  lastAddedRow +   (_maxSkipRows+1))
	
      	 && 
	//    3.hit[0]is not the same  y position as last added hit
	(firstSeedRow  != lastAddedRow )
	
  	 &&	
	//     4.hit[0] is within deltaX and deltaZ search window, i. e. it actually is on the track
	     ( ( (hitVec[0]) ->getPosition()[0]  >= 
	            straightLine.a*(hitVec[0]) ->getPosition()[1] + straightLine.b - _deltaX ) &&
	       ( (hitVec[0]) ->getPosition()[0]  <= 
	            straightLine.a*(hitVec[0]) ->getPosition()[1] + straightLine.b + _deltaX ) &&
	       ( (hitVec[0]) ->getPosition()[2]  >= 
	            straightLine.c*(hitVec[0]) ->getPosition()[1] + straightLine.d - _deltaZ ) &&
	       ( (hitVec[0]) ->getPosition()[2]  <= 
		    straightLine.c*(hitVec[0]) ->getPosition()[1] + straightLine.d + _deltaZ )     )
         )// if conditions


	   
	{
	    // Track candidate is valid.

	    // Add it's hit to the array of hits which are in a track
	    addTrackCandidateFlags();

	    // Add the track candidate to the collection of seed tracks
	    seedTracksCol->addElement(trackCandidate);
		
            // Now that the track candidate has been handed over to the seed track collection
	    // we allocate new memory for the next candidate
	    trackCandidate=new TrackImpl;

	    if (_verbosityLevel>=3) 
	      cout << "Adding track." << endl;
	    
	    // leave the j loop which loops all possible starting point
	    // the algorithm will continue with the next i, which is the end point of a new track
	    break; 
	}
	else
	{
	  // The track candidate was not valid.
	  // As LCIO objects don't have a clear function (at least none I know of)
	  // we simply delete the candidate and allocate a new, empty one
	  delete trackCandidate;
	  trackCandidate=new TrackImpl;

	  if (_verbosityLevel>=3) 
	    cout << "Rejecting track candidate." << endl;

	}
      } //for j, the loop over the start hits
    }//for i, the loop over the end hits

    // All loops are finished, add the collection to the event
    evt->addCollection( seedTracksCol,_outputColName );

    // Some debug output
    if (_verbosityLevel>=2) 
      cout << "TrackFinderRectangularProcessor:  found " << seedTracksCol->getNumberOfElements()  << " tracks." << endl;

    // Do the cleaning up.
    // We have to delete the objects we allocated
    // but we must not delete the collection, which has been
    // handed over to the event.
    delete xhit;
    delete trackCandidate;

    delete []isInTrack;
    delete []isInTrackCandidate;
}



void TrackFinderRectangularProcessor::check( LCEvent * evt ) { 
  // nothing to check here - could be used to fill checkplots in reconstruction processor
}


void TrackFinderRectangularProcessor::end(){ 
  if (_verbosityLevel>0)
  {
    std::cout << "TrackFinderRectangularProcessor::end()  " << name() 
	      << " processed " << _nEvt << " events in " << _nRun << " runs "
	      << std::endl ;
  }
}

void TrackFinderRectangularProcessor::doRegression(Track *trackCandidate)
{
  // Do the actual linear regression 


  // Reset the straight line parameters.
  // In case there is only one or no hits (which should not
  // be the case, but who knows...) these parameters are kept.
  // It's a straight line along the y axis 
  // through the origin
  straightLine.a=0;
  straightLine.b=0;
  straightLine.c=0;
  straightLine.d=0;


  // These are the sums needed to calculate the
  // track parameters (see documentation in .h file)
  float sumX=0; float sumY=0;float sumXY=0;
  float sumZ=0; float sumYZ=0;
  float sumYsqare=0;  // = Sum (Y^2)

  // get the vector of hits from the track candidate
  EVENT::TrackerHitVec hitVec=trackCandidate->getTrackerHits();
  
  // save the number of hits in a separate variable
  // makes the code nicer
  unsigned int n=hitVec.size();

  // some debug output
  if (_verbosityLevel>=5)
    cout << "hitVec.size() is "<<hitVec.size()<<endl;

  // Regression can only be performed if there are at least
  // two hits on the track. Otherwise we get a divide by zero
  if (n>=2)
  {
    // debug:
    if (_verbosityLevel>=5)
      cout << "n is "<<n<<endl;

    // loop all hits and calculate the sums needed
    for (unsigned int i=0;i<n;i++)
    {
        // get the particular hit from the vector
        TrackerHit *hit=hitVec[i];

	// check whether the hit pointer is valid
	// If the algorithm is working this should always
	// be the case. Otherwise throw an exception
	// to simplify debugging.
	if (hit==NULL)
	{
	    cerr << "something went wrong, got empty hit "<<i<<"from the vector"<<endl;
	    throw std::runtime_error(string("TrackFinderRectangularProcessor::doRegression hit is NULL "));
	}

	// add the variables to the sums
        sumX+=hit->getPosition()[0];
        sumY+=hit->getPosition()[1];
        sumXY+=hit->getPosition()[0]*hit->getPosition()[1];
        sumZ+=hit->getPosition()[2];
        sumYZ+=hit->getPosition()[1]*hit->getPosition()[2];
        sumYsqare+=hit->getPosition()[1]*hit->getPosition()[1];
    }

    // Now that the sums are calculated we can compute the
    // track parameters
    straightLine.a=(sumX*sumY-n*sumXY)/(sumY*sumY-n*sumYsqare);
    straightLine.b=(sumY*sumXY-sumYsqare*sumX)/(sumY*sumY-n*sumYsqare);
    straightLine.c=(sumZ*sumY-n*sumYZ)/(sumY*sumY-n*sumYsqare);
    straightLine.d=(sumY*sumYZ-sumYsqare*sumZ)/(sumY*sumY-n*sumYsqare);

  }// if  n>=2
}

TrackerHitImpl TrackFinderRectangularProcessor::extrapolateNextHit(double y)
{
    // caclulate x and z on the straight line from the y 
    // coordinate via the parametrisation:
    // x = a*y +b
    // z = c*y +d

    // the instance of the hit we want to return
    TrackerHitImpl hit;

    // an array to store the three coordinates
    double position[3];

    // debug:
    if (_verbosityLevel>=5)
      cout << "straight line is " <<straightLine.a<<" "<<straightLine.b<<" "
	   <<  straightLine.c<<" "<<straightLine.d<<endl;
    
    // calculate the 3D position
    position[0]=straightLine.a*y+straightLine.b;
    position[1]=y;
    position[2]=straightLine.c*y+straightLine.d;

    // set the hit's position and return it
    hit.setPosition(position);

    return hit;
}

} //namespace marlintpc
