
// STL
#include <vector>
#include <map>
#include <iomanip>

#include <cmath>

// LCIO
#include "IMPL/LCCollectionVec.h"
#include "IMPL/LCFlagImpl.h"
#include "IMPL/TrackImpl.h"
#include "IMPL/TrackerHitImpl.h"
#include "IMPL/TrackerPulseImpl.h"
#include "EVENT/LCIO.h"
#include "EVENT/LCCollection.h"
#include "EVENT/LCObject.h"
#include "EVENT/Track.h"
#include "EVENT/TrackerHit.h"
#include "EVENT/TrackerPulse.h"
#include "EVENT/TrackerData.h"
#include "Exceptions.h"

// GEAR
#include "gear/GEAR.h"
#include "gear/TPCParameters.h"
#include "gear/PadRowLayout2D.h"
#include "gearxml/GearXML.h"

// tpcconddata
#include "TPCConditions.h"

// CLHEP
#include "CLHEP/Matrix/Matrix.h"
#include "CLHEP/Matrix/Vector.h"
#include "CLHEP/Vector/LorentzVector.h"
#include "CLHEP/Vector/TwoVector.h"

// AIDA
#include <AIDA/AIDA.h>

// Marlin
#include "marlin/Global.h"
#include "marlin/AIDAProcessor.h"
#include "marlin/Exceptions.h"
#include "TrackFitterLikelihoodProcessor.h"

// MINUIT
/* NOTE this uses the C++ version of Minuit - not the Fortran version */
#include "Minuit2/MnPrint.h"
#include "Minuit2/MnUserParameters.h"
#include "Minuit2/MnUserCovariance.h"
#include "Minuit2/MnStrategy.h"
#include "Minuit2/MnMinimize.h"
#include "Minuit2/FunctionMinimum.h"

using namespace marlin;
using namespace ROOT;
using namespace Minuit2;

using namespace std;

namespace marlintpc
{

TrackFitterLikelihoodProcessor aTrackFitterLikelihoodProcessor;

TrackFitterLikelihoodProcessor::TrackFitterLikelihoodProcessor():
	Processor("TrackFitterLikelihoodProcessor"),
	_seed_track_collection_name("TPCSeedTracks"),
	_fitted_track_collection_name("TPCTracks"),
	_tpc_conditions_collection_name(tpcconddata::TPCConditions::getDefaultColName()),
	_fit_sigma0(false),
	_p_noise(0.01),
	_maximum_drift_length(0.),
	_sigma_z(0.) {

	_description = "Fits a Likely Track to the signals collected on the readout geometry.";

	registerOptionalParameter("SeedTrackCollection",
				"The name of the Seed Track collection.",
				_seed_track_collection_name,
				std::string("TPCSeedTracks"));

	registerOptionalParameter("FittedTrackCollection",
				"A name for the collection the fitted track will be stored with.",
				_fitted_track_collection_name,
				std::string("TPCTracks"));

	registerOptionalParameter( "TPCConditionsCollection" ,
				    "Name of the collection containing the TPC conditions",
				    _tpc_conditions_collection_name,
				    tpcconddata::TPCConditions::getDefaultColName() );

	registerProcessorParameter("Sigma0",
				"The constant diffusion for each electron cloud in the TPC",
				_sigma0,
				0.5);

	registerProcessorParameter("FitSigma0?",
				"Controls whether the minimization should try and fit Sigma0 as a free parameter (true) or leave it as a constant (false - default)",
				_fit_sigma0,
				false);

	registerProcessorParameter("Noise",
				"A parameter to reduce the affect of spurious signals.",
				_p_noise,
				0.01);

	registerProcessorParameter("SigmaZ",
				"The standard deviation of the primary ionizations in the z direction (mm)",
				_sigma_z,
				0.);
				
	registerOptionalParameter("ProduceTrackTuples",
				"Controls whether tuples for the track paramters and their error are produced.",
				_produce_tuples,
				true);
}


TrackFitterLikelihoodProcessor::~TrackFitterLikelihoodProcessor() {

}


void TrackFitterLikelihoodProcessor::init() {

	// usually a good idea to
	printParameters();

	// get the position of the electronics
	const gear::TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();
	_maximum_drift_length = tpc_parameters.getMaxDriftLength();
	
	if (_produce_tuples)
	{
	  AIDA::ITupleFactory *tupleFactory = AIDAProcessor::tupleFactory(this);
	  assert( tupleFactory );
	  
	  _track_result_tuples = tupleFactory->create("Track Results", "Results of the Track Fitter", "int event_number, boolean valid = false, int F_calls, double F_minimum, sigma0, sigma0_err, edm");
	  assert( _track_result_tuples );
	}
}


void TrackFitterLikelihoodProcessor::processRunHeader(LCRunHeader *run) {

}

IMPL::TrackImpl TrackFitterLikelihoodProcessor::find_most_likely_fit(EVENT::Track* seed_track) {

	streamlog_out( MESSAGE4 ) << "Fitting a track to data.\n";
	// set the member variable _current_seed_track to allow access to the data during the MINUIT minimization
	_current_seed_track = seed_track;

	/* these are parameters which are minimized with Minuit. For more information see "Track Parameters for the marlin Reconstruction Framework" by Thomas Kraemer.
	*	param#. name: description
	*	0.	phi: the azimuthal angle of the momentum of the particle at the p.c.a. in {x,y}
	*	1.	omega: the curvature of the track where omega = 1/R where
	*				R is the radius of the curvature of the track
	*				sign(omega) is positive (negative) if moving along the momentum of the particle it travels in a (counter-)clockwise direction
	*	2.	d0: the signed impact parameter in the xy plane.
	*	3.	tan(lambda): the slope dz/ds of the straight line in the sz plane
	*	4.	z0: the z position of the track at the p.c.a. with respect to the reference point
	*	5.	sigma: the standard deviation of the transverse diffusion of the electron clouds
	*/

	MnUserParameters user_parameters;
	
	// omega should be sign(charge) * eE-04 * (B) / transverseP ... ie) -3E-04 * 4 / 25 >>>>>>>>>>>>>0.00029805
	
	user_parameters.Add("phi", _current_seed_track->getPhi(), 1.E-04);
	user_parameters.Add("omega", _current_seed_track->getOmega(), 1.E-08);
	user_parameters.Add("d0", _current_seed_track->getD0(), 1.E-03);
	user_parameters.Add("tanLambda", _current_seed_track->getTanLambda(), 1.E-02);
	user_parameters.Add("z0", _current_seed_track->getZ0(), 1.E-01);

	if(_fit_sigma0) {
		// fit sigma0 with a range of [0, 10]
		user_parameters.Add("sigma0", _sigma0, 1.E-02, 0., 10.);
	}
	else {
		user_parameters.Add("sigma0", _sigma0);
	}

	MnMinimize minimizer(*this, user_parameters);

	FunctionMinimum minimum = minimizer();
	
	if( !minimum.IsValid() ) {

		minimum = minimizer();
	}

	streamlog_out( MESSAGE4 ) << minimum;

	// SOME STUFF FOR DEBUGGING THE PROCESSOR
	/*std::vector<double> fake_parameters(6);
	fake_parameters[0] = 1.5706725;
	fake_parameters[1] = -4.7964287E-05;
	fake_parameters[2] = 0.032295986;
	fake_parameters[3] = 4.1693919E-05;
	fake_parameters[4] = 9.4378003;
	fake_parameters[5] = 0.63375213;
	
	streamlog_out( MESSAGE ) << "FAKE TRACK PARAMETERS BELOW" << std::endl;
	this->operator()(fake_parameters);
	streamlog_out( MESSAGE ) << "END OF FAKE TRACK PARAMETERS" << std::endl;*/

	IMPL::TrackImpl best_fit_track;

	const MnUserParameters &fit_parameters = minimum.UserParameters();

	best_fit_track.setPhi(fit_parameters.Value("phi"));
	best_fit_track.setOmega(fit_parameters.Value("omega"));
	best_fit_track.setD0(fit_parameters.Value("d0"));
	best_fit_track.setTanLambda(fit_parameters.Value("tanLambda"));
	best_fit_track.setZ0(fit_parameters.Value("z0"));
	best_fit_track.addTrack( seed_track );

	best_fit_track.setNdf(fit_parameters.VariableParameters());

	if( minimum.HasCovariance() ) {
		// record the covariances... kindof ugly but this is how it has to be in case some of the parameters are fixed
		std::map<std::string, int> cov_index_by_name;
		int index = 1;

		const std::vector<MinuitParameter> &param_vec = fit_parameters.Parameters();
		for( std::vector<MinuitParameter>::const_iterator parameter = param_vec.begin(); parameter != param_vec.end(); parameter++ ) {
	
			if( ! (*parameter).IsFixed() ) {
	
				cov_index_by_name.insert( std::pair<std::string, int>(std::string( (*parameter).Name()), index++ ) );
			}
		}
	
		const int omega_index = cov_index_by_name["omega"];
		const int phi_index = cov_index_by_name["phi"];
		const int d0_index = cov_index_by_name["d0"];
		const int tanLambda_index = cov_index_by_name["tanLambda"];
		const int z0_index = cov_index_by_name["z0"];

		const MnUserCovariance& mn_covariance = minimum.UserCovariance();
		EVENT::FloatVec covariance(15);
	
		covariance[0] = ( omega_index )? mn_covariance(omega_index-1, omega_index-1): 0.;
		covariance[1] = ( omega_index && phi_index )? mn_covariance(omega_index-1, phi_index-1): 0.;
		covariance[2] = ( omega_index && d0_index )? mn_covariance(omega_index-1, d0_index-1): 0.;
		covariance[3] = ( omega_index && z0_index )? mn_covariance(omega_index-1, z0_index-1): 0.;
		covariance[4] = ( omega_index && tanLambda_index )? mn_covariance(omega_index-1, tanLambda_index-1): 0.;
		covariance[5] = ( phi_index )? mn_covariance(phi_index-1, phi_index-1): 0.;
		covariance[6] = ( phi_index && d0_index )? mn_covariance(phi_index-1, d0_index-1): 0.;
		covariance[7] = ( phi_index && z0_index )? mn_covariance(phi_index-1, z0_index-1): 0.;
		covariance[8] = ( phi_index && tanLambda_index )? mn_covariance(phi_index-1, tanLambda_index-1): 0.;
		covariance[9] = ( d0_index )? mn_covariance(d0_index-1, d0_index-1): 0.;
		covariance[10] = ( d0_index && z0_index )? mn_covariance(d0_index-1, z0_index-1): 0.;
		covariance[11] = ( d0_index && tanLambda_index )? mn_covariance(d0_index-1, tanLambda_index-1): 0.;
		covariance[12] = ( z0_index )? mn_covariance(z0_index-1, z0_index-1): 0.;
		covariance[13] = ( z0_index && tanLambda_index )? mn_covariance(z0_index-1, tanLambda_index-1): 0.;
		covariance[14] = ( tanLambda_index )? mn_covariance(tanLambda_index-1, tanLambda_index-1): 0.;

		streamlog_out( DEBUG3 ) << "Covariance:" << "\n\t" << covariance[0] << '\t' << covariance[1] << '\t' << covariance[2] << '\t' << covariance[3] << '\t' << covariance[4] << "\n\t" << covariance[5] << '\t' << covariance[6] << '\t' << covariance[7] << '\t' << covariance[8] << "\n\t" << covariance[9] << '\t' << covariance[10] << '\t' << covariance[11] << "\n\t" << covariance[12] << '\t' << covariance[13] << "\n\t" << covariance[14] << std::endl;

		best_fit_track.setCovMatrix( covariance );
	}
	else {

		EVENT::FloatVec covariance(15);

		covariance[0] = pow( fit_parameters.Error("omega"), 2 );
		covariance[5] = pow( fit_parameters.Error("phi"), 2 );
		covariance[9] = pow( fit_parameters.Error("d0"), 2 );
		covariance[12] = pow( fit_parameters.Error("z0"), 2 );
		covariance[14] = pow( fit_parameters.Error("tanLambda"), 2 );

		best_fit_track.setCovMatrix( covariance );
	}

	best_fit_track.setChi2( -1. );

	if( _produce_tuples ) {

		_track_result_tuples->fill( _track_result_tuples->findColumn("valid"), minimum.IsValid() );
		_track_result_tuples->fill( _track_result_tuples->findColumn("F_calls"), minimum.NFcn() );
		_track_result_tuples->fill( _track_result_tuples->findColumn("F_minimum"), minimum.Fval() );
		_track_result_tuples->fill( _track_result_tuples->findColumn("sigma0"), minimum.UserParameters().Value("sigma0") );
		_track_result_tuples->fill( _track_result_tuples->findColumn("sigma0_err"), minimum.UserParameters().Error("sigma0") );
		_track_result_tuples->fill( _track_result_tuples->findColumn("edm"), minimum.Edm() );
		_track_result_tuples->addRow();
	}

	const EVENT::TrackerHitVec& data_hits = seed_track->getTrackerHits();

	for(EVENT::TrackerHitVec::const_iterator current_hit = data_hits.begin(); current_hit != data_hits.end(); current_hit++) {

		best_fit_track.addHit( *current_hit );
	}

	return best_fit_track;
}


void TrackFitterLikelihoodProcessor::processEvent(LCEvent *evt) {

	streamlog_out( MESSAGE4 ) << "Processing Event#" << evt->getEventNumber() << std::endl;

	if( _produce_tuples ) {

		_track_result_tuples->fill( _track_result_tuples->findColumn("event_number"), evt->getEventNumber() );
	}

	try {

		EVENT::LCCollection* seed_tracks = evt->getCollection(this->_seed_track_collection_name);

		// get the tpc conditions
		LCCollection* tpc_conditions_collection = evt->getCollection( _tpc_conditions_collection_name );
		tpcconddata::TPCConditions tpc_conditions( tpc_conditions_collection->getElementAt(0) );

		// in sqrt(mm)
		_gas_trans_diffusion_constant = tpc_conditions.getTransDiffusionCoef();

		// in mm
		_gas_trans_defocussing_constant = tpc_conditions.getTransDefocussing();

		// create a collection to hold the best fit tracks
		IMPL::LCCollectionVec* tpc_tracks = new IMPL::LCCollectionVec(EVENT::LCIO::TRACK);

		for(int current_seed_counter = 0; current_seed_counter < seed_tracks->getNumberOfElements(); current_seed_counter++) {

			EVENT::Track* current_seed_track = dynamic_cast<EVENT::Track*>(seed_tracks->getElementAt(current_seed_counter));

			if(! current_seed_track) {
				// skip the track and continue
				continue;
			}

			// FIXME for some reason this seg faults
			// check that the TrackerHits have been flagged as stored in the seed track
			/*IMPL::LCFlagImpl seed_track_flag(_current_seed_track->getType());

			if(!seed_track_flag.bitSet(LCIO::TRBIT_HITS)) {

				throw ParseException("TrackerHits not flagged as being stored with the seed track in collection " + _seed_track_collection_name);
			}*/

			try {
				// now call minuit with this seed track
				tpc_tracks->addElement(new IMPL::TrackImpl(find_most_likely_fit(current_seed_track)));
			}
			catch (Exception &e) {

				streamlog_out( WARNING ) << e.what() << std::endl;
			}
		}

		// now that we have a collection we need to add it to the event
		try {

			evt->addCollection(tpc_tracks, _fitted_track_collection_name);
		}
		catch (EVENT::EventException e) {

			streamlog_out( ERROR ) << e.what() << "Collection " << _fitted_track_collection_name << " is being deleted." << std::endl;

			delete tpc_tracks;
		}

	}
	catch (DataNotAvailableException &e) {

		streamlog_out( ERROR ) << "void TrackFitterLikelihoodProcessor::processEvent( LCEvent * evt ) aborting. " << e.what() << std::endl;
	}
}


double TrackFitterLikelihoodProcessor::operator()(const std::vector<double>& parameters) const {

	// calculate the negative log-likelihood of observing the number of electrons distributed by this track

	// first thing to do is to pre-calculate the expected distribution of electrons on the readout geometry by this track

	IMPL::TrackImpl expected_track;

	// the ordering of the parameters is defined in find_most_likely_fit
	expected_track.setPhi(parameters.at(0));
	expected_track.setOmega(parameters.at(1));
	expected_track.setD0(parameters.at(2));
	expected_track.setTanLambda(parameters.at(3));
	expected_track.setZ0(parameters.at(4));
	const double sigma0 = parameters.at(5);

	streamlog_out( DEBUG4 ) << std::setprecision(8) << "MINUIT Track Parameters: Phi: " << std::setw(8) << parameters.at(0) << "rads, Omega: " << std::setw(8) << parameters.at(1) << "mm^-1, d0: " << std::setw(8) << parameters.at(2) << "mm, tan(Lambda): " << std::setw(8) << parameters.at(3) << " , z0: " << std::setw(8) << parameters.at(4) << "mm, sigma0 " << std::setw(8) << parameters.at(5) << "\n";

	expected_track = calculate_expected_track(expected_track, sigma0);

	// now that we have the expected distribution and the actual distribution (_current_seed_track) we can compare the two to come up with a likelihood

	streamlog_out( DEBUG0 ) << "Calculating the likelihood: \n";

	const double charge_distribution_log_likelihood = calculate_charge_distribution_log_likelihood(expected_track);
	const double primary_ionization_location_log_likelihood = calculate_primary_ionization_location_log_likelihood(expected_track, _sigma_z);
	
	const double log_likelihood = charge_distribution_log_likelihood + primary_ionization_location_log_likelihood;

	streamlog_out( DEBUG4 ) << "Complete negative log likelihood: " << std::setprecision(16) << -1. * log_likelihood << "\n";
	
	// delete the hits and pulses in the expected track (not done automatically)
	const EVENT::TrackerHitVec& tracker_hits = expected_track.getTrackerHits();

	for(EVENT::TrackerHitVec::const_iterator hit_iter = tracker_hits.begin(); hit_iter != tracker_hits.end(); hit_iter++) {
	
		const EVENT::LCObjectVec& tracker_pulses = (*hit_iter)->getRawHits();
	
		for(EVENT::LCObjectVec::const_iterator pulse_iter = tracker_pulses.begin(); pulse_iter != tracker_pulses.end(); pulse_iter++) {
		
			if( *pulse_iter ) {
			
				delete *pulse_iter;
			}
		}
		
		if( *hit_iter ) {
		
			delete *hit_iter;
		}
	}
	
	if( isnan(log_likelihood) ) {
		
		std::cout << "logL is nan, hit enter to continue... \n";
		std::cin.get();
	}
	
	if( streamlog::out.write<DEBUG>() ) {
	
		/*std::cout << "Hit Enter to continue... \n";
		std::cin.get();*/
	}

	return -1. * log_likelihood;
}


double TrackFitterLikelihoodProcessor::Up() const { return 0.5; }


const double TrackFitterLikelihoodProcessor::calculate_charge_distribution_log_likelihood(const EVENT::Track& expected_track) const {

	streamlog_out( DEBUG2 ) << "Calculating the (r, phi) likelihood using the charge distribution\n";
	
	const gear::PadRowLayout2D& pad_layout = Global::GEAR->getTPCParameters().getPadLayout();

	// calculate the log likelihood
	double log_likelihood = 0;

	const EVENT::TrackerHitVec& data_hits = _current_seed_track->getTrackerHits();
	const EVENT::TrackerHitVec& expected_hits = expected_track.getTrackerHits();

	for(EVENT::TrackerHitVec::const_iterator data_hit = data_hits.begin(), expected_hit = expected_hits.begin(); data_hit != data_hits.end(); data_hit++, expected_hit++) {

		// calculate the likelihood for each row
		double log_row_likelihood = 0;

		// retrieve the expected charge for the entire row (hit)
		const double row_expected = (*expected_hit)->getdEdx();

		EVENT::TrackerPulse* first_pulse = dynamic_cast<EVENT::TrackerPulse*>(*((*data_hit)->getRawHits().begin()));
		assert(first_pulse);

		streamlog_out( DEBUG2 ) << "Row #" << pad_layout.getRowNumber(first_pulse->getCellID0()) << ":\n";

		const int expected_pads = (*data_hit)->getType();

		// loop over every pad with a hit in this row calculating the likelihood of seeing the data on the expected value
		// FIXME this might not calculate the number of original electrons correctly at the moment (awaiting more information)
		/* right now it assumes that getTotalCharge() only calculates the number of electrons deposited by the current_seed_track, not any other tracks. The way it is supposed to work is you find the highest peak from _this_ _track_ in this row and create a small time interval around it. To find the number of electrons deposited (or charge) for each pad all you have to do is integrate the charge over this interval (which is the same for each pad in the row). Unfortunately I can't think of a way to determine the highest peak in a row if there are multiple signals per pad... */

		streamlog_out( DEBUG2 ) << "\tRow expected " << row_expected << " primary electrons\n";

		for(EVENT::LCObjectVec::const_iterator expected_pulse = (*expected_hit)->getRawHits().begin(), data_pulse = (*data_hit)->getRawHits().begin(); data_pulse != (*data_hit)->getRawHits().end(); data_pulse++, expected_pulse++) {

			// calculate the probability of an electron from this row being deposited on this pad

			double pad_expected = 0;
			double probability = 0.;

			if(row_expected) {

				pad_expected = (dynamic_cast<EVENT::TrackerPulse*>(*expected_pulse))->getCharge();
				probability = pad_expected/row_expected;
			}

			//std::cout << "probability: " << probability << " number of pads in the row: " << number_pads_in_row << " row# " << pad_layout.getRowNumber((dynamic_cast<EVENT::TrackerPulse*>(*expected_pulse))->getCellID0()) << " pad# " << (dynamic_cast<EVENT::TrackerPulse*>(*expected_pulse))->getCellID0() << std::endl;
			// now factor in some noise
			probability = (probability + _p_noise) / (1 + expected_pads * _p_noise);

			streamlog_out( DEBUG1 ) << "\tCellID #" << (dynamic_cast<EVENT::TrackerPulse*>(*data_pulse))->getCellID0() << ": "
			<< "probability: " << probability
			<< "; pad expected: " << pad_expected
			<< "; pad actual electrons: " << (dynamic_cast<EVENT::TrackerPulse*>(*data_pulse))->getCharge() << ";\n";

			// the log(likelihood) for the row is the summed over each pad in the row
			log_row_likelihood += std::log(probability) * fabs( (dynamic_cast<EVENT::TrackerPulse*>(*data_pulse))->getCharge() );
		}

		streamlog_out( DEBUG2 ) << "\t\trow (hit) log likelihood log(L): " << std::setprecision(10) << log_row_likelihood << "\n";

		// the total log(likelihood) of observing the track is the product of the likelihood of each row
		log_likelihood += log_row_likelihood;
	}
	
	streamlog_out( DEBUG3 ) << "\ttrack (r, phi) log likelihood: " << std::setprecision(10) << log_likelihood << "\n";

	return log_likelihood;
}


IMPL::TrackImpl TrackFitterLikelihoodProcessor::calculate_expected_track(const EVENT::Track& trackEstimate, double sigma0) const {

	streamlog_out( DEBUG3 ) << "Calculating the expected distribution\n";
	
	const gear::PadRowLayout2D& pad_layout = Global::GEAR->getTPCParameters().getPadLayout();

	switch( pad_layout.getPadLayoutType() ) {

		case gear::PadRowLayout2D::POLAR:
			return calculate_expected_track_polar( trackEstimate, sigma0 );

		case gear::PadRowLayout2D::CARTESIAN:
			return calculate_expected_track_cartesian( trackEstimate, sigma0 );

		default: throw Exception("Unsupported geometry type");
	 }
}


IMPL::TrackImpl TrackFitterLikelihoodProcessor::calculate_expected_track_cartesian(const EVENT::Track& trackEstimate, double sigma0) const {

	const gear::PadRowLayout2D& pad_layout = Global::GEAR->getTPCParameters().getPadLayout();

	const double estimated_omega = trackEstimate.getOmega();
	const double estimated_d0 = trackEstimate.getD0();
	const double estimated_phi0 = trackEstimate.getPhi();
	const double estimated_tan_lambda = trackEstimate.getTanLambda();
	const double estimated_z0 = trackEstimate.getZ0();

	// get the extent of the readout electronic plane (xmin, xmax, ymin, ymax)
	const std::vector<double> &plane_extent = pad_layout.getPlaneExtent();

	// get the reference point of the track
	const float* ref_ptr = trackEstimate.getReferencePoint();
	CLHEP::Hep3Vector reference_point(ref_ptr[0], ref_ptr[1], ref_ptr[2]);

	// calculate the point of closest approach to the reference point
	const CLHEP::Hep3Vector n_pca( -1*sin(estimated_phi0), cos(estimated_phi0), 0 );
	const CLHEP::Hep3Vector point_of_closest_approach = reference_point + (estimated_d0 * n_pca);

	// this track represents the way the electrons would be distributed based on the parameters and chosen pads in the input track estimate
	IMPL::TrackImpl expected_track;
	expected_track.setPhi(estimated_phi0);
	expected_track.setOmega(estimated_omega);
	expected_track.setD0(estimated_d0);
	expected_track.setTanLambda(estimated_tan_lambda);
	expected_track.setZ0(estimated_z0);

	// get a list of all the TrackerHits used in the Track
	const EVENT::TrackerHitVec& tracker_hits = _current_seed_track->getTrackerHits();

	for(EVENT::TrackerHitVec::const_iterator hit_iter = tracker_hits.begin(); hit_iter != tracker_hits.end(); hit_iter++) {

		// FIXME modify this in the future to make sure TrackerPulses are used to create this hit
		/*if((*hit_iter).getType() != LCIO::TRACKERPULSE) {
			// the likelihood processor needs the original pulses to calculate the likelihood
			throw(SkipEventException(this));
		}*/

		// the raw objects used to create this hit... it's assumed that they are TrackerPulses
		const EVENT::LCObjectVec& tracker_pulses = (*hit_iter)->getRawHits();

		// get the location of the current hit
		const double* position = (*hit_iter)->getPosition();
		CLHEP::Hep3Vector data_hit_position( position[0], position[1], position[2] );

		// and the mirror which will be used to store what this hit is expected to look like given the track parameters
		IMPL::TrackerHitImpl* expected_hit = new IMPL::TrackerHitImpl();

		// find the row number
		EVENT::TrackerPulse* first_pulse = dynamic_cast<EVENT::TrackerPulse*>(tracker_pulses.at(0));
		assert(first_pulse);

		streamlog_out( DEBUG1 ) << "Row #" << pad_layout.getRowNumber(first_pulse->getCellID0()) << ":\n";

		// find the middle point of this row that the data 'hit' belongs to
		const double row_middle_y = (pad_layout.getPadCenter(first_pulse->getCellID0()))[1];

		// b0 is the x position of the centre of the track in the vertical middle of this row
		double b0 = 0.;

		// phi is the local azimuthal angle of the track in the vertical middle of this row
		double phi = 0.;

		// and use it to calculate the path integral of the primary particle
		double expected_path_integral = 0.;

		// determine if this row was intersected
		bool row_was_not_intersected = false;

		if( estimated_omega == 0 ) {

			// this track is straight, treat it as a straight line
			phi = estimated_phi0;

			const double delta_y = row_middle_y - point_of_closest_approach.y();
			const double delta_x = delta_y / tan(estimated_phi0);

			b0 = point_of_closest_approach.x() + delta_x;

			expected_path_integral = sqrt( (delta_y*delta_y) + (delta_x*delta_x) );
			// to determine the sign of the path integral the direction of the primary must be known and compared to the position of this hit. If the hit is in the opposite direction of the particle momentum then the path integral should be negative (to show that the particle is moving backwards)
			expected_path_integral *= ((data_hit_position.z() - ( estimated_z0 + reference_point.z())*(estimated_tan_lambda)) >= 0)? 1: -1;
		}
		else {

			// calculate the centre of the helix relative to (0,0) in the pad plane
			const double helix_centre_x = reference_point.x() + ((1/estimated_omega) - estimated_d0) * sin(estimated_phi0);
			const double helix_centre_y = reference_point.y() - ((1/estimated_omega) - estimated_d0) * cos(estimated_phi0);

			const double delta_y_squared = (row_middle_y - helix_centre_y) * (row_middle_y - helix_centre_y);
			const double radius_squared = 1/(estimated_omega*estimated_omega);

			if( delta_y_squared > radius_squared ) {

				expected_path_integral = HUGE_VAL;
				row_was_not_intersected = true;
			}
			else {

				const double offset = sqrt( radius_squared - delta_y_squared );
	
				// there could be two intersection points... pick the one that is closer to the 'hit' position
				const double x_candidate_one = helix_centre_x - offset;
				const double x_candidate_two = helix_centre_x + offset;

				const double estimated_inverse_radius = fabs(estimated_omega);
	
				b0 = ( fabs(x_candidate_one - data_hit_position.x()) < fabs(x_candidate_two - data_hit_position.x()) )? x_candidate_one: x_candidate_two;

				const CLHEP::Hep3Vector intersection_point( b0, row_middle_y, 0 );
				const CLHEP::Hep3Vector helix_centre( helix_centre_x, helix_centre_y, 0 );

				phi = (intersection_point - helix_centre).phi();
				streamlog_out( DEBUG1 ) << "\thelix centre: " << helix_centre << "mm, intersection point: " << intersection_point << "mm, phi to hit: " << phi << "rads\n";
				phi = fabs(phi);
				phi = (phi >= M_PI/2)? M_PI - phi: phi;

				const double delta_phi = atan2( point_of_closest_approach.y() - helix_centre_y, point_of_closest_approach.x() - helix_centre_x ) - atan2( row_middle_y - helix_centre_y, b0 - helix_centre_x );

				const double delta_z = data_hit_position.z() - (estimated_z0 + reference_point.z());
				double delta_s =  delta_phi / estimated_inverse_radius;
	
				//FIXME fix this later
// 				const int path_integral_multiplier = (int)( (estimated_inverse_radius)/(2*M_PI) * ((delta_z/estimated_tan_lambda) - delta_s) );
				const int path_integral_multiplier = 0;
				
				if( path_integral_multiplier != 0 ) {
					streamlog_out( DEBUG1 ) << "\tpath_integral_multiplier: " << path_integral_multiplier << ", delta_phi: " << delta_phi << ", delta_z: " << delta_z << ", delta_s: " << delta_s << "\n";
				}

				expected_path_integral = (delta_phi + 2*M_PI*path_integral_multiplier)/estimated_inverse_radius;
				//expected_path_integral = estimated_z0 + reference_point.z() + (estimated_tan_lambda*(delta_phi + 2*M_PI*path_integral_multiplier)/estimated_inverse_radius);
			}
		}

		// now that the row parameters are known use the information to calculate the expected distribution of primary electrons on each pad in the row
		// now calculate, and set, the z position of the hit for use in the primary ionization likelihood function
		double expected_hit_position[3] = {0., 0., (estimated_z0 - reference_point.z()) + expected_path_integral * estimated_tan_lambda};
		expected_hit->setPosition( expected_hit_position );

		streamlog_out( DEBUG1 ) << "\tintersected at: (" << b0 << ", " << row_middle_y << ")mm, path length: " << expected_path_integral << "mm.\n";

		// ensure the original ionization is within the TPC
		if( fabs(expected_hit_position[2]) > fabs(_maximum_drift_length) ) {

			row_was_not_intersected = true;
			streamlog_out( DEBUG1 ) << "\tThe estimated track has left the TPC at this point\n";
		}

		if( (b0 < plane_extent[0]) || (b0 > plane_extent[1]) ) {

			row_was_not_intersected = true;
		}

		if( row_was_not_intersected ) {
			streamlog_out( DEBUG1 ) << "\tThe estimated track is not expected to deposit any charge on this row\n";
			// use the dedx field of the hit to store the row sum of primary electrons expected (not very correct but it works)
			expected_hit->setdEdx( 0 );
			// now as the row (expected_hit) to the track (expected_track)
			expected_track.addHit(expected_hit);
			continue;
		}

		const double electron_drift_distance = fabs(_maximum_drift_length - expected_hit_position[2]);

		const double drift_diffusion = sqrt(electron_drift_distance) * _gas_trans_diffusion_constant;
		const double row_sigma = sqrt(_gas_trans_defocussing_constant*_gas_trans_defocussing_constant + drift_diffusion*drift_diffusion);

		/* loop over every pad that collected data and calculate the expected number of primary electrons it would collect based on these track parameters. Also, keep track of the index number of the pads which were used so when the total number of primary electrons collected by the row is calculated these ones aren't double counted */
		std::map<int, bool> status_by_pad_index;
		double expected_row_sum_of_deposits = 0.;
		int expected_pads = 0;

		for(EVENT::LCObjectVec::const_iterator pulse_iter = tracker_pulses.begin(); pulse_iter != tracker_pulses.end(); pulse_iter++) {

			// I think we can assume these are TrackerPulses...
			EVENT::TrackerPulse* data_pulse = dynamic_cast<EVENT::TrackerPulse*>(*pulse_iter);

			if(!data_pulse) {
				// the data stored with the hits were not pulses so skip this event
				throw(Exception("TrackFitterLikelihoodProcessor requires that TrackerPulses are stored with the TrackerHits."));
			}

			// and the mirror which will be stored in the expected_track
			// the expected number of electrons will be stored / accessed with getCharge() / setCharge()
			IMPL::TrackerPulseImpl* expected_pulse = new IMPL::TrackerPulseImpl();

			// the cell ID stays the same
			expected_pulse->setCellID0(data_pulse->getCellID0());

			streamlog_out( DEBUG1 ) << "\tCellID #" << expected_pulse->getCellID0() << ": ";

			// retrieve the location and size of the current pad (in mm)
			const double pad_x = (pad_layout.getPadCenter(data_pulse->getCellID0()))[0];
			const double pad_y = (pad_layout.getPadCenter(data_pulse->getCellID0()))[1];
			const double pad_width = pad_layout.getPadWidth(data_pulse->getCellID0());

			streamlog_out( DEBUG1 ) << "\tpad x: " << pad_x << "mm; pad y: " << pad_y << "mm;\n";

			// these 2 parameters need to be calculated for this pad
			const double local_b = b0 - pad_x;
			const double local_phi = phi;

			// now calculate the expected charge on this pad based on the trackEstimate

			streamlog_out( DEBUG1 ) << "\tlocal b: " << std::setprecision(6) << local_b << "mm; local phi: " << std::setprecision(6) << local_phi << "rads; local sigma: " << std::setprecision(6) << row_sigma << "mm; height: " << pad_layout.getPadHeight(data_pulse->getCellID0()) << "mm; width: " << pad_width << "mm; ";

			double expected_distribution = calculate_expected_distribution(
				local_b, // b
				local_phi,
				sigma0,
				pad_layout.getPadHeight(data_pulse->getCellID0()),
				pad_width);

			expected_distribution = std::fabs(expected_distribution);
			expected_pulse->setCharge(expected_distribution);

			streamlog_out( DEBUG1 ) << "distribution: " << expected_pulse->getCharge() << "\n";

			// now add the pad (expected_pulse) to the row (expected_hit)
			expected_hit->rawHits().push_back(expected_pulse);

			// and record that the pad from this hit has contributed to the expected row total number of primary electrons collected
			status_by_pad_index[expected_pulse->getCellID0()] = true;
			expected_row_sum_of_deposits += expected_pulse->getCharge();
			expected_pads += 1;
		}

		// determine which pads in the row had primary electrons deposited by this track estimate but did not receive any during the event. They will still contribute to the row expected total
		// the centre pad is the one which probably collected the most primaries in the row
		const int centre_pad = pad_layout.getNearestPad(b0, row_middle_y);

		// move to the left (0) and right (1) summing up the expected primary electron deposits (if they weren't already done)
		for( int direction = 0; direction < 2; direction++ ) {

			int current_pad = centre_pad;

			//FIXME this could potentially run forever ... a terminating condition should be put in
			while( true ) {

				// calculate the expected primary electron deposits
				// retrieve the location and size of the current pad (in mm)
				const double pad_x = (pad_layout.getPadCenter(current_pad))[0];
				const double pad_width = pad_layout.getPadWidth(current_pad);
	
				// these 2 parameters need to be calculated for this pad
				const double local_b = b0 - pad_x;
				const double local_phi = phi;
	
				// now calculate the expected charge on this pad based on the trackEstimate
	
				streamlog_out( DEBUG1 ) << "\tlocal b: " << std::setprecision(6) << local_b << "mm; local phi: " << std::setprecision(6) << local_phi << "rads; local sigma: " << std::setprecision(6) << row_sigma << "mm; height: " << pad_layout.getPadHeight(current_pad) << "mm; width: " << pad_width << "mm; ";

				double expected_distribution = calculate_expected_distribution(
					local_b, // b
					local_phi,
					sigma0,
					pad_layout.getPadHeight(current_pad),
					pad_width);

				streamlog_out( DEBUG1 ) << "distribution: " << expected_distribution << "\n";

				expected_distribution = std::fabs(expected_distribution);

				if( expected_distribution <= 0.000001 ) {

					break;
				}

				if( !status_by_pad_index[current_pad] ) {

					expected_row_sum_of_deposits += expected_distribution;
					status_by_pad_index[current_pad] = true;
				}

				try {
				
					current_pad = (direction == 0)? pad_layout.getLeftNeighbour(current_pad): pad_layout.getRightNeighbour(current_pad);
				}
				catch(...) {

					break;
				}
			}
		}

		// use the dedx field of the hit to store the row sum of primary electrons expected (not very correct but it works)
		expected_hit->setdEdx( expected_row_sum_of_deposits );
		// use the type field of the hit to store the number of pads which were used to make a probablility of 1 (not very correct but it works)
		expected_hit->setType( expected_pads );

		// now as the row (expected_hit) to the track (expected_track)
		expected_track.addHit(expected_hit);
	}

	return expected_track;
}


IMPL::TrackImpl TrackFitterLikelihoodProcessor::calculate_expected_track_polar(const EVENT::Track& trackEstimate, double sigma0) const {

	const gear::PadRowLayout2D& pad_layout = Global::GEAR->getTPCParameters().getPadLayout();

	const double estimated_omega = trackEstimate.getOmega();
	const double estimated_d0 = trackEstimate.getD0();
	const double estimated_phi0 = trackEstimate.getPhi();
	const double estimated_tan_lambda = trackEstimate.getTanLambda();
	const double estimated_z0 = trackEstimate.getZ0();

	const float* ref_ptr = trackEstimate.getReferencePoint();
	CLHEP::Hep3Vector reference_point(ref_ptr[0], ref_ptr[1], ref_ptr[2]);

	// calculate the point of closest approach to the reference point
	const CLHEP::Hep3Vector n_pca( -1*sin(estimated_phi0), cos(estimated_phi0), 0 );
	const CLHEP::Hep3Vector point_of_closest_approach = reference_point + (estimated_d0 * n_pca);

	// this track represents the way the electrons would be distributed based on the parameters and chosen pads in the input track estimate
	IMPL::TrackImpl expected_track;
	expected_track.setPhi(estimated_phi0);
	expected_track.setOmega(estimated_omega);
	expected_track.setD0(estimated_d0);
	expected_track.setTanLambda(estimated_tan_lambda);
	expected_track.setZ0(estimated_z0);

	// get a list of all the TrackerHits used in the Track
	const EVENT::TrackerHitVec& tracker_hits = _current_seed_track->getTrackerHits();

	for(EVENT::TrackerHitVec::const_iterator hit_iter = tracker_hits.begin(); hit_iter != tracker_hits.end(); hit_iter++) {

		// FIXME modify this in the future to make sure TrackerPulses are used to create this hit
		/*if((*hit_iter).getType() != LCIO::TRACKERPULSE) {
			// the likelihood processor needs the original pulses to calculate the likelihood
			throw(SkipEventException(this));
		}*/

		// the raw objects used to create this hit... it's assumed that they are TrackerPulses
		const EVENT::LCObjectVec& tracker_pulses = (*hit_iter)->getRawHits();

		// get the location of the current hit
		const double* position = (*hit_iter)->getPosition();
		CLHEP::Hep3Vector data_hit_position( position[0], position[1], position[2] );

		// and the mirror which will be used to store what this hit is expected to look like
		IMPL::TrackerHitImpl* expected_hit = new IMPL::TrackerHitImpl();

		// find the row number
		EVENT::TrackerPulse* first_pulse = dynamic_cast<EVENT::TrackerPulse*>(tracker_pulses.at(0));

		if(first_pulse) {

			streamlog_out( DEBUG1 ) << "Row #" << pad_layout.getRowNumber(first_pulse->getCellID0()) << ":\n";
		}

		// calculate where this row was intersected by the expected track
		const double row_radius = (pad_layout.getPadCenter(first_pulse->getCellID0()))[0];

		// and use it to calculate the path integral of the primary particle
		double expected_path_integral = 0.;

		// determine if this row was intersected
		bool row_was_not_intersected = false;

		// determine the location where the expected track crosses this row
		double row_intersection_phi = 0.;

		if( estimated_omega == 0 ) {

			if( fabs(estimated_d0) > row_radius ) {

				expected_path_integral = HUGE_VAL;
				row_was_not_intersected = true;
			}
			else {

				// determine the two locations this track could have crossed the row and estimate the best one
				const double reference_phi = asin(estimated_d0 / row_radius);
				const double first_phi = reference_phi + estimated_phi0;
				const double second_phi = M_PI - reference_phi + estimated_phi0;

				// find the phi of this hit
				const double hit_phi = data_hit_position.phi();

				// assume the row was intersected at the point closest to the hit
				row_intersection_phi = ( fabs(hit_phi - first_phi) <= fabs(hit_phi - second_phi) )? first_phi: second_phi;

				// calculate the path_integral (straight distance between the p.c.a. and the intersection point)
				const double pca_r = fabs( estimated_d0 );
				const double pca_phi = asin( estimated_d0 / fabs(estimated_d0) ) + estimated_phi0;

				expected_path_integral = sqrt( pca_r*pca_r + row_radius*row_radius - 2 * row_radius * pca_r * cos(row_intersection_phi - pca_phi) );

				const double delta_z = data_hit_position.z() - (estimated_z0 + reference_point.z());
				// to determine the sign of the path integral the direction of the primary must be known and compared to the position of this hit. If the hit is in the opposite direction of the particle momentum then the path integral should be negative (to show that the particle is moving backwards)
				expected_path_integral *= ( (delta_z)*(estimated_tan_lambda) >= 0 )? 1: -1;
			}
		}
		else {

			// calculate the centre of the helix
			const double helix_centre_x = reference_point.x() + ((1/estimated_omega) - estimated_d0) * sin(estimated_phi0);
			const double helix_centre_y = reference_point.y() - ((1/estimated_omega) - estimated_d0) * cos(estimated_phi0);
			const CLHEP::Hep3Vector helix_centre( helix_centre_x, helix_centre_y, reference_point.z() );

			// calculate the angle of the intersection point of the track and this pads radius (using the law of cosines)
			const double a = row_radius;
			const double b = (helix_centre - reference_point).mag();
			const double c = fabs( 1 / (estimated_omega) );

			const double acos_parameter = ((a*a + b*b) - (c*c)) / (2. * a * b);

			if( (acos_parameter <= -1) || (acos_parameter >= 1) ) {

				streamlog_out( DEBUG1 ) << "\thelix centre: " << helix_centre << "mm, acos param: " << acos_parameter << ", a: " << a << ", b: " << b << ", c: " << c << "\n";
				expected_path_integral = HUGE_VAL;
				row_was_not_intersected = true;
			}
			else {

				const double hit_delta_phi = acos( acos_parameter );

				// find the phi of this hit
				const double hit_phi = data_hit_position.phi();

				const double estimated_inverse_radius = fabs( estimated_omega );

				const double helix_centre_phi = fmod(helix_centre.phi() + 2*M_PI, 2*M_PI);

				// find the closest of the two intersection points, the phi is relative to the centre of the geometry (0, 0)
				row_intersection_phi = ( fabs(helix_centre_phi - hit_delta_phi - hit_phi) < fabs(helix_centre_phi + hit_delta_phi - hit_phi) )? helix_centre_phi - hit_delta_phi: helix_centre_phi + hit_delta_phi;

				streamlog_out( DEBUG1 ) << "\thelix centre phi: " << helix_centre_phi << "rads, phi1: " << helix_centre_phi - hit_delta_phi << "rads, phi2: " << helix_centre_phi + hit_delta_phi << "rads.\n";

				const CLHEP::Hep3Vector intersection_location( row_radius*cos(row_intersection_phi), row_radius*sin(row_intersection_phi), 0 );

				const double delta_phi = ((point_of_closest_approach - helix_centre).phi() - (intersection_location - helix_centre).phi()) * (estimated_omega / fabs(estimated_omega));
				const double delta_z = data_hit_position.z() - (estimated_z0 + reference_point.z());
				double delta_s =  delta_phi / estimated_inverse_radius;

				const int path_integral_multiplier = (int)( (estimated_inverse_radius)/(2*M_PI) * ((delta_z/estimated_tan_lambda) - delta_s) );
				streamlog_out( DEBUG1 ) << "\tpath_integral_multiplier: " << path_integral_multiplier << ", delta_phi: " << delta_phi << ", delta_z: " << delta_z << ", delta_s: " << delta_s << "\n";

				expected_path_integral = (delta_phi + 2*M_PI*path_integral_multiplier)/estimated_inverse_radius;
			}
		}

		// now calculate, and set, the z position of the hit for use in the primary ionization likelihood function
		double expected_hit_position[3] = {0., 0., (estimated_z0 - reference_point.z()) + expected_path_integral * estimated_tan_lambda};
		expected_hit->setPosition( expected_hit_position );

		streamlog_out( DEBUG1 ) << "\tintersected at: " << row_radius << "mm, " << row_intersection_phi << "rads, path length: " << expected_path_integral << "mm.\n";

		// ensure the original ionization is within the TPC
		if( fabs(expected_hit_position[2]) > fabs(_maximum_drift_length) ) {

			row_was_not_intersected = true;
			streamlog_out( DEBUG1 ) << "\tThe estimated track has left the TPC at this point\n";
		}

		if( row_was_not_intersected ) {
			streamlog_out( DEBUG1 ) << "\tThe estimated track is not expected to deposit any charge on this row\n";
			// use the dedx field of the hit to store the row sum of primary electrons expected (not very correct but it works)
			expected_hit->setdEdx( 0 );
			// now as the row (expected_hit) to the track (expected_track)
			expected_track.addHit(expected_hit);
			continue;
		}
		
		const double electron_drift_distance = fabs(_maximum_drift_length - expected_hit_position[2]);

		const double drift_diffusion = sqrt(electron_drift_distance) * _gas_trans_diffusion_constant;
		const double row_sigma = sqrt(_gas_trans_defocussing_constant*_gas_trans_defocussing_constant + drift_diffusion*drift_diffusion);

		/* loop over every pad that collected data and calculate the expected number of primary electrons it would collect based on these track parameters. Also, keep track of the index number of the pads which were used so when the total number of primary electrons collected by the row is calculated these ones aren't double counted */
		std::map<int, bool> status_by_pad_index;
		double expected_row_sum_of_deposits = 0.;
		int expected_pads = 0;

		// loop over every pad in this hit (and a few neighbours) calculating the expected charge based on the parameters
		for(EVENT::LCObjectVec::const_iterator pulse_iter = tracker_pulses.begin(); pulse_iter != tracker_pulses.end(); pulse_iter++) {

			// I think we can assume these are TrackerPulses...
			EVENT::TrackerPulse* data_pulse = dynamic_cast<EVENT::TrackerPulse*>(*pulse_iter);

			if(!data_pulse) {
				// the data stored with the hits were not pulses so skip this event
				throw(Exception("TrackFitterLikelihoodProcessor requires that TrackerPulses are stored with the TrackerHits."));
			}

			// and the mirror which will be stored in the expected_track
			// the expected number of electrons will be stored / accessed with getCharge() / setCharge()
			IMPL::TrackerPulseImpl* expected_pulse = new IMPL::TrackerPulseImpl();

			// the cell ID stays the same
			expected_pulse->setCellID0(data_pulse->getCellID0());

			streamlog_out( DEBUG1 ) << "\tCellID #" << expected_pulse->getCellID0() << ": ";

			// retrieve the location and size of the current pad (in mm and rads)
			const double pad_radius = (pad_layout.getPadCenter(data_pulse->getCellID0()))[0];
			const double pad_phi = (pad_layout.getPadCenter(data_pulse->getCellID0()))[1];
			const double pad_width = pad_layout.getPadWidth(data_pulse->getCellID0()) * pad_radius;

			streamlog_out( DEBUG1 ) << "\tpad phi: " << pad_phi << "rads; pad radius: " << pad_radius << "mm;\n";

			// these 2 parameters need to be calculated for this pad
			double local_b = 0.;
			double local_phi = 0.;

			// check if the track is straight or curved
			if( estimated_omega == 0 ) {

				// b has been found
				local_b = (pad_phi - row_intersection_phi) * pad_radius;

				// local_phi is found
				local_phi = fmod( acos( sin(pad_phi)*sin(row_intersection_phi) + cos(pad_phi)*cos(row_intersection_phi) ), M_PI);

			} else {

				// calculate the centre of the helix
				const double helix_centre_x = reference_point.x() + ((1/estimated_omega) - estimated_d0) * sin(estimated_phi0);
				const double helix_centre_y = reference_point.y() - ((1/estimated_omega) - estimated_d0) * cos(estimated_phi0);
				const CLHEP::Hep3Vector helix_centre( helix_centre_x, helix_centre_y, reference_point.z() );

				// calculate the angle of the intersection point of the track and this pads radius (using the law of cosines)
				const double a = pad_radius;
				const double b = (helix_centre - reference_point).mag();
				const double c = fabs( 1 / (estimated_omega) );

				const double cos_parameter = (c*c - (a*a + b*b)) / (-2. * a * b);

				const double delta_phi = acos( cos_parameter );

				// b has been found
				local_b = (pad_phi - row_intersection_phi) * pad_radius;

				// estimate phi
				local_phi = (M_PI / 2) - fmod( asin( b * sin(delta_phi) / c ), M_PI / 2);
			}

			// now calculate the expected charge on this pad based on the trackEstimate

			streamlog_out( DEBUG1 ) << "\tlocal b: " << std::setprecision(6) << local_b << "mm; local phi: " << std::setprecision(6) << local_phi << "rads; local sigma: " << std::setprecision(6) << row_sigma << "mm; height: " << pad_layout.getPadHeight(data_pulse->getCellID0()) << "mm; width: " << pad_width << "mm; ";

			double expected_distribution = calculate_expected_distribution(
				local_b, // b
				local_phi,
				sigma0,
				pad_layout.getPadHeight(data_pulse->getCellID0()),
				pad_width);

			expected_distribution = std::fabs(expected_distribution);
			expected_pulse->setCharge(expected_distribution);

			streamlog_out( DEBUG1 ) << "distribution: " << expected_pulse->getCharge() << "\n";

			// now add the pad (expected_pulse) to the row (expected_hit)
			expected_hit->rawHits().push_back(expected_pulse);

			// and record that the pad from this hit has contributed to the expected row total number of primary electrons collected
			status_by_pad_index[expected_pulse->getCellID0()] = true;
			expected_row_sum_of_deposits += expected_pulse->getCharge();
			expected_pads += 1;
		}

		//FIXME add teh rest of the pads in the row
		//// determine which pads in the row had primary electrons deposited by this track estimate but did not receive any during the event. They will still contribute to the row expected total
		// the centre pad is the one which probably collected the most primaries in the row
		const int centre_pad = pad_layout.getNearestPad(row_radius, row_intersection_phi);

		// move to the left (0) and right (1) summing up the expected primary electron deposits (if they weren't already done)
		for( int direction = 0; direction < 2; direction++ ) {

			int current_pad = centre_pad;

			while( true ) {

				// calculate the expected primary electron deposits
				// retrieve the location and size of the current pad (in mm and rads)
				const double pad_radius = (pad_layout.getPadCenter(current_pad))[0];
				const double pad_phi = (pad_layout.getPadCenter(current_pad))[1];
				const double pad_width = pad_layout.getPadWidth(current_pad) * pad_radius;
	
				// these 2 parameters need to be calculated for this pad
				double local_b = 0.;
				double local_phi = 0.;
				
				// check if the track is straight or curved
				if( estimated_omega == 0 ) {
	
					// b has been found
					local_b = (pad_phi - row_intersection_phi) * pad_radius;
	
					// local_phi is found
					local_phi = fmod( acos( sin(pad_phi)*sin(row_intersection_phi) + cos(pad_phi)*cos(row_intersection_phi) ), M_PI);
	
				} else {
	
					// calculate the centre of the helix
					const double helix_centre_x = reference_point.x() + ((1/estimated_omega) - estimated_d0) * sin(estimated_phi0);
					const double helix_centre_y = reference_point.y() - ((1/estimated_omega) - estimated_d0) * cos(estimated_phi0);
					const CLHEP::Hep3Vector helix_centre( helix_centre_x, helix_centre_y, reference_point.z() );
	
					// calculate the angle of the intersection point of the track and this pads radius (using the law of cosines)
					const double a = pad_radius;
					const double b = (helix_centre - reference_point).mag();
					const double c = fabs( 1 / (estimated_omega) );
	
					const double cos_parameter = (c*c - (a*a + b*b)) / (-2. * a * b);
	
					const double delta_phi = acos( cos_parameter );
	
					// b has been found
					local_b = (pad_phi - row_intersection_phi) * pad_radius;
	
					// estimate phi
					local_phi = (M_PI / 2) - fmod( asin( b * sin(delta_phi) / c ), M_PI / 2);
				}
	
				// now calculate the expected charge on this pad based on the trackEstimate
	
				streamlog_out( DEBUG1 ) << "\tlocal b: " << std::setprecision(6) << local_b << "mm; local phi: " << std::setprecision(6) << local_phi << "rads; local sigma: " << std::setprecision(6) << row_sigma << "mm; height: " << pad_layout.getPadHeight(current_pad) << "mm; width: " << pad_width << "mm; ";

				double expected_distribution = calculate_expected_distribution(
					local_b, // b
					local_phi,
					sigma0,
					pad_layout.getPadHeight(current_pad),
					pad_width);

				streamlog_out( DEBUG1 ) << "distribution: " << expected_distribution << "\n";

				expected_distribution = std::fabs(expected_distribution);

				if( expected_distribution <= 0.00001 ) {
					
					break;
				}
				
				if( !status_by_pad_index[current_pad] ) {
				
					expected_row_sum_of_deposits += expected_distribution;
					status_by_pad_index[current_pad] = true;
				}

				try {
				
					current_pad = (direction == 0)? pad_layout.getLeftNeighbour(current_pad): pad_layout.getRightNeighbour(current_pad);
				}
				catch(...) {

					break;
				}
			}
		}


		// use the dedx field of the hit to store the row sum of primary electrons expected (not very correct but it works)
		expected_hit->setdEdx( expected_row_sum_of_deposits );
		// use the type field of the hit to store the number of pads which were used to make a probablility of 1 (not very correct but it works)
		expected_hit->setType( expected_pads );
		// now as the row (expected_hit) to the track (expected_track)
		expected_track.addHit(expected_hit);
	}

	return expected_track;
}

// a helper function for calculate_expected_distribution
double xi(double upsilon, double sigma) {

	const double exp_param = std::max( -100., (-1 * upsilon*upsilon) / (2 * sigma*sigma) );

	return ((upsilon / 2) * erf(upsilon * M_SQRT1_2 / sigma)) + ((sigma / sqrt(2 * M_PI)) * exp(exp_param));
}

// a helper function for calculate_expected_distribution
double eta_old(double b, double phi, double sigma, double height, double width) {

	return xi( ((b + (width / 2)) * cos(phi)) + ((height / 2) * sin(phi)), sigma);
}

// a helper function for calculate_expected_distribution
double eta( const double a, const double del, const double sigma, const double phi ) {

	return xi( a + del, sigma);
}

// a helper function for calculate_expected_distribution
double lambda( const double a, const double del, const double sigma, const double phi ) {

	const double sin_phi = sin(phi);
	const double cos_phi = 1 - (sin_phi*sin_phi);
	const double sinphi_times_cosphi = sin_phi * cos_phi;

	return (eta(a, del, sigma, phi) - eta(a, -1*del, sigma, phi)) / sinphi_times_cosphi;
}

// a helper function for calculate_expected_distribution
double lambda_taylor( const double a, const double height, const double sigma, const double phi ) {

	const double sin_phi = sin(phi);
	const double sin_phi2 = sin_phi*sin_phi;
	const double cos_phi = 1 - sin_phi2;

	// pre-calculate some powers
	const double a2 = a*a;

	const double h2 = height*height;

	const double s2 = sigma*sigma;
	const double s3 = sigma*s2;
	const double s4 = sigma*s3;

	const double erf_term = erf( a * M_SQRT1_2 / sigma );
	const double exp_term = exp( -1 * (a2) / (2*s2) );

	// calculate the first term
	const double term_1 = 0.5 * height / cos_phi;

	// calculate the second term
	const double term_2_3 = 1 - ( h2 * (3*s2 - a2) * sin_phi2 / (80 * s4) );

	const double term_2_2_1 = M_SQRT1_2 * M_2_SQRTPI / 24.;	// dividing by 24 because the math constant is 2/sqrt(pi)
	const double term_2_2_2 = h2 * a / s3;
	const double term_2_2 = term_2_2_1 * term_2_2_2 * sin_phi2 * exp_term;

	const double term_2 = erf_term - (term_2_2 * term_2_3);

	return term_1 * term_2;
}

double TrackFitterLikelihoodProcessor::calculate_expected_distribution( const double b, const double phi, const double sigma,
	const double height, const double width) const {

	const double sin_phi = sin(phi);
	const double cos_phi = 1 - (sin_phi*sin_phi);

	const double a_plus = (b + (0.5*width)) * cos_phi;
	const double a_minus = (b - (0.5*width)) * cos_phi;
	const double del = (0.5*height) * sin_phi;
	
	// FIXME: debug when sinphi ~ the crossover value print them out
	// check to see if the integral needs to be Taylor expanded
	if( fabs(sin_phi) < (sigma/(6*height)) ) {

		return lambda_taylor(a_plus, height, sigma, phi) - lambda_taylor(a_minus, height, sigma, phi);
	}
	else {
		
		return lambda(a_plus, del, sigma, phi) - lambda(a_minus, del, sigma, phi);
	}
}


const double TrackFitterLikelihoodProcessor::calculate_primary_ionization_location_log_likelihood(const Track& track_estimate, const double sigma_z) const {

	streamlog_out( DEBUG2 ) << "Calculating the (r, z) likelihood:\n";
	
	//const gear::PadRowLayout2D& pad_layout = Global::GEAR->getTPCParameters().getPadLayout();

	// calculate the log likelihood
	double log_likelihood = 0;

	const EVENT::TrackerHitVec& data_hits = _current_seed_track->getTrackerHits();
	const EVENT::TrackerHitVec& expected_hits = track_estimate.getTrackerHits();

	// loop through the hits
	for(EVENT::TrackerHitVec::const_iterator data_hit = data_hits.begin(), expected_hit = expected_hits.begin(); data_hit != data_hits.end(); data_hit++, expected_hit++) {

		EVENT::TrackerPulse* first_pulse = dynamic_cast<EVENT::TrackerPulse*>(*((*data_hit)->getRawHits().begin()));
		assert(first_pulse);

		// get the position of this hit
		const double actual_mean_position = (*data_hit)->getPosition()[2];

		// and get the position of the expected hit
		const double expected_position = (*expected_hit)->getPosition()[2];

		const double displacement = (isinf(expected_position) || isnan(expected_position)) ? (_maximum_drift_length): (actual_mean_position - expected_position);

		streamlog_out( DEBUG1 ) << "\tmean z: " << actual_mean_position << "mm, expected z: " << expected_position << "mm, displacement: " << displacement << "mm.\n";

		const double log_position_likelihood = -1 * ( displacement*displacement ) / ( 2 * sigma_z * sigma_z );

		streamlog_out( DEBUG2 ) << "\t\trow (hit) log likelihood log(L): " << std::setprecision(10) << log_position_likelihood << "\n";

		// the total log(likelihood) of observing the track is the product of the likelihood of each row
		log_likelihood += log_position_likelihood;
	}
	
	streamlog_out( DEBUG3 ) << "\ttrack (r, z) log likelihood: " << std::setprecision(10) << log_likelihood << "\n";
	
	return log_likelihood;
}


Processor* TrackFitterLikelihoodProcessor::newProcessor() {

	return new TrackFitterLikelihoodProcessor();
}


void TrackFitterLikelihoodProcessor::check(LCEvent *evt) {


}


void TrackFitterLikelihoodProcessor::end() {

}

} // namespace marlintpc

