#include "TrackFitterSimpleChiSquarePads.h"
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <CLHEP/Vector/ThreeVector.h>

#include <EVENT/LCRunHeader.h>

#include <streamlog/streamlog.h>

// global constants from Marlin, used for the global pointer to the GearMgr
#include <marlin/Global.h>

#include "IntersectionCalculator.h"
#include "TrackFitterFactory.h"

//#define CHI_SQUARE_PADS_TYPE_STRING "SimpleChiSquarePads"

namespace marlintpc{

  TrackFitterSimpleChiSquarePads * TrackFitterSimpleChiSquarePads::_theSimpleChiSquarePadsFitter = 0;

  TrackFitterSimpleChiSquarePads::TrackFitterSimpleChiSquarePads(gear::TPCParameters const *gearTPCParameters,
						     double transDefocussing, double longDefocussing,
						     double transDiffusion, double longDiffusion)
    : TrackFitterSimpleChiSquare(transDefocussing, longDefocussing, transDiffusion, longDiffusion)
    {
	initialiseGEARGeometry(gearTPCParameters);
    }

  // This block of code has been copied from TrackFitterSimpleChiSquare 
  // I have no idea how to avoid this code duplication because the function 
  // is implementation specific and static. 
    TrackFitterBase * TrackFitterSimpleChiSquarePads::getInstance(LCParameters const * parameters)
    {
      // try to read the fitter parameters from the LCParameters
      EVENT::FloatVec transDefocusingVals;
      parameters->getFloatVals("SimpleChiSquare_TransDefocussing",transDefocusingVals);
      
      float transDefocussing;
      if ( !transDefocusingVals.empty() )
      {
	transDefocussing = transDefocusingVals[0];
      }
      else
	transDefocussing=1.;

      EVENT::FloatVec longDefocusingVals;
      parameters->getFloatVals("SimpleChiSquare_LongDefocussing",longDefocusingVals);

      double longDefocussing;
      if ( !longDefocusingVals.empty() )
      {
	longDefocussing = longDefocusingVals[0];
      }
      else
	longDefocussing=1.;

      // for the diffusion we can accept the default of 0.0 if parameter not found
      double transDiffusion = parameters->getFloatVal("SimpleChiSquare_TransDiffusionCoef");
      double longDiffusion  = parameters->getFloatVal("SimpleChiSquare_LongDiffusionCoef");

      // now the GEAR parameters. Use the global GearMgr defined in Marlin
      gear::TPCParameters const & theTPCParameters = marlin::Global::GEAR->getTPCParameters();

      if (_theSimpleChiSquarePadsFitter )
      {
	// reinitialise the fitter
	new( _theSimpleChiSquarePadsFitter ) TrackFitterSimpleChiSquarePads(& theTPCParameters,
								transDefocussing, longDefocussing,
								transDiffusion, longDiffusion);
      }
      else
      {
	_theSimpleChiSquarePadsFitter = new TrackFitterSimpleChiSquarePads(& theTPCParameters,
							       transDefocussing, longDefocussing,
							       transDiffusion, longDiffusion);
      }

      return _theSimpleChiSquarePadsFitter;
    }

  
//    TrackFitterSimpleChiSquarePads::TrackFitterSimpleChiSquarePads(EVENT::LCRunHeader const *rh,
//				     gear::TPCParameters const *gearTPCParameters)
//	: TrackFitterSimpleChiSquare(rh)
//    {
//	initialiseGEARGeometry(gearTPCParameters);
//    }

     EVENT::DoubleVec  TrackFitterSimpleChiSquarePads::calculateResiduals(EVENT::Track *testTrack,
								    unsigned int testHitNumber,
								    EVENT::Track *referenceTrack)
     {
	// get hit testHitNumber from track
	EVENT::TrackerHitVec const &hitVector = testTrack->getTrackerHits();

	if ( testHitNumber >= hitVector.size() )
	{
	    streamlog_out(ERROR1) << "Cannot calulate residual, not hit number " 
				  << testHitNumber <<" in track."<< std::endl;
	    throw EVENT::DataNotAvailableException("Could not find hit in track.");
	}
	
	EVENT::TrackerHit *hit = hitVector[testHitNumber];

	// now simply call calculateResidual(testHit, trackWithoutTestHit),
	// functionality is the same for this basic implementation

	if (referenceTrack) // residual wrt. referenceTrack has been requested
	    return calculateResiduals( hit, referenceTrack );

	else // return residual wrt. track which contains the hit
	    return calculateResiduals( hit, testTrack );
	
    }

    EVENT::DoubleVec TrackFitterSimpleChiSquarePads::calculateResiduals(EVENT::TrackerHit *testHit,
								  EVENT::Track *trackWithoutTestHit)
    {
	// just for convenience: alias for trackWithoutTestHit
	EVENT::Track *track = trackWithoutTestHit;

	// Description of the algorithm:
	// - Calculate the centre of the pad row ( y or r ). Do this by determining the
	//   nearest pad and getting its centre.
	// - Calculate the intersection of the track (helix or straight line) with
	//   the pad row centre (straight line or circle)
	// - Calculate distance of measured coordinate and intersection point along the pad row (x or r*pi)

	// As the implementations have different syntax for polar and cartesian coordinates (or other pad layouts to come)
	// the code is separated in two blocks rather than introducing a switch statement every time the layout differ.

	// But first calculate some track parameters from track:

	// unit vector from reference point to point of closest approach (pca)
	CLHEP::Hep3Vector n_pca( -sin( track->getPhi() ), cos ( track->getPhi() ) );

	// reference point in xy projection
	CLHEP::Hep3Vector reference_point( track->getReferencePoint ()[0], 
					   track->getReferencePoint ()[1] );

	// point of closest approach
	CLHEP::Hep3Vector pca =  reference_point +  track->getD0() *  n_pca ;

	// centre of the helix projection
	// (n_pca is unit vercot pointing outwards)
	CLHEP::Hep3Vector helix_centre; // is initialised with 0., 0. by default
	// can only be calculated if curcature != 0 
	// leave it at the origin otherwise
	if (  track->getOmega() != 0. ) 
	    helix_centre = pca - 1./ fabs( track->getOmega() ) * n_pca ;
	streamlog_out(DEBUG2) <<"helix_centre" << helix_centre << std::endl;
	
	EVENT::DoubleVec retVal; // the return value, a DoubleVec

	// Now the two blocks for the different pad layouts
	switch ( _padLayoutType )
	{
	    case gear::PadRowLayout2D::CARTESIAN:
	    {
		// Step one: calculate row centre, in y in this case

		    // get xy position of hit
		    double const *position = testHit->getPosition();

		    //get pad index of pad nearest to hit
		    int nearestPadIndex = _padLayout->getNearestPad( position[0] , position[1] );

		    //get the centre y coordinate of the hit
		    double rowCentre = _padLayout->getPadCenter(nearestPadIndex)[1];
		    streamlog_out(DEBUG2) << "nearest pad row has y = "<<rowCentre << std::endl;

		// Step two: calculate intersection of track with horizontal straight line
		    
		    double intesectionX ;
		    if ( trackWithoutTestHit->getOmega() == 0 ) // track is a straight line
		    {
			// A horizontal line does not intersect with another horizontal line
			if (  track->getPhi() == 0. ||  track->getPhi() == M_PI ) 
			    throw  IntersectionCalculator::NoIntersectionException("Track is along x direction");

			// inclination wrt. to y axis is cotan (phi)
			double inclination = 1./ tan( track->getPhi() );
			intesectionX = pca[0] + (rowCentre - pca[1]) * inclination;
		    }
		    else // track is a helix, two possible intersections
		    {
		        std::pair<double,double> intersections =  
			    IntersectionCalculator::horizonalLine_circle( rowCentre ,
									  helix_centre[0],  helix_centre[1],
									  1./fabs(  track->getOmega() ) );
			
			streamlog_out(DEBUG2) << "Possible intersection: x1 = "<< intersections.first 
					      << " ; x2 = " << intersections.second << std::endl;
		    
			// chose the intersection which is closer to the hit position
			intesectionX = ( fabs(intersections.first - position[0]) 
					 < fabs(intersections.second - position[0])
					 ? intersections.first : intersections.second );
		    }
		    
		// Step tree: calculate distance between intersection point and hit
		    retVal.push_back( fabs( intesectionX - position[0] ));

	    } // end of block for cartesian pad layout
	    break;
	    case gear::PadRowLayout2D::POLAR:
	    {
		// Step one: calculate row centre, in r in this case

		    // convert hit position from x/y to r/phi
		    CLHEP::Hep3Vector pos3D( testHit->getPosition()[0] , testHit->getPosition()[1] );
		    double position[2] = { pos3D.mag() , pos3D.phi() };

		    //get pad index of pad nearest to hit
		    int nearestPadIndex = _padLayout->getNearestPad( position[0] , position[1] );

		    //get the centre r coordinate of the hit
		    double rowCentre = _padLayout->getPadCenter(nearestPadIndex)[0];

		// Step two: calculate intersection of track with circle around origin with radius r
		    
		    std::pair<double,double>  intersectionsPhi; // there are always two intersectios, if any
			//( in case the circles are touching both angles are the same, which dows not affect 
			//  the calculated residual ) 
		    
		    if ( trackWithoutTestHit->getOmega() == 0 ) // track is a straight line
		    {
			// handle horizontal line separately, cannot be described as a function of y
			if (  track->getPhi() == 0. ||  track->getPhi() == M_PI ) 
			{
			    // caclulate intersections of horizontal line (in this case the track) with circle
			    // (in this case the pad row, so the centre is 0,0)
			    std::pair<double,double> intersections =  
				IntersectionCalculator::horizonalLine_circle( pca[1] , // pca is always on track
									      0., 0.,
									      rowCentre);
			    CLHEP::Hep3Vector intersextionXY1(intersections.first, pca[1] );
			    CLHEP::Hep3Vector intersextionXY2(intersections.second, pca[1] );
			    
			    intersectionsPhi.first  =  intersextionXY1.phi();
			    intersectionsPhi.second =  intersextionXY2.phi();
			}
			else // track is not horizontal
			{   // calculate straight line as a function of y
			    // inclination wrt. to y axis is cotan (phi)
			    double inclination = 1./ tan( track->getPhi() );
			    double xAxisIntersept  = pca[0] - pca[1] * inclination;
			    
			    intersectionsPhi = IntersectionCalculator::circle_straightLine_phi( rowCentre, // radius of the pad row centre
												inclination,
												xAxisIntersept);
			}
		    }
		    else // track is a helix
		    {
		        intersectionsPhi =  
			    IntersectionCalculator::circle_circle( rowCentre, // radius of the pad row centre
								   helix_centre[0], helix_centre[1],
								   1./fabs( track->getOmega() ) );
		    }

		    // the correct angle is the one with the smaller angular distance to the hit
		    double intersectionPhi = ( ( distanceOnCircle( intersectionsPhi.first , position[1] ) 
					  < distanceOnCircle( intersectionsPhi.second , position[1] ) )
					? intersectionsPhi.first : intersectionsPhi.second );

		// Step tree: calculate distance on the pad row between intersection point and hit 
		    // it is deltaPhi*R
		    retVal.push_back( distanceOnCircle( intersectionPhi , position[1] )/ rowCentre );

	    } // end of block for polar pad layout
	    break;
	    default:
		throw gear::Exception( "Unknown pad layout type" );
	};

	retVal.push_back(0);
  
	return retVal;

    }// calculateResidual( testHit, trackWithoutTestHit )

    std::string TrackFitterSimpleChiSquarePads::getRevision()
    {
	return std::string( "$Rev: 1721 $" );
    }

    unsigned char TrackFitterSimpleChiSquarePads::getFitterType() const
    {
        return TrackFitterBase::FitterTypes::SIMPLECHISQUAREPADS;      
    }

    IMPL::TrackImpl * TrackFitterSimpleChiSquarePads::fitTrack(EVENT::Track const * seedTrack)
    {
      // use the implementation from TrackFitterSimpleChiSquare
      IMPL::TrackImpl *fittedTrack = TrackFitterSimpleChiSquare::fitTrack(seedTrack);

      // now set the correct fitter type
      setTrackFitterType( this->getFitterType(), fittedTrack );

      return fittedTrack;
    }

    void  TrackFitterSimpleChiSquarePads::initialiseGEARGeometry(gear::TPCParameters const *gearTPCParameters)
    {
	_TPCParameters=gearTPCParameters;
	_padLayout = &(_TPCParameters->getPadLayout());
	_padLayoutType = _padLayout->getPadLayoutType();
    }

    // calculate angular distance on a circle, max angle can be pi
    double TrackFitterSimpleChiSquarePads::distanceOnCircle(double phi1, double phi2)
	{
	    double deltaPhi=std::fmod( std::fabs(phi1-phi2) , 2*M_PI); 
	    if ( deltaPhi > M_PI )
		deltaPhi = 2 * M_PI - deltaPhi;
	    
	    return deltaPhi;
	}
    
}//namespace marlintpc
