//// LCIO
#include "EVENT/LCIO.h"
#include "EVENT/LCCollection.h"
#include "EVENT/TrackerHit.h"
#include <IMPL/LCCollectionVec.h>
#include "IMPL/TrackImpl.h"
#include <IMPL/LCFlagImpl.h>

#include <iostream>
#include <iomanip>
#include <cmath>
#include <algorithm>

// Marlin
#include "marlin/Global.h"
#include "TrackFitterSimpleChiSquareProcessor.h"
#include "TrackFitterBase.h"

// Aida
#ifdef MARLIN_USE_AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>

//marlin
#include <marlin/ConditionsProcessor.h>


#endif

using namespace marlin;
using namespace lcio;

namespace marlintpc
{

TrackFitterSimpleChiSquareProcessor aTrackFitterSimpleChiSquareProcessor;

TrackFitterSimpleChiSquareProcessor::TrackFitterSimpleChiSquareProcessor() 
  : Processor("TrackFitterSimpleChiSquareProcessor"),
    _trackCollectionParameters(0), _nGoodFits(0), _nBadFits(0)
{

	_description = "Calculates seed track parameters from  in the TrackerHits in the track candidates collection!";

	registerInputCollection(     LCIO::TRACK,
				    "InputSeedTracks",
				    "The the name of the input collection of track candidates (default: TPCSeedTracks)",
				    _input_seed_tracks_collection_name,
				    std::string("TPCSeedTracks"));

	registerOutputCollection(    LCIO::TRACK,
				    "OutputTracks",
				    "The name of the output collection with the fitted tracks(default: TPCTracks)",
				    _output_tracks_collection_name,
				    std::string("TPCTracks"));

	registerProcessorParameter( "SetOutputTransient" , 
				    "if not 0 the output hits collection is set transient (default: 0)",
				    _outputIsTransient,
				    int(0));

	registerProcessorParameter( "TransDefocussing",
				    "The transverse defocussing in the readout structure sigma_0,trans(in mm) (default: 1)",
				    _transDefocussing,
				    float(1.));
  
	registerProcessorParameter( "TransDiffusionCoef",
				    "The transverse diffusion coefficient (in mm/sqrt(mm) ) (default: 0)",
				    _transDiffusionCoef,
				    float(0.));

	registerProcessorParameter( "LongDefocussing",
				    "The longitudinal defocussing in the readout structure sigma_0,long (in mm) (default: 1)",
				    _longDefocussing,
				    float(1.));

	registerProcessorParameter( "LongDiffusionCoef",
				    "The longitudinal diffusion coefficient (in mm/sqrt(mm) ) (default: 0)",
				    _longDiffusionCoef,
				    float(0.));

	registerOptionalParameter( "D0Start",
					    "Use this start value for D0 instead of the value from the seed track",
					    _d0Start,
					    float(0.));

	registerOptionalParameter( "PhiStart",
					    "Use this start value for phi instead of the value from the seed track",
					    _phiStart,
					    float(0.));

	registerOptionalParameter( "OmegaStart",
					    "Use this start value for omega instead of the value from the seed track",
					    _omegaStart,
					    float(0.));

	registerOptionalParameter( "TanLambdaStart",
					    "Use this start value for tanLambda instead of the value from the seed track",
					    _tanLambdaStart,
					    float(0.));

	registerOptionalParameter( "Z0Start",
					    "Use this start value for z0 instead of the value from the seed track",
					    _z0Start,
					    float(0.));

	registerProcessorParameter( "FixD0",
				    "Do not vary d0 when fitting (default: false)",
				    _fixD0,
				    bool(false) );
	
	registerProcessorParameter( "FixPhi",
				    "Do not vary phi when fitting (default: false)",
				    _fixPhi,
				    bool(false) );
	
	registerProcessorParameter( "FixOmega",
				    "Do not vary omega when fitting (default: false)",
				    _fixOmega,
				    bool(false) );
	
	registerProcessorParameter( "FixTanLambda",
				    "Do not vary tanLambda when fitting (default: false)",
				    _fixTanLambda,
				    bool(false) );
	
	registerProcessorParameter( "FixZ0",
				    "Do not vary z0 when fitting (default: false)",
				    _fixZ0,
				    bool(false) );
	
	registerProcessorParameter( "NOZFit",
				    "Do not fit in z direction, only use residuals in x-y  / r-phi (default: false)",
				    _noZFit,
				    bool(false) );
}


TrackFitterSimpleChiSquareProcessor::~TrackFitterSimpleChiSquareProcessor() 
{
  delete _trackCollectionParameters;
}


void TrackFitterSimpleChiSquareProcessor::init() {

  printParameters();
  
  // initialise the collection parameters
  _trackCollectionParameters = new IMPL::LCParametersImpl();

  // set the pamameters for this fitter
  _trackCollectionParameters->setValue("TrackFitterType",
				       "SimpleChiSquare");

  _trackCollectionParameters->setValue("SimpleChiSquare_TransDefocussing",
				       _transDefocussing);
  _trackCollectionParameters->setValue("SimpleChiSquare_LongDefocussing",
				       _longDefocussing);
  _trackCollectionParameters->setValue("SimpleChiSquare_TransDiffusionCoef",
				       _transDiffusionCoef);
  _trackCollectionParameters->setValue("SimpleChiSquare_LongDiffusionCoef",
				       _longDiffusionCoef);

}


void TrackFitterSimpleChiSquareProcessor::processRunHeader(LCRunHeader *run) {

    run->parameters().setValue( _processorName + "_revision", "$Rev:$");

    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
        if ( ! i->second->isOptional() || i->second->valueSet() ) {
            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }

}

void TrackFitterSimpleChiSquareProcessor::processEvent(LCEvent *evt) {

    streamlog_out(DEBUG4) << "processEvent"<<std::endl;
    
    IMPL::LCCollectionVec* seedTracks=0;
    try {
	// the tracker hit collection
	seedTracks = 
	  dynamic_cast<IMPL::LCCollectionVec*>
	  (evt->getCollection(_input_seed_tracks_collection_name));
    }
    catch (DataNotAvailableException &) {
	streamlog_out(MESSAGE1) << "No seed tracks collection \""<<_input_seed_tracks_collection_name
				<<"\" in event "<< evt->getEventNumber()<<std::endl;
	return;
    }

    // a collection to hold the new track
    IMPL::LCCollectionVec* outputTrackCollection = new IMPL::LCCollectionVec(EVENT::LCIO::TRACK);

    // set flag that the hits are stored in with the track
    // make sure the hit collection is not transient
    LCFlagImpl trkFlag(0) ;
    trkFlag.setBit( LCIO::TRBIT_HITS ) ;
    outputTrackCollection->setFlag( trkFlag.getFlag() );

    // set parameters for the output collection
    // (rather complicated because the assignment operator cannot be use (it's lcio :-( )
    
    StringVec integerKeys;
    _trackCollectionParameters->getIntKeys(integerKeys);
    for (StringVec::iterator it=integerKeys.begin();
	 it < integerKeys.end(); it ++)
    {
      outputTrackCollection->parameters().setValue( *it,
			      _trackCollectionParameters->getIntVal(*it) );
    }

    StringVec floatKeys;
    _trackCollectionParameters->getFloatKeys(floatKeys);
    for (StringVec::iterator it=floatKeys.begin();
	 it < floatKeys.end(); it ++)
    {
      outputTrackCollection->parameters().setValue( *it,
			      _trackCollectionParameters->getFloatVal(*it) );
    }

    StringVec stringKeys;
    _trackCollectionParameters->getStringKeys(stringKeys);
    for (StringVec::iterator it=stringKeys.begin();
	 it < stringKeys.end(); it ++)
    {
      outputTrackCollection->parameters().setValue( *it,
			      _trackCollectionParameters->getStringVal(*it) );
    }

    streamlog_out(DEBUG3) << "requesting fitter"<<std::endl;

    // request a fitter from the factory
    // a pointer to this event is passed so conditions data can be taken into
    // account
    TrackFitterBase *fitter = TrackFitterFactory::getTrackFitter( TrackFitterBase::FitterTypes::SIMPLECHISQUARE,
								 _trackCollectionParameters);

    if ( parameterSet("D0Start") ){ fitter->setD0Start(_d0Start);}
    if ( parameterSet("PhiStart") ){ fitter->setPhiStart(_phiStart);}
    if ( parameterSet("OmegaStart") ){ fitter->setOmegaStart(_omegaStart);}
    if ( parameterSet("TanLambdaStart") ){ fitter->setTanLambdaStart(_tanLambdaStart);}
    if ( parameterSet("Z0Start") ){ fitter->setZ0Start(_z0Start);}

    fitter->fixD0( _fixD0 );
    fitter->fixPhi( _fixPhi );
    fitter->fixOmega( _fixOmega );
    fitter->fixTanLambda( _fixTanLambda );
    fitter->fixZ0( _fixZ0 );

    fitter->setNoZFitFlag( _noZFit );

    streamlog_out(DEBUG3) << "fitter is at "<<fitter << " fitter type "
			  << static_cast<unsigned short>(fitter->getFitterType())<<std::endl;

    //loop all tracks
    for (IMPL::LCCollectionVec::iterator trackIter=seedTracks->begin(); 
	 trackIter < seedTracks->end() ; trackIter++)
    {
      EVENT::Track *inputTrack = dynamic_cast<EVENT::Track *>(*trackIter);

      if (inputTrack == 0)
      {
	streamlog_out(WARNING2) << "bad_cast: Track candidates collection"
				<< _input_seed_tracks_collection_name
				<< " does not contain EVENT::Tracks!"<< std::endl;
	throw DataNotAvailableException("No track candidates found.");
      }
	

      streamlog_out(DEBUG3) << "fitting track"<<std::endl;

      IMPL::TrackImpl* fittedTrack = fitter->fitTrack(inputTrack);

      // check if fitting was successful
      if ( LCFlagImpl(fittedTrack->getType()).bitSet( TrackFitterBase::FITFAILEDBIT ) )
      {
	_nBadFits++;
      }
      else
      {
	_nGoodFits++;
      }

      streamlog_out(DEBUG3) << "fitting done"<<std::endl;

     //determine dEdx:
      //===============
      //FIXME: This currently calculates the mean charge per hit,
      // not the dE/dx because it is difficult to calculate the actual
      // track length in a general formula
      
      TrackerHitVec const & hitVec = fittedTrack->getTrackerHits();
      
      //      double phi =  outputTrack->getPhi();
      //      double d0 =  outputTrack->getD0();
      double Q_sum = 0;
      
      for ( TrackerHitVec::const_iterator hit = hitVec.begin(); 
	    hit < hitVec.end(); hit++ )
      {
	Q_sum += (*hit)->getdEdx();
      }
      
      fittedTrack->setdEdx( Q_sum/hitVec.size() );
      
      outputTrackCollection->push_back(fittedTrack);

    } // loop tracks

    evt->addCollection(outputTrackCollection, _output_tracks_collection_name);

}// processEvent


Processor* TrackFitterSimpleChiSquareProcessor::newProcessor() {

	return new TrackFitterSimpleChiSquareProcessor();
}


void TrackFitterSimpleChiSquareProcessor::check(LCEvent *) {
}


void TrackFitterSimpleChiSquareProcessor::end() {
  streamlog_out(MESSAGE4) << "Number of good fits: " << _nGoodFits 
			 << "number of bad fits: " << _nBadFits << std::endl;
}



} // namespace marlintpc
