 
// CLHEP
#include <CLHEP/Random/RandGauss.h>
#include <CLHEP/Random/RandFlat.h>

// std
#include <cmath>
#include <cassert>

// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>

// Marlin
#include <marlin/Exceptions.h>

// MarlinTPC
#include "TPCConditions.h"
#include "ElectronCloud.h"
#include "ElectronCloudGEMAmplificationProcessor.h"

using namespace std;
using namespace EVENT;
using namespace IMPL;
using namespace marlin;
using namespace CLHEP;
using namespace tpcconddata;

namespace marlintpc {

// create one to register it with Marlin
ElectronCloudGEMAmplificationProcessor anElectronCloudGEMAmplificationProcessor;


ElectronCloudGEMAmplificationProcessor::ElectronCloudGEMAmplificationProcessor():
	Processor("ElectronCloudGEMAmplificationProcessor"),
	_input_collection_name("TPCElectronClouds"),
	_output_collection_name("TPCElectronClouds"),
	_tpc_conditions_collection_name(TPCConditions::getDefaultColName()),
	_gem_name(""),
	_collection_efficiency(0.),
	_extraction_efficiency(0.),
	_amplification(0.),
	_longitudinal_defocussing(0.),
	_transverse_defocussing(0.),
	_width(0.) {

	registerInputCollection(
		LCIO::LCGENERICOBJECT,
		"InputTPCElectronClouds",
		"Name of the input electron cloud collection",
		_input_collection_name,
		string("TPCElectronClouds") );

	registerOutputCollection(
		LCIO::LCGENERICOBJECT,
		"OutputTPCElectronClouds",
		"Name of the output electron cloud collection",
		_output_collection_name,
		string("TPCElectronClouds") );

	registerOptionalParameter(
		"TPCConditionsName",
		"The name to use for the tpc conditions collection.",
		_tpc_conditions_collection_name,
		TPCConditions::getDefaultColName());

	registerOptionalParameter(
		"GEMName",
		"The name of this GEM.",
		_gem_name,
		string("") );

	registerProcessorParameter(
		"CollectionEfficiency",
		"The parameterized collection efficiency of the GEM [0..1]",
		_collection_efficiency,
		0. );

	registerProcessorParameter(
		"ExtractionEfficiency",
		"The parameterized extraction efficiency of the GEM [0..1]",
		_extraction_efficiency,
		0. );

	registerProcessorParameter(
		"Amplification",
		"The parameterized amplification of the GEM",
		_amplification,
		0. );

	registerProcessorParameter(
		"LongitudinalDefocussing",
		"The parameterized longitudinal defocussing of this GEM (mm)",
		_longitudinal_defocussing,
		0. );

	registerProcessorParameter(
		"TransverseDefocussing",
		"The parameterized transverse defocussing in this section of the TPC (mm)",
		_transverse_defocussing,
		0.);

	registerProcessorParameter(
		"Width",
		"The width of the GEM (mm)",
		_width,
		0.);

}


ElectronCloudGEMAmplificationProcessor::~ElectronCloudGEMAmplificationProcessor() {}


void ElectronCloudGEMAmplificationProcessor::processRunHeader( LCRunHeader* run) {

	// store the processor parameters in the event
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

        if ( ! i->second->isOptional() || i->second->valueSet() ) {

            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
}


void ElectronCloudGEMAmplificationProcessor::processEvent( LCEvent * evt ) {

	// grab the collection of electron clouds
	LCCollection *electron_cloud_collection = NULL;
	try {

		electron_cloud_collection = evt->getCollection( _input_collection_name );
	}
	catch (DataNotAvailableException &e) {

		// the collection of electron clouds with that name wasn't in the LCIO event file, skip simulating this event
		throw marlin::SkipEventException( this );
	}

	streamlog_out( MESSAGE0 ) << "Amplifying " << electron_cloud_collection->getNumberOfElements() << " electron clouds through GEM " << _gem_name << "\n";

	// add this amplification to the tpc conditions data
	TPCConditions* tpc_conditions = NULL;

	try {
		LCCollection *tpc_conditions_collection = evt->getCollection( _tpc_conditions_collection_name );
		tpc_conditions = dynamic_cast<TPCConditions*>( tpc_conditions_collection->getElementAt(0) );
		assert(tpc_conditions);
	}
	catch (DataNotAvailableException &e) {

		LCCollectionVec *tpc_conditions_collection = new LCCollectionVec( LCIO::LCGENERICOBJECT );
		tpc_conditions_collection->setTransient(false);
		tpc_conditions_collection->setSubset(false);
		tpc_conditions = new TPCConditions();
		tpc_conditions_collection->addElement( tpc_conditions );
		evt->addCollection( tpc_conditions_collection, _tpc_conditions_collection_name );
	}

	tpc_conditions->setLongDefocussing( tpc_conditions->getLongDefocussing() + _longitudinal_defocussing );
	tpc_conditions->setTransDefocussing( tpc_conditions->getTransDefocussing() + _transverse_defocussing );
	tpc_conditions->setAmplification( tpc_conditions->getAmplification() * _extraction_efficiency * _amplification );

	// and create a collection for the new, moved clouds (don't write this collection at the end)
	LCCollectionVec *new_electron_cloud_collection = new LCCollectionVec( LCIO::LCGENERICOBJECT );
	new_electron_cloud_collection->setTransient();

	// now loop through the energy deposits creating clouds of electrons
	for( int cloud_number = 0; cloud_number < electron_cloud_collection->getNumberOfElements(); cloud_number++ ) {

		ElectronCloud *current_cloud = dynamic_cast<ElectronCloud*>( electron_cloud_collection->getElementAt( cloud_number ) );
		assert(current_cloud);

		// the number of electrons is stored in dedx
		const int number_of_electrons = current_cloud->get_electron_count();

		streamlog_out( DEBUG3 ) << "Amplifying a cloud with " << number_of_electrons << " electrons\n";

		for( int i = 0; i < number_of_electrons; i++ ) {

			// determine if the electron made it into a hole
			if( (_collection_efficiency >= 1.) || (RandFlat::shoot() > _collection_efficiency) ) {

				const int new_electrons =  (int) ( -1. * _amplification * _extraction_efficiency * std::log(RandFlat::shoot()) );

				if( new_electrons <= 0 ) {

					continue;
				}

				streamlog_out( DEBUG2 ) << "Created a new cloud with " << new_electrons << " electrons\n";

				ElectronCloud *new_cloud = new ElectronCloud();

				// calculate a position for the new cloud
				new_cloud->set_x_position( current_cloud->get_x_position() + (RandGauss::shoot() * _transverse_defocussing) );
				new_cloud->set_y_position( current_cloud->get_y_position() + (RandGauss::shoot() * _transverse_defocussing) );
				new_cloud->set_z_position( current_cloud->get_z_position() + (RandGauss::shoot() * _longitudinal_defocussing) );
				new_cloud->set_transverse_stdev( current_cloud->get_transverse_stdev() );
				new_cloud->set_longitudinal_stdev( current_cloud->get_longitudinal_stdev() );

				new_cloud->set_time( current_cloud->get_time() );
				new_cloud->set_electron_count( new_electrons );

				// and add this hit to the new collection
				new_electron_cloud_collection->addElement( new_cloud );
			}
		}
	}
	
	// now add the collection to the event so other processors may access it
	try {

		// if they have the same name try replacing the input collection (experimental)
		if( this->_output_collection_name == this->_input_collection_name ) {

			LCCollection *old_collection = evt->takeCollection( this->_input_collection_name );

			// remove the old collection...
			for( int cloud_number = 0; cloud_number < old_collection->getNumberOfElements(); cloud_number++ ) {

				ElectronCloud *current_cloud = dynamic_cast<ElectronCloud*>( old_collection->getElementAt( cloud_number ) );
				assert(current_cloud);

				delete current_cloud;
			}
			delete old_collection;
		}
		evt->addCollection( new_electron_cloud_collection, _output_collection_name );
	}
	catch( ... ) {

		throw SkipEventException( this );
	}
	
}


void ElectronCloudGEMAmplificationProcessor::init() {

	// usually a good idea to
	printParameters();
}


void ElectronCloudGEMAmplificationProcessor::check( LCEvent * evt ) { 

}


void ElectronCloudGEMAmplificationProcessor::end() { 

}

}
