#ifndef PrimaryIonisationProcessor_h
#define PrimaryIonisationProcessor_h 1

//LCIO
#include <EVENT/MCParticle.h>
#include <lcio.h>

//MARLIN
#include "marlin/Processor.h"

//C++
#include <string>

//GSL
//#include <gsl/gsl_rng.h>
//#include <gsl/gsl_randist.h>

namespace marlintpc{

/* The following comment is tranfered to doxygen documentation
 * which begins with a second asterix, or in case of on line comment an extra third slash 
 */

/** Processor to create primary ionisation from MCParticles passing
 *  the TPC.
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of MCParticles
 *
 *  <h4>Output</h4> 
 *  Collection of SimTrackerHits
 *
 * @param InputCollectionName Name of input collection, MCParticles
 * @param OutputCollectionName Name of output collection, SimTrackerHits
 * @param Transient If not 0 the output collection is set transient
 * @param OutputType If not 0 all ions in the TPC are returned
 * @param FullCurler Switch detailed treatment of curlers on or off
 * @param ClusterInfo Path to file containing information on ionisation clusters
 * @param GroupElectrons Flag whether to group electrons at the same position
 *        to a single hit,  default: false <br>
 *        WARNING: Turning on the \c GroupElectrons flag currently breaks the default
 *        digitisation chain (DriftProcessor and GEMProcessor). It only works with the
 *        MokkaToVoxelProcessor.
 * @author Astrid Muennich
 */


class PrimaryIonisationProcessor : public marlin::Processor {
  
 public:
  
  virtual Processor*  newProcessor() { return new PrimaryIonisationProcessor ; }
  
  
  PrimaryIonisationProcessor() ;
  virtual  ~PrimaryIonisationProcessor(){ }
  
/** Called at the begin of the job before anything is read.
   */
  virtual void init() ;
  
  /** Called for every run.
   */
  virtual void processRunHeader( lcio::LCRunHeader* run ) ;
  
  /** Called for every event - the working horse.
   */
  virtual void processEvent( lcio::LCEvent * evt ) ; 
  
   
  /*! Reads in values from generator particle needed for track, like vertex, energy, momentum etc.*/
  virtual void SetParticleParameters(lcio::MCParticle* mcparticle);

  /*! Calculates parameters for helix like radius, so tracks can be curled if magnetic field is present. */
  virtual void InitializeTrackPoint();

  /*! Determines number of electrons on track segment using parametrisation obtained with HEED. */
  virtual  void CalculateNumberOfElectrons();
  virtual void CalculateNumberOfClusters();
  virtual void ReadClusterInfoFile();

  /*! Calculates end point of track segment with magnetic field. */
  virtual void CalculateTrackPointwithBfield();

  /*! Calculates end point of track segment without magnetic field. */
  virtual void CalculateTrackPointwithoutBfield();
  int GetTotalNumberOfElectrons(){return InTPC;}

  /*! distributes number of electrons randomly along track segment. */
  virtual  void CalculateClusterDistance();

  /*! Calculates energy loss for helix. */
  virtual  void CalculateEnergyLoss();

  /*! Rotation for delta rays. */
  virtual  void RotateToTrack(double x, double y, double z);

  /*! Rotation for delta rays. */
  virtual  void RotateToDelta(double x, double y, double z);

  /*! Multiple scattering for delta rays. */
  virtual  void MultipleScattering(double phi, double theta);
  virtual  void SetNextStartPoint();
  
  


 protected:

  // Input/output collection names.
  std::string _inputCollectionName;///<Name of input collection
  std::string _outputCollectionName;///<Name of output collection


  std::string clusterfile;  ///<File with parametrisation of clustering processes. 

  //For internal bookkeeping
  int _nRun ;
  int _nEvt ;
 
  //TPC parameters
  float _TPCRadius; ///<Outer radius of the TPC
  float _TPCInnerRadius; ///<Inner radius of the TPC
  float _BField;///<Magnetic field strength

  float _distance;
  int _transient;

  bool _groupElectrons; // flag whether to group electrons at the same position to one hit

  /** For full simulation of Curler set to 0, otherwise standard 1 means simplified
   *  treatment of low momentum particles.
   */
  int Curler;
  /*! Random generator for landau and gaussian. */
  //  gsl_rng *r; 

  /** Fitparameters for calculation of # clusters from Parameterisation from HEED. 
   */
  double a0,a1,a2,a3,a4,a5,a6,a7;
  int numberofelectrons;
  double numberofclusters;
  double clusterdistance;
  double electrons[7633][2];
  /*! #e- in one event */
  int InTPC; 
  // coordinates and variables for track
  /*! Start point of track segment. */
  double xa,ya,za; 
  /*! Vertex of particle. */
  double xv,yv,zv; 
  /*! End point of track segment. */
  double xend,yend,zend, tend; 
  /*! e- coordinate on track segment. */
  double x_e,y_e,z_e,t_e; 
  /*! Length of track segment. */
  double dx,dy,dz,dr; 
  /*! Center of helix . */
  double xm,ym; 
  /*! Scale factor for radius in case of dE/dx.*/
  double dp;
  /*! Angle corresponding to track segment on helix. */
  double alpha;
  // particle variables
  /*! Particle impulses. */
  double px,py,pz,pt,p,p_const; 
  double energy;
  double mass;
  double charge;
  int pdg;
  double delta_x,delta_y,delta_z;
} ;

}//end of namespace marlintpc

#endif

