
// LCIO
#include <EVENT/LCIO.h>

// Marlin
#include <marlin/ConditionsProcessor.h>

// MarlinTPC
#include "Field.h"
#include "GlobalFieldProcessor.h"

using namespace EVENT;
using namespace lccd;
using namespace CLHEP;
using namespace std;
using namespace marlin;

namespace marlintpc {

// create one to register it with Marlin
GlobalFieldProcessor aGlobalFieldProcessor;


GlobalFieldProcessor *GlobalFieldProcessor::_instance = 0;


GlobalFieldProcessor::GlobalFieldProcessor():
	Processor("GlobalFieldProcessor") {

	registerOptionalParameter(
		"MagneticFields",
		"An array of pairs of strings of the format: (field type, lcio collection) defining magnetic fields.",
		_input_magnetic_fields,
		std::vector< std::string >() );

	registerOptionalParameter(
		"ElectricFields",
		"An array of pairs of strings of the format: (field type, lcio collection) defining electric fields.",
		_input_electric_fields,
		std::vector< std::string >() );

	registerOptionalParameter(
		"MainCrossingAngle",
		"The main crossing angle of the ILC (mrads).",
		_main_crossing_angle,
		0. );
}


GlobalFieldProcessor::~GlobalFieldProcessor() {}


Processor *GlobalFieldProcessor::newProcessor() {

	return (&get_instance());
}


void GlobalFieldProcessor::init() {

	// FIXME determine if the steering file is consistant with the simulation
	if( parameterSet("MainCrossingAngle") ) {

		// (assumed that the Mokka parameter was stored with the lcio or GEAR file but I don't see it)
	}

	// parse the processor parameters and register the active magnetic fields
	for( vector<string>::const_iterator input = _input_magnetic_fields.begin(); input != _input_magnetic_fields.end(); input++ ) {

		const string field_typename = (*input++);
		const string collection_name = (*input);

		// create the field
		if( _field_by_typename.find(field_typename) == _field_by_typename.end() ) {

			streamlog_out( WARNING ) << "Field of type: " << field_typename << " has not been registered with GlobalFieldProcessor." << endl;
			continue;
		}

		Field *active_field = _field_by_typename[field_typename]->new_field();

		_active_magnetic_fields.push_back( active_field );

		// now create the object which is responsible for telling the field when it has changed
		FieldNotifier *field_notifier = new FieldNotifier( *active_field );

		if( !ConditionsProcessor::registerChangeListener( field_notifier, collection_name ) ) {

			delete active_field;
			_active_magnetic_fields.pop_back();

			delete field_notifier;

			streamlog_out( WARNING ) << "Could not register collection " << collection_name << " with the LCCD conditions processor." << endl;
		}
		else {

			streamlog_out( MESSAGE4 ) << "Activating " << field_typename << " using collection " << collection_name << '\n';
			_field_notifiers.push_back( field_notifier );
		}
	}

	// parse the processor parameters and register the active electric fields
	for( vector<string>::const_iterator input = _input_electric_fields.begin(); input != _input_electric_fields.end(); input++ ) {

		const string field_typename = (*input++);
		const string collection_name = (*input);

		// create the field
		if( _field_by_typename.find(field_typename) == _field_by_typename.end() ) {

			streamlog_out( WARNING ) << "Field of type: " << field_typename << " has not been registered with GlobalFieldProcessor." << endl;
			continue;
		}

		Field *active_field = _field_by_typename[field_typename]->new_field();

		_active_electric_fields.push_back( active_field );

		// now create the object which is responsible for telling the field when it has changed
		FieldNotifier *field_notifier = new FieldNotifier( *active_field );

		if( !ConditionsProcessor::registerChangeListener( field_notifier, collection_name ) ) {

			delete active_field;
			_active_electric_fields.pop_back();

			delete field_notifier;

			streamlog_out( WARNING ) << "Could not register collection " << collection_name << " with the LCCD conditions processor." << endl;
		}
		else {

			streamlog_out( MESSAGE4 ) << "Activating " << field_typename << " using collection " << collection_name << '\n';
			_field_notifiers.push_back( field_notifier );
		}
	}

}


void GlobalFieldProcessor::processRunHeader( EVENT::LCRunHeader* run ) {}


void GlobalFieldProcessor::processEvent( EVENT::LCEvent * evt ) {}


void GlobalFieldProcessor::check( EVENT::LCEvent * evt ) {}


void GlobalFieldProcessor::end() {

	for( std::vector<FieldNotifier*>::iterator field_notifier = _field_notifiers.begin(); field_notifier != _field_notifiers.end(); field_notifier++ ) {

		if( *field_notifier ) {

			delete *field_notifier;
		}
	}

	for( std::vector<Field*>::iterator field = _active_magnetic_fields.begin(); field != _active_magnetic_fields.end(); field++ ) {

		if( *field ) {

			delete *field;
		}
	}

	for( std::vector<Field*>::iterator field = _active_electric_fields.begin(); field != _active_electric_fields.end(); field++ ) {

		if( *field ) {

			delete *field;
		}
	}

}


/** registers a field with the field manager
*/
bool GlobalFieldProcessor::register_field( Field* field ) {

	const string field_typename = field->get_typename();

	if( _field_by_typename.find(field_typename) == _field_by_typename.end() ) {

		_field_by_typename[field_typename] = field;
	}

	return true;
}


/** Returns the magnetic field (in T) at a given point and time in the detector
*/
CLHEP::HepVector GlobalFieldProcessor::get_global_magnetic_field( const CLHEP::HepVector& x ) const {

	HepVector total_field(3);
	
	for( std::vector<Field*>::const_iterator field = _active_magnetic_fields.begin(); field != _active_magnetic_fields.end(); field++ ) {

		total_field += (*field)->get_field_value( x );
	}

	return total_field;
}


/** Returns the electric field (in V) at a given point and time in the detector
*/
CLHEP::HepVector GlobalFieldProcessor::get_global_electric_field( const CLHEP::HepVector& x ) const {

	HepVector total_field(4);
	
	for( std::vector<Field*>::const_iterator field = _active_electric_fields.begin(); field != _active_electric_fields.end(); field++ ) {

		total_field += (*field)->get_field_value( x );
	}

	return total_field;
}


/** returns the main crossing angle of the ILC model
*/
double GlobalFieldProcessor::get_main_crossing_angle() const {

	return _main_crossing_angle;
}


/** return the instance of the global field processor
	*/
GlobalFieldProcessor &GlobalFieldProcessor::get_instance() {

	if( !_instance ) {

		_instance = new GlobalFieldProcessor();
	}

	return (*_instance);
}


GlobalFieldProcessor::FieldNotifier::FieldNotifier( Field& field ):
	_field(field) {

}


/** from IConditionsChangeListener. this function is called whenever a condition has changed in this collection
*/
void GlobalFieldProcessor::FieldNotifier::conditionsChanged( lcio::LCCollection* col ) {

	_field.construct( col );
}


}	// namespace marlintpc
