//##########################################################################
//                   Programm to merge events together
//
//                   Thorsten Krautscheid 06/11/2007
//
//##########################################################################

/** This programm transforms .slcio files from Mokka to files usable for 
 *  input in the MarlinTPC digitisation. Mokka output is one bunchcrossing 
 *  in one LCIO run consisting of several thousand LCIO events. The MarlinTPC 
 *  digitisation takes one event for one bunchcrossing. This makes a res-
 *  structuring of the runs and events necessary. Only the collections 
 *  usefull for the digitisation are taken (MCParticle and the SimTrackerHits
 *  in the TPC), all other collections are dropped. This program is very 
 *  rudimental and therefore only takes collections with the right names, this
 *  are "MCParticle" for the MCParticle collections and "tpc06_TPC" for the 
 *  SimTrackerHits in the TPC.
 */

//LCIO
#include <lcio.h>
#include <IO/LCWriter.h>
#include <IO/LCReader.h>
#include <EVENT/LCIO.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCTOOLS.h>
#include <IMPL/MCParticleImpl.h>
#include <IMPL/SimTrackerHitImpl.h>

//C++
#include <iostream>
#include <string>
#include <map>

//MarlinTPCTools
#include "LCObjectCopier.h"

using namespace std ;
using namespace lcio ;



//Function to add the parent/child dependencies to the copied MCParticles.
void addFamilyDependencies(map<MCParticleImpl*,MCParticleImpl*> partMap){

  for(map<MCParticleImpl*, MCParticleImpl*>::iterator it = partMap.begin();
      it!=partMap.end();it++){

    it->second->setGeneratorStatus(1);

    for(unsigned int j=0;j<it->first->getDaughters().size();j++){

      map<MCParticleImpl*, MCParticleImpl*>::iterator daughter
        =partMap.find(dynamic_cast<MCParticleImpl*>
                                  (it->first->getDaughters()[j]));
      daughter->second->addParent(it->second);
    }
  }
}


//Function to add the MCParticle dependencies to the copied SimTrackerHits.
void addParticleDependencies(map<SimTrackerHitImpl*,SimTrackerHitImpl*> hitMap
                            ,map<MCParticleImpl*,MCParticleImpl*> partMap){

  for(map<SimTrackerHitImpl*,SimTrackerHitImpl*>::iterator it=hitMap.begin();
      it!=hitMap.end();it++){

    MCParticleImpl* myMCParticle=dynamic_cast<MCParticleImpl*>
                                             (it->first->getMCParticle());

    it->second->setMCParticle(partMap[myMCParticle]);                        
  }
}


void printUsage()
{
  cout << endl  << "MokkaToDigitisationConverter [infile.slcio] [outfile.slcio]" << endl
                << "outfile name is optional, "                                   << endl  
                << "default is ConvertedBackground.slcio"                         << endl;
}


int main(int argc,char **argv){

  if(argc!=2&&argc!=3) {
    printUsage();
    return(1);
  }

  //Get infile name
  string inFileName = argv[1];
  
  //Get outfile name
  string outFileName;
  if (argc==3)
    outFileName=argv[2];
  else
    outFileName="ConvertedBackground.slcio";

  try{
    // create sio writer/reader
    LCWriter* lcWrt=LCFactory::getInstance()->createLCWriter();
    LCReader* lcRd =LCFactory::getInstance()->createLCReader();
    try{ 
       //Open outfile/infile.
       lcWrt->open( outFileName, LCIO::WRITE_NEW ) ;
       lcRd->open( inFileName );
    }catch(IOException& e){    
      cout << e.what() << endl ;
      return 0 ;
    }
    //Get old runheader.
    LCRunHeader* oldRunHdr=lcRd->readNextRunHeader();

    //Create runheader and write it.
    LCRunHeaderImpl* runHdr=new LCRunHeaderImpl;
    runHdr->setRunNumber(0);
    runHdr->setDetectorName(oldRunHdr->getDetectorName());

    //Calculate event number. The -1 is to start from zero.
    int evtNr=oldRunHdr->getRunNumber()-1;

    //Create ouput collections.
    LCCollectionVec* tpcHitsVec=new LCCollectionVec(LCIO::SIMTRACKERHIT);    
    LCCollectionVec* mcPartVec=new LCCollectionVec(LCIO::MCPARTICLE);

    while(1){
      LCEvent *mergeEvt=lcRd->readNextEvent();
      if(mergeEvt==NULL)
	break;
      else{
	try{
	  //Get the collection with TPC hits. If the collection name
	  //differs from "tpc06_TPC" it must be changed!
	  LCCollection* hitsCol = mergeEvt->getCollection("tpc06_TPC");
 
	  //Fill hits with copies in map.
          map<SimTrackerHitImpl*,SimTrackerHitImpl*> hitMap;
	  for(int i = 0; i < hitsCol->getNumberOfElements(); i++){

            hitMap[dynamic_cast<SimTrackerHitImpl*>(hitsCol->getElementAt(i))]
		=LCObjectCopier::copy(dynamic_cast<SimTrackerHitImpl*>
                                                  (hitsCol->getElementAt(i)));
	  }
	  //Get collection with MCParticles. If the collection name differs
	  //from "MCParticle" it must be changed!
	  LCCollection* partCol = mergeEvt->getCollection("MCParticle");

          //Fill MCParticles and copies in map.
	  map<MCParticleImpl*,MCParticleImpl*> partMap;
	  for(int i=0;i<partCol->getNumberOfElements();i++){

	    partMap[dynamic_cast<MCParticleImpl*>(partCol->getElementAt(i))]
                =LCObjectCopier::copy(dynamic_cast<MCParticleImpl*>
                                                  (partCol->getElementAt(i)));
            }


	  //Function name speaks for itself.
	  addFamilyDependencies(partMap);

	  //Add the McParticle dependencies of the SimTrackerHits.
	  addParticleDependencies(hitMap,partMap);

	  //Transform energie depositions into charge values and add hits 
	  //to collection. Also transform time from [ns] to [us].
	  for(map<SimTrackerHitImpl*,SimTrackerHitImpl*>::iterator it
              =hitMap.begin();it!=hitMap.end();it++){

	    //Set charge. This is dEdx divided by the mean ionisation energy
	    //of argon.
	    int charge=static_cast<int>(it->first->getdEdx()/0.000000026);
	    it->second->setdEdx(charge);

	    //Transform time from [ns] to [us].
	    it->second->setTime(0.001*it->first->getTime());

	    //Add hits to to collection.
            tpcHitsVec->addElement(it->second);
	  }

	  //Add MCParticle to collection.
	  for(map<MCParticleImpl*,MCParticleImpl*>::iterator it
              =partMap.begin();it!=partMap.end();it++){
            mcPartVec->addElement(it->second);	    
	  }
	}catch(DataNotAvailableException& e){}
      }
    }
    //Close reader.
    lcRd->close();

    //Create new lcio event.
    LCEventImpl* evt=new LCEventImpl();

    evt->setRunNumber(0);
    evt->setEventNumber(evtNr);
    evt->setDetectorName(runHdr->getDetectorName());

    //Add the collections to the event.
    evt->addCollection(tpcHitsVec,"MokkaTPCHits") ;
    evt->addCollection(mcPartVec,"MokkaMCParticles");
      
    //Write the event to the file
    lcWrt->writeEvent(evt) ;

    //Close writer.
    lcWrt->close();
	
    //Dump the event to the screen.
    LCTOOLS::dumpEvent( evt ) ;

    //Clean up.
    delete evt ;
    delete runHdr ;

  }catch(std::exception& ex){
    cout << " an excpetion occured: " << endl
         << "  " << ex.what()         << endl ;
    return 1 ;
  }
     
  return 0;

}
